# ls.py
# Functions to support DVR+ LinearSchedule calls
# Copyright June 2020, Timothy Goihl

# SD API reference
# https://github.com/SchedulesDirect/JSON-Service/wiki/API-20141201

from time import time, gmtime, strftime, mktime
import json
import copy
import logging
from datetime import timedelta, datetime, timezone
from dateutil import parser
import sys
import time
#import gc  # garbage collection, not needed?

from ut import SendEmailAlert
from ut import LocalTimeStr

# Calls we need to make to SD
from sd import SDGetStationSchedule, SDSetAPIVersion
from sd import SDGetProgramDetails
from sv import GetServiceList


# The is the message header DVR+ expects when reuesting a schedule
# The Linear Schedule Result Header, a python dictionary
lsr = {
  "LinearScheduleResult": {
    # String GUID
    "RequestId": "6be49fd4-46b7-4429-acc2-1d385a441864",
    "TimeStamp": "2020-05-02T16:04:06.396Z",
    "Status": "PNE",  #PNE Processed No Errors
    "Errors": [],
    "EndTimestamp": "0001-01-01T00:00:00",
    "Build": "Development-v9-dev-9.15.08-173-r34606-34606",
    "Schedule": {
      "Locale": "en-US",
      "ServiceId": 12345,
      "Name": "Madison Area Channels - Broadcast TV",
      "ServiceType": "Broadcast",
      "StartDate": "2020-05-02T16:00:00Z",
      "Duration": 240,
      "TimeZones": [
        {
          "StartDateTime": "2020-03-08T08:00:00Z",
          "EndDateTime": "2020-11-01T07:00:00Z",
          "Offset": -300
        }
      ],
      "Airings": []    # empty list of airingDefs
    }
  }
}        

# airingDef
# Definition of the one 'program airing' object.  Fields will be updated with SD data
# Each display slot on the DVR+ Guide is one of these.  I doubt that DVr+ uses it all.
# DVR+ probably copies out the few things it needs and discards the rest.

COPY_MAX_CHARS = 512    # measured as 512 chars by expiriment

airingDef = {
      "ProgramId": "21085335",  # <10 chars?, seems to work with SD 12-char Program ID's, luckily. 
      "SeriesId":"1234",        # not used or changed
      "ServiceId": 12345,       # must match service id 99999 everywhere
      "SourceId": 11833,        # must match SourceID in services (StationID)
      "ParentNetworkId": 2,     # this will be set to the channels 'order' for a unique #
      "Channel": "3.1",
      "CallLetters": "WISCHD",
      "SourceDisplayName": "xx",
      "SourceLongName": "xx",
      "Title": "Prm Title", 
      "Copy": "~512 chars.",    # measured as 512 chars by experiment
      "AiringTime": "",         # 2020-05-02T05:00:00Z
      "Duration": 1440,         #DVR+ wants mins, SD supplies seconds, /60
      "AiringType": "Unknown",  # this is the magic 'New' status that allows 'record all new programs'
      "TVRating": "None",       # TV-14 etc
      "IconAvailable": False,
      "EpisodeTitle": "",
      "CC": False,              # These are the defaults, if SD keys are found
      "Stereo": False,          # they will be updated
      "SAP": False,
      "Dolby": False,
      "DVS": False,
      "HD": False,
      "HDLevel": "HD Level Unknown",  # HD Level Unknown,480i,480p,720p,1080i,1080p
        # I dont mess with any of these below this line
      "InstanceId": None,
      "UnknownInstance": False,
      "UPId": "autofill",
      "Color": "Color",
      "LetterBox": False,
      "ShowingType": "Unknown",  #Finale,Paid For,Premiere,Season Finale,Season Premiere,Series Finale,Series Premiere,Unknown
      "DSS": False,
      "Sports": False,
      "Category": "Other",
      "Subcategory": "other",
      "TiVoSupported": True,
      "DvbTriplet": "",
      "SL": "false",            # green text "false" in wireshark from CM
      "Subtitles": "false"      # green text "false" in wireshark from CM
    }

# Program details if there is a SD error

SDPD_Error = {
    "programID": "EP000000060003",
    "titles": [{"title120": "Error Title"}],
    "descriptions": {
        "description100": [{ "description": "Description not available. SD Code()"}]
    },
    #"episodeTitle150": "PiGS is Great!",
    "md5": "Jo5NKxoo44xRvBCAq8QT2A"
}

from ut import total_size
#
# Build a DVR+ compatibe schedule by downloading the full SD schedule
# for all of the users favorite stations.  The downloaded SD data  
# will be reformatted into the json format required by the DVR+ and then
# added to a airing-time sorted list.  The sorted list will be easy to 
# traverse for GetLinearSchedule() to return data to DVR+.

# This will be a time-ordered list of SD programs reformatted into
# DVR+ json Airings format.  It is stored internally as a Python dictionary
# but converting a dict to json is 1 line of code.

from operator import itemgetter  #, attrgetter  # for sorting channels/airings into time order

# The ProgramSchedule list will be built from SD data and then kept in memory.
# A DVR request for schedule data does not go out to SD in realtime, it just uses this internal database
# The ProgramSchedule will be updated by a timer process about an hour before a DVR wants it

ProgramSchedule = []
LastBuiltTime = 0
BuildProgress = 0

# SD's new server model limits how often an Episode or Show ID can be requested
# v1.25 Need to keep a cache of everything requested during this build, maybe on disk???

ProgramsCache = {}   # { programID : {program details}, ...}

#https://www.w3schools.com/python/python_file_handling.asp

import pickle
from ut import PigsPath

SCHED_FILE = 'ProgramSched'
CACHE_FILE = 'ProgramCache'

# SaveLevel: 0=Dont Save, 1=SaveCache, 2=Save Boths Progs & Cache

# Pickle(Save) an object to disk in 'ObjectName.pkl'

def utSaveObjectToDisk(ObjectToSave, ObjectName) :
   
    if len(ObjectToSave) : 
        logging.info("  Saving %s to disk (%d items)", ObjectName, len(ObjectToSave))
        
        psf = PigsPath(ObjectName+'.pkl')
        
        with open(psf, "wb") as outfile: 
            pickle.dump(ObjectToSave, outfile, protocol=4)
            outfile.close()
    else :
        logging.info("  %s - Nothing to Save.", ObjectName)  #, len(ProgramSchedule))


def utLoadObjectFromDisk(ObjectToLoad, ObjectName) :

    psf = PigsPath(ObjectName+'.pkl')
    
    try:
        f = open(psf, "rb")       
        ObjectToLoad = pickle.load(f)  
        f.close()             

    except FileNotFoundError:
        logging.info("  %s file not found. Nothing loaded.", psf)
        
    except Exception as ex:
        logging.warning("  %s Exception. Err=%s", ObjectName, ex)
    else :
        logging.info("  %s Loaded from disk (%d items)", ObjectName, len(ObjectToLoad))

    pcf = PigsPath(CACHE_FILE)
    

def xxxLSSaveProgramsToDisk(SaveLevel = 2 ) :
    global ProgramsCache
    global ProgramSchedule
    
    if SaveLevel == 2 :
        if len(ProgramSchedule) : 
            logging.info("  Saving ProgramSchedule to disk (%d items)", len(ProgramSchedule))
            
            psf = PigsPath(SCHED_FILE)
            with open(psf, "wb") as outfile: 
                pickle.dump(ProgramSchedule, outfile, protocol=4)
                outfile.close()
        else :
            logging.info("  ProgramSchedule - Nothing to Save.")  #, len(ProgramSchedule))
    
    if SaveLevel >= 1 :
        if len(ProgramsCache) :
            logging.info("  Saving ProgramCacheto disk (%d items) ", len(ProgramsCache))

            pcf = PigsPath(CACHE_FILE)
            with open(pcf, "wb") as outfile: 
                pickle.dump(ProgramsCache, outfile, protocol=4)
                outfile.close()
        else :
            logging.info("  ProgramsCache - Nothing to Save.")


def LSLoadProgramsFromDisk() :
    
    global ProgramsCache
    global ProgramSchedule

    utLoadObjectFromDisk(ProgramSchedule, SCHED_FILE)

    utLoadObjectFromDisk(ProgramsCache,   CACHE_FILE)
    
def xxxLSLoadProgramsFromDisk() :
    global ProgramsCache
    global ProgramSchedule

    psf = PigsPath(SCHED_FILE)
    
    try:
        f = open(psf, "rb")       
        ProgramSchedule = pickle.load(f)  
        f.close()             

    except FileNotFoundError:
        logging.info("  ProgramSchedule file [%s] not found.", SCHED_FILE)
        
    except Exception as ex:
        logging.warning("  ProgramSchedule Exception. Err=%s", ex)
    else :
        logging.info("  ProgramSchedule Loaded from disk (%d items)", len(ProgramSchedule))

    pcf = PigsPath(CACHE_FILE)
    
    try:
        f = open(pcf, "rb")
        ProgramsCache = pickle.load(f)
        f.close()

    except FileNotFoundError:
        logging.info("  ProgramCache file [%s] not found...", CACHE_FILE)
        
    except Exception as ex:
        logging.warning("  ProgramCache Exception. Err=%s", ex)
    else :
        logging.info("  ProgramCache Loaded from disk (%d items)", len(ProgramsCache))
    

class BuildStatus :
    
    sectionHeader = 'Schedule Build Info'
    
    _BuildInfo = { # create template here so stuff is in nice order
       sectionHeader : {  # section header for display
        "BuilderAlive": False, 
        "DaysLimit": 0,
        "StartTime": "",
        "BuildTime": "",
        "LastBuild": "",
        "Programs_": "",
        "SchdStart": "",
        "SchedEnds": "",
        "NextBuild": "",
        "AutoBuild": ""
      }
    }
    
    def __init__(self) :
        self.bi = self._BuildInfo[self.sectionHeader]
        self.reset(0,"","")
        
    def reset(self, dl, tstr, bt) :
        self.bi['DaysLimit'] = dl
        self.bi['StartTime'] = tstr
        self.bi['BuildTime'] = bt 
        #self.bi["LastBuild"] = ""  #only updated when build succeeds
        #self.bi["Programs_"] = ""
        #self.bi["SchdStart"] = ""
        #self.bi["SchedEnds"] = ""

    def SetAlive(self, state) :
        self.bi['BuilderAlive'] = state
    def SetDaysLimit(self, days) :
        self.bi['DaysLimit'] = state
    def SetStartTime(self, startT) :
        self.bi['StartTime'] = startT
    def SetBuildTime(self, buildT) :
        self.bi['BuildTime'] = buildT
    def SetLastBuildTime(self, buildT) :
        self.bi["LastBuild"] = buildT
    def SetPrograms(self, programs) :
        self.bi['Programs_'] = programs
    def SetSchStart(self, SchStart) :
        self.bi['SchdStart'] = SchStart
    def SetSchEnd(self, SchEnd) :
        self.bi['SchedEnds'] = SchEnd
    def SetNextBuild(self, buildT) :
        self.bi['NextBuild'] = buildT
    def SetAutoBuild(self, buildT) :
        self.bi['AutoBuild'] = buildT


# Build State for information

LSBuildStat = BuildStatus()

#*****************************************************************************
# Key Mapping between various json objects
#
# ServicesResult    ServiceDetailResult   Linear Schedule Result     SD
# 
#  ServiceID
#
# Reference: SD API
# https://github.com/SchedulesDirect/JSON-Service/wiki/API-20141201
#
# BuildLinearSchedule creates the ProgramSchedule[]


#def LSClearSchedule() :
#    ProgramSchedule.clear()
#    return

def LSGetProgramSchedule() :
    return ProgramSchedule

# Return value
# 0 = ok
# 100 No Service List
#
from enum import IntEnum, auto

class BuildError(IntEnum) :
    Success        = 0    # (SD code 200 converts to 0)
    NoServiceList  = 10   # cant build without it
    AbortEvent     = 20   # external abort request 
    GetPrgSchedule = 30   # SD error causing abort
    ProgramDetails = 40   # SD error causing abort
 


AllPrgSet = set()  # list of unique programID's found during this build

def BuildLinearSchedule(PiGScfg, serviceID, AbortEvent,updateTxt):

    global ProgramSchedule
    global LastBuiltTime
    global DictOfDVRs
    global BuildInfo

    BuildResult = BuildError.Success  # Default to Success
    
    # make sure code is a function level variable
    code = 0
    
    # Build schedule into Temp list, then update Master list if success
    
    NewSchedule = []  # build schedule into a temp list
    
    # get list of users Favorite Stations / channels 
    # it is a dict, the same list sent to DVR+ when it requests it in Message Type 2
    stationList = GetServiceList()

    if not stationList :
        logging.warning("LS:BLS: There are no configured Stations! Nothing to do!")
        return BuildError.NoServiceList

    # for computing % complete.  +1 for computing the final size
    TotalChannels = stationList[-1]["Order"] + 1   # = len(stationList)

    BuildStartTime = datetime.utcnow()
    ts,tz = LocalTimeStr( BuildStartTime,showSeconds=True )

    logging.info("\nLS:BLS: ****** Start Schedule build %s ****** ", PiGScfg['PiGS_Version'])
    logging.info("LS:BLS: Start Time : %s", ts)
    logging.info("LS:BLS: SD API Version : '%s'", SDSetAPIVersion() )
                 
    # check if there is a DaysLimit in the config file (limit data days for testing)
    DaysLimit = PiGScfg.get('DaysLimit', 0)
    if DaysLimit :
        logging.info("LS:BLS: Limiting SD request to %d days(plus yesterday).", DaysLimit)

    # Clear out Build Status
    LSBuildStat.reset(DaysLimit, ts, bt="Rebuilding... 0%")

    # build a fresh new list - complete refresh  BuildModmeDelete=true, BuildModeUpdate(false)
    if PiGScfg["BuildModeDelete"] :
        ProgramSchedule.clear()
    
    # build a list of program ID's as we go. will make getting descriptions faster
    PgmIdSet = set()
    
    AllPrgSet.clear()  # Clear list at start of each build.
    
    # Clear or Load the list from disk?
    # Should it be cleared when it gets Stale?
    if PiGScfg["BuildModeDelete"] :
        ProgramsCache.clear()
        
    # program airings will be built in a temp list, then be added to the master Airing list
    # we do this so we can add the program details just to this partially complete list
    # of airings
    ChanSchedule = []
    
    #Description Enhancements?
    iconnames = ['None(0)','Schedule Direct Icons(1)','CM/Rovi Icons(2)','???(3)','???(4)'] 
    logging.info("LS:BLS: CAST List : %s", PiGScfg['AppendCast'])
    logging.info("LS:BLS: Icon Type : %s", iconnames[PiGScfg['IconType']])
    logging.info("LS:BLS: Episd Info: %s", PiGScfg['PrependEpisode'])


    #==================================================================================
    # Loop through the station/channel objects in the Services List
    # The stationList has the Channel ID integers we need to get data for.
    
    BuildAbort = False   # Method to break out of Build process
    
    for aStation in stationList :

        # Method to break out of Build process if a processing (SD) error occurs below
        if BuildAbort :
            #BuildResult = BuildError.AbortEvent
            break   

        # Check for Abort Event, and convert it into a Build Abort
        if AbortEvent.is_set() :
            BuildAbort = True
            BuildResult = BuildError.AbortEvent
            #AbortEvent.clear()  Builder will clear it
            break

        # SD treats Station ID as a string, CM wants an integer
        
        iSourceId = aStation["SourceId"]  #int  DVR+ call stations SourceId as an integer
        
        # skip channels marked as 0 so that the channels can exist in the Service Lineup
        # This allows new stations to have an image icon before the schedule is available
        
        if iSourceId == 0 : continue
        
        # Part1 ========================================================================
        # Get the full Schedule for 1 station from Schedules Direct
        # Then scan thru SD's json and build a DVR+ compatible 'airings' list
        
        # SDGetStationSchedule tries to get yestday plus all days forward (limited by DaysLimit)
        # the returned data is all dates available for 1 station
        # DVR+ can only use 14 days of data, so ndays should be called DaysLimit as 14

        logging.info ("  Downloading Schedule for Channel %s", aStation["Channel"])
        
        code, schj = SDGetStationSchedule( str(iSourceId), DaysLimit )
        #print('st sch ', iSourceId)
        
        # A 200 return, provides either an error block
        #[
        #  {
        #    "stationID": "25124",
        #    "serverID": "20141201.web.1",
        #    "code": 7020,
        #    "response": "SCHEDULE_RANGE_EXCEEDED",
        #    "minDate": null,
        #    "maxDate": null,
        #    "requestedDate": "2020-09-22",
        #    "message": "Date requested (2020-09-22) not within -> for stationID 25124."
        #  }
        #]
        # or program data
        #[
        #  {
        #    "stationID": "20454",
        #    "programs": [
        #            { prog keys },
        #    ],
        #  }
        #]

        if code != 200 :
            logging.warning("LS:BLS: SDGetStationSchedule() http Error(%d)- Schedule Not Updated!", code)
            BuildAbort = True   # Method to break out of Build process
            BuildResult = BuildError.GetPrgSchedule
            break
        
        # clear the temp lists we use on each station iteration
        PgmIdSet.clear()
        ChanSchedule.clear()
 
        # SD Schedule data is in multiple blocks by date with a stationID header
        # There is also a differnt type of error header with a non-zero code.
        # we dont know which we will get

        StaleMd5CacheSet = set()  # Maintain list of items Removed from Cache due to md5
        
        for aSdStationId in schj :  # aSdStationId for Station Date element, find the stationID's
            
            #check for an error code, good blocks dont have an error code
            # get() does not fail. if 'code' not available, its sched data, default to 0=ok
            rcode = aSdStationId.get('code',0)
            
            # check for error code in json data
            # its probably a 7020 no-date error since I tried to get yesterday
            # 'yesterday' exists for a while (and is needed around 7pm cst)
            if rcode > 0 :
                logging.debug("LS:Build Found code %d", rcode)
                continue;
             
            # Get References to the 3 sections of the schedule data
            # ToDo, do i need try/catch or use Get() to prevent key error crashes?
            
            sids= aSdStationId['stationID']  # SD stationID or the Airing SourceId
            prl = aSdStationId['programs']   # programs list
            met = aSdStationId['metadata']   # metadata, md5 stuff

            # for each date in the returned schedule list, parse the programs
            
            for aProgram in prl :  # aProgram scan the SD Programs results List
                
                # append a copy of a new airing object (the object format DVR+ needs)
                ChanSchedule.append( copy.deepcopy(airingDef) )

                # SD 'Schedule' has no program details - must request separately
                # Compile a list of programID's for Stage 2
                
                pidstr = copy.copy( aProgram["programID"] )  #programID is a string
                
                PgmIdSet.add(pidstr) # Python Sets automatically strip duplicates!
                
                # if pid is Cached, see if md5 changed, remove from cache so it will be refetched
                # with a persistent Cache between builds, item in 1st Channel could be in Cache
                
                cpd = ProgramsCache.get(pidstr, None)
                                  
                if cpd :    # if pidstr in our Cache, check if md5 matches
                    if cpd['md5'] != aProgram['md5'] :
                        logging.debug("  Removing Stale Program(%s) from Cache...", pidstr)
                        StaleMd5CacheSet.add(pidstr)  # keep track of removed Cache entries
                        ProgramsCache.pop(pidstr)
                
                # Now Copy what we can from the SD Schedule items
                
                # Attempt to get SD 'new' program key, default to Not new, does not always exist
                
                newflag= aProgram.get('new', False)
                
                # DVR+ text must be one of these strings: Unknown New Live Repeat Taped
                newstr = "New" if newflag else "Unknown"  # ternary assignment
                ChanSchedule[-1]["AiringType"]= newstr
               
                #ToDo what if any these keys do not exist

                ChanSchedule[-1]["ProgramId"]       = pidstr
                ChanSchedule[-1]["ParentNetworkId"] = aStation['ParentNetworkId'];               
                
                #DVR+ requires the channel number in each Airing
                ChanSchedule[-1]["Channel"]   = aStation['Channel']   # like 3.1, copy from config
                ChanSchedule[-1]["ServiceId"] = serviceID    # 99999 must match in airings
                ChanSchedule[-1]["SourceId"]  = iSourceId    # SourceID must match service header
                # Airing time format matches directly - one thing I dont have to modify!!
                ChanSchedule[-1]["AiringTime"]= copy.copy( aProgram["airDateTime"])
                # DVR+ expects minutes, SD supplies seconds
                ChanSchedule[-1]["Duration"]  = copy.copy( int(aProgram["duration"]/60) )
                
                # copy from Settings file info about a station
                ChanSchedule[-1]["CallLetters"]       = aStation["CallLetters"]
                ChanSchedule[-1]["SourceDisplayName"] = aStation["DisplayName"]
                ChanSchedule[-1]["SourceLongName"]    = aStation["FullName"]

                # If SD has audio properties... they are optional
                aplst = aProgram.get("audioProperties")  
                if aplst :
                    #logging.info('audio %s', aplst)
                    ChanSchedule[-1]["CC"]     = 'cc' in aplst   # True shows CC flag, OTA over-rides
                    ChanSchedule[-1]["Stereo"] = 'stereo' in aplst
                    ChanSchedule[-1]["SAP"]    = 'SAP' in aplst
                    ChanSchedule[-1]["DVS"]    = 'dvs' in aplst
                
                # If SD has video properties... they are optional
                vplst = aProgram.get("videoProperties") 
                if vplst :
                    # SD list looks like this ['HD 1080i', 'hdtv']
                    # logging.info('ch %s: video %s', aStation['Channel'], vplst)
                    ChanSchedule[-1]["HD"] = 'hdtv' in vplst  # True="HD", False="SD"  OTA over-rides
                    
                    # DVR+ wants 'HD Level Unknown','480i','480p','720p','1080i','1080p'
                    # just one of these should be true - brute force seems easiest
                    if 'HD 480i'  in vplst : ChanSchedule[-1]["HDLevel"] = "480i"
                    if 'HD 480p'  in vplst : ChanSchedule[-1]["HDLevel"] = "480p"
                    if 'HD 720p'  in vplst : ChanSchedule[-1]["HDLevel"] = "720p"
                    if 'HD 1080i' in vplst : ChanSchedule[-1]["HDLevel"] = "1080i"
                    if 'HD 1080p' in vplst : ChanSchedule[-1]["HDLevel"] = "1080p"

                # content ratings - dont know if the SD ratings are compatible, i'm just
                # going to stuff whatever SD has into the DVR+ field
                # DVR+ seems to want: None,TV-G,TV-PG,TV-14,TV-Y works TV-MA does not
                # "ratings": [
                #    {
                #        "body": "USA Parental Rating",
                #        "code": "TV14"
                #    }
                # ]
                
                rtlst = aProgram.get("ratings")  # Ratings properties are optional
                if rtlst :
                    if len(rtlst) > 1 :
                        logging.info('Ch %s: has %d ratings.', aStation['Channel'], len(rtlst) )
                    
                    # search List for "USA Parental Rating" 
                    for rating in rtlst :
                        if rating['body'] == "USA Parental Rating" :
                            break
                    
                    if rating['body'] == "USA Parental Rating" :
                        #logging.info('ratings %s', rtlst)
                        ratingstr = rating.get('code')  # SD "body" : "USA Parental Rating"
                        ChanSchedule[-1]["TVRating"] = ratingstr
                    else :
                        #ratingstr = rtlst[0].get('code')  # otherwise take the first one?
                        ChanSchedule[-1]["TVRating"] = "None"     # fill with empty 

        # Part1 Post Process =================================================================
        # In Apr 2022 SD limits number of duplicate ID's requested in 5 minutes to approx 10
        # Decide what to request from SD, whats already cached, what to save in cache
        
        # Cache algorithm is :
        # 1st time a PgmID is encountered it is looked up and it is added to AllPrgSet
        # 2nd time a PgmID is encountered we know its a Repeat, we look it up, but we will save it
        # 3nd time a PgmID is encountered we know its a Repeat, and in cache, we DONT look it up
        #   it will be retrieved from Cache going forward
        # Since AllPrgSet is reset on each build, there is a minor issue with items
        #  already in Cache, if Md5 changed and Cache item is removed, it may not be a 'Repeat'
        #  on first few Chans, but we know we need to resave StaleMd5CacheSet items in Cache
        
        CachedSet = set(ProgramsCache)  # get a set of pid's of Cached Programs

        # Determine what to lookup, what we have Cached, what we need to save in Cache
        LookupSet = PgmIdSet  - CachedSet     # Lookup all ID's except those we have Cached
        RepeatSet = AllPrgSet & PgmIdSet      # List of ID's we've seen before (in both sets) Intersection
        ToSaveSet = RepeatSet - CachedSet     # Remove those already Cached, subset to save in Cache
        ToSaveSet.update(StaleMd5CacheSet)    # If just removed from Cache due to md5, ensure re-save
        AllPrgSet.update(PgmIdSet)            # Add this channels ID's to AllPrgSet
        
        PgmIdList = list(LookupSet)           # Create a List() for SD lookup API
        
        if len(StaleMd5CacheSet) :
            logging.info("  Removed (%d) Stale Programs from Cache...", len(StaleMd5CacheSet) )
            #logging.info("LS:BLS: CachedSet(%d) StaleMd5CacheSet(%d) ToSaveSet(%d) SCS:%s TSS:%s", \
            #             len(CachedSet), len(StaleMd5CacheSet), len(ToSaveSet), StaleMd5CacheSet, ToSaveSet )
        
        logging.debug("LS:BLS: CachedSet(%d) PgmIdSet(%d) RepeatSet(%d) LookupSet(%d) ToSaveSet(%d) AllPrgSet(%d)", \
            len(CachedSet), len(PgmIdSet), len(RepeatSet), len(LookupSet), len(ToSaveSet), len(AllPrgSet) )
        
        #if '20.1' == aStation['Channel'] :  # print actual contents of lists 
        #    print('CachedSet', (CachedSet),'RepeatSet', (RepeatSet), 'LookupSet', (LookupSet), \
        #          'ToSaveSet', (ToSaveSet),'PgmIdList', (PgmIdList),'AllPrgSet', len(AllPrgSet) )

        StaleMd5CacheSet.clear()
        
        # Part2 ========================================================================
        # Reqest SD program details for each program airing we just added and fill in
        # 
        # There are about 336(24 1-HR listing*14days) to 672(48 30-MIN listing*14days)
        # programs in the list (550 typical) for each day
        
        # request program details from Sched Direct for this channel in one big block
        # The SD call lets you request multiple ProgramID at one time, up to 5000 ID's
    
        logging.info("  Retrieving Program Details from Schedules Direct..." )

        BlockLen   = 1000   # my limit per api call - robert K says up to 5000 should work
        BlockStart = 0
        ListLen  = len(PgmIdList)

        pcode = 200  # Default in case loop doesnt execute
        pdj   = []
        
        while ListLen > BlockStart : # loop until all ID retrieved
            
            if BlockStart > 0 : # msg when doing more than 1 block
                logging.info("LS:BLS: Chan %s has %s programs! Fetching block %d", \
                             aStation['Channel'],ListLen, int(BlockStart/BlockLen)+1 )
                
            list1 = PgmIdList[BlockStart:BlockStart+BlockLen] # slice out a block of ID's
            
            # Get list of programs from SD
            pcode, pdj1,msg = SDGetProgramDetails( list1  )
            
            if pcode != 200 :   # some kind of bad error - clean up and exit
                break
            
            pdj.extend(pdj1)    # add this block to list
            
            BlockStart += BlockLen
            
        if pcode != 200 : # some kind of bad error - clean up and exit
            logging.warning("LS:BLS: Could not get program details. Code(%d)", pcode)
            BuildAbort = True       # Method to break out of Build loops
            BuildResult = BuildError.ProgramDetails
            break
            
        # ========================================================================
        # The recieved list could have actual Program data, or an error block with code=600x
        # Scan received programs for errs & build a fast lookup dict{} of the returned list
        # Also save Programs flagged in ToSaveSet in ProgramsCache for repeat use
        
        ProgramsLookupDict = {}    # use a Dict for fast/easy lookup into SD's returned list

        for pd in pdj: # loop and add/replace programs, could be 0 items if all in Cache
            
            pidstr = pd.get('programID')
            
            code   = pd.get('code', 0)  # 'code' key does not exist in valid response
            
            #if pidstr == 'SH000000010000' : code = 6003 # fake error for testing 'Paid Programming'
            
            if code == 0 :

                ProgramsLookupDict[pidstr] = pd     # reference to a list item
                
                if pidstr in ToSaveSet :  # Save repeat Programs in Cache, stamp time added
                    pd['LastAccessed']    = datetime.utcnow()  # add timestamp to new items
                    ProgramsCache[pidstr] = copy.deepcopy(pd)  
            else :
                logging.warning("LS:BLS: Program %s returned code(%d)", pidstr, code)
               
                # Create an error 'Program' that will display on guide as clue to user
                SDPD_Error["programID"] = pidstr
                SDPD_Error["titles"][0]["title120"] = pidstr + '(' + str(code) + ')'
                SDPD_Error["descriptions"]['description100'][0]["description"] = \
                                            "Description not available. SD Code" + '(' + str(code) + ')'
                # Save error description 
                ProgramsLookupDict[pidstr] = copy.deepcopy(SDPD_Error) 
                
        # ========================================================================
        # Now cycle through temporary list of airings and fill in Pgm detail
        
        for airing in ChanSchedule :
    
            aPidstr = airing['ProgramId']   # Program ID is the key to it all
            
            # Attempt to find the matching programID in Prog list we just requested
            pd = ProgramsLookupDict.get(aPidstr, None)
            
            if not pd :  # if not in the 'new' list, then...
                
                logging.debug("LS:BLS: ProgramId (%s) not found! Checking Cache", aPidstr)
                
                pd = ProgramsCache.get( aPidstr, None)
                
                if pd :
                    
                    pd['LastAccessed'] = datetime.utcnow()  # update timestamp 
                    
                else : # should have been found in list or Cache!
                    
                    logging.warning("LS:BLS: ProgramId (%s) not found!", aPidstr)
                    
                    # if PiD not found, create an error message 'Program'
                    
                    SDPD_Error["programID"] = pidstr
                    SDPD_Error["titles"][0]["title120"] = "Missing Program [" + pidstr + "]"
                    SDPD_Error["descriptions"]['description100'][0]["description"] = \
                                                "Missing Program ID [" + pidstr + "]"
                    
                    # Use these temp Program details, code below will copy it 
                    pd = SDPD_Error
                
            #ToDo what if key does not exist?  it is a mandatory key
                
            try :
                title = airing["Title"] = copy.copy( pd['titles'][0]['title120'] )
            except :
                logging.warning("LS:BLS: pd['titles'] error for ProgramId(%s)!", aPidstr)
                
                print('[AIRINGp] ',airing)
                print('[PDp]:',pd)
               # print('[MDp]:',md)
            
            # Make an enhanced program description
            # SH type programID's dont have descriptions
            # ToDo there could be an option to select 100 or 1000 
            # DVR+ can display up to 512 chars.  COPY_MAX_CHARS=512
            
            # How do we pick which description to use? d100 or d1000?
            # We have 3 things to consider
            # 1) do they want the Episode Title, S#E#, AirDate:?
            # 2) How long of a description can we add?
            # 3) do they want the cast list?
            # That seems to be the priority....
            # 1) Give them the Episode info if they asked it, then
            # 2) give them the longest descripton that will fit, staying under 512 chars
            # 3) then fit in as many full cast names as we can in 512 chars

            summary = ""
            
            # Default the description with '(New)' for new episodes
            newEpisodeFlag = (airing["AiringType"] == "New")
            
            if newEpisodeFlag :
                summary = "(New) " 

            #1 Build the Episode info string, might use it below
            oAirDate = pd.get("originalAirDate","")
            
            #if oAirDate == "1999-09-05" :  # remove strange date from certain news broadcasts
            #    oAirDate = ""
                
            if oAirDate : oAirDate = "Aired: " + oAirDate + ". "
            
            # not all listings have episode titles
            EpisodeTitle = pd.get("episodeTitle150","")
            if EpisodeTitle : EpisodeTitle = '"' + EpisodeTitle + '."  '
            
            # ET does nothing for DVR+ but there is a place for it so I will put it there 
            airing["EpisodeTitle"] = copy.copy(EpisodeTitle);
                            
            # build a Season Episode S#E# string if we can
            seasonStr = ""
            md = pd.get("metadata")
            
            if (md) : #if metadata exists
                if ( md[0].get("Gracenote") ) :  # only want gracenote meta
                    # sometimes there is 0,0 data which seems to be bogus b/c
                    # the same 0,0 occurs for multiple episodes
                    # epi title is usually missing and airdate seems like a default
                    
                    sn = md[0]["Gracenote"].get('season',"?")
                    en = md[0]["Gracenote"].get('episode',"?")  # optional

                    # Make a compact string so as not to use up too much description
                    seasonStr = "S" + str(sn) + ", E" + str(en) + ". "
                    
                    # if its an invalid season #, clear the Epi,SN,EN
                    if sn == 0 :
                        EpisodeTitle = oAirDate = seasonStr = ""

            # Now the New flag is set and Summary is initialized, add episode info
            if PiGScfg['PrependEpisode'] :
                summary += EpisodeTitle + seasonStr + oAirDate
                
            #2, handle description, also called 'copy'
                
            desc = "No summary available(PiGS)."  # default value, if we find no description
                                                  # this will give DVR+ something to display
            de = pd.get('descriptions')           # reference to the program description stuff
            
            # there can be a 1000 char descr, a 100 char descr or both, decide which to use

            d1000s = d100s = ""  # make sure strings are clear from prior iteration
            if de : # if there is a description section, in returned data
                d1000  = de.get('description1000')  # chk avail for long and short versions
                d100   = de.get('description100')
                
                # Then pull the strings for short and long if possible, or "" if not
                d1000s = d1000[0]['description'] if d1000 else ""
                d100s  =  d100[0]['description'] if d100  else ""
                    
            # Now we can decide which desc to use based on available space remaining in 512 limit
            #charsAvail = COPY_MAX_CHARS - len(summary)- 5 # 5 for end count below
            charsAvail = COPY_MAX_CHARS - len(summary)

            # now append the main description
            if d1000s and len(d1000s) < charsAvail :  # if the long desc will fit, use it
                summary += d1000s
            elif len(d100s):  # else if there is a d100s, use that (even if its too long)
                #logging.debug("Using d100s CharsAvail(%d), d1000s(%d), d100s(%d)", charsAvail, len(d1000s), len(d100s) )
                summary += d100s  
            else :           #but... if there is no d100s, then use the long one anyway
                logging.debug("        Default d1000 CharsAvail(%d), d1000s(%d), d100s(%d)", charsAvail, len(d1000s), len(d100s) )
                summary += d1000s  # longer than 512 will just get truncated by DVR+

            # Truncate to 511, append "..."
            if len(summary) > 511 :
                summary = summary[:508] + "..."
                 
            # now fill in any remaining space with cast info, if they asked for it
            
            # If they want the castlist, try and build one
            caststr = ""
            if PiGScfg['AppendCast'] :    # does user want one?
                castlist = pd.get('cast') # is there a castlist in the SD Data?
                if castlist :
                    caststr    = " Cast: "  # Label
                    
                    # adjust chars avail  less current len, less cast header
                    #charsAvail = COPY_MAX_CHARS - len(summary) - len(caststr) - 5 # 5 for end count below
                    charsAvail = COPY_MAX_CHARS - len(summary) - len(caststr)
                    
                    clen = len(caststr)          # len of just the label for testing below
                    
                    # fit as many full names as we can
                    for person in castlist :                        
                        actorname = person['name'] +", "  # get a name, append a ','
                        alen = len(actorname)
                        if alen > charsAvail :   # check if name will fit, 
                            break                # quit when out of space
                        caststr += actorname     # otherwise add name and continue
                        charsAvail -= alen       # account for name len
                    
                    if clen == len(caststr) :    # if we could not add any names, remove the label
                        caststr = ""
                    else :
                        #replace the trailing comma with a period, slice off last two chars, add a period 
                        caststr = caststr[:-2] + "."
            
            # Append the cast list if one was built
            summary += caststr

            # message if we created a string longer than 512 chars
            
            slen = len(summary)
            if slen > COPY_MAX_CHARS :
                logging.warning("LS:BLS: Desc=%d %s %s", slen, airing["AiringTime"], airing["Title"])
                
            #------------------------------------------------------------------------------------
            # Set New Flag for users list of Shows in list MarkAsNew{}
            # e.g. Fix "America's Test Kitchen"  "S21" not having the new flag
            
            for markitem in PiGScfg['MarkAsNew'] :
                
                titleMatch = (markitem[0].lower() in title.lower())
                descrMatch = (markitem[1].lower() in summary.lower())
                
                # if Title matches and Unmark flag is True, Set show as Not New
                # Example is title "NFL Football", description = "Packers vs Stealers"
                
                if titleMatch and newEpisodeFlag and newEpisodeFlag and markitem[2] :
                    airing["AiringType"] = "Unknown"         # Unmark 
                    summary = "Was"+summary           # add Was(New)' to indicate PiGS made it UNnew
                    
                # if "America's Test" in title and sn==21 and airing["AiringType"] != "New" :
                # Only test shows not marked New by SD
                    
                if (titleMatch and descrMatch ) and not newEpisodeFlag : 
                    airing["AiringType"] = "New"
                    summary = "(New+)" + summary          # add '(New+)' to indicate PiGS made it new
                     
                    logging.warning("LS:BLS: Marking (New+) %s,%s%s", title, EpisodeTitle, seasonStr)
            # Done with Mark as new
            
            # summary was created for debugging so I could view the string easily
            airing["Copy"] = copy.copy( summary )
            
            
            
        # Part3 ========================================================================
        # Appends the ChanSchedule airings to the Master list, then sorts by date
        # https://docs.python.org/3/howto/sorting.html
        # https://www.geeksforgeeks.org/ways-sort-list-dictionaries-values-python-using-itemgetter/?ref=rp

        # add Channel Schedule to master Program Schedule
        NewSchedule.extend(ChanSchedule)
 
        # sort the list by ascending airing list, compute sort time.  it is quick
        t1=datetime.utcnow()
        NewSchedule.sort( key=itemgetter('AiringTime') )
        t2=datetime.utcnow()
        
        tdiff=t2-t1
        ts=str.format('{:1}.{:06}', tdiff.seconds,tdiff.microseconds)
        logging.debug("LS:BLS: %d airings addded. %d total items. Sort time: %ss", len(ChanSchedule),len(NewSchedule), ts)
        
        # Done with 1 station, report the completion, update progress        
        
        #logging.info("LS:BLS: DupListLen=%s, RepeatsListLen=%s", len(DupIdList), len(RepeatsList) )
        #logging.info("  Added %s programs for channel %s, RfChan=%s (StationID=%d) CacheLen(%d)", \
        logging.info("LS:BLS: Added %s programs for channel %s, RfChan=%s (StationID=%d) CacheLen(%d)", \
              len(ChanSchedule), aStation['Channel'], aStation['ChannelLocation'], iSourceId,len(ProgramsCache))
        
        if len(ChanSchedule) == 0 :
            logging.info("LS:BLS: 0 progs: code = %s", code)
            
        # compute build percent
        pct = aStation["Order"] / TotalChannels * 100
        LSBuildStat.SetBuildTime("Rebuilding... " + str(int(pct)) + "%")
        
        # *** End of Stage 3 ***
        
    # Part 4 - Build complete, Finish up ====================================
        
    # update schedule build clock time 
    LastBuiltTime = datetime.utcnow()  # time of this request
    lt,lz = LocalTimeStr(LastBuiltTime, showSeconds=True)

    # incremental time for this build
    buildMinSecDif = datetime.utcnow() - BuildStartTime
    buildMinSecStr = str.format('{:02}:{:02}s', buildMinSecDif.seconds//60, buildMinSecDif.seconds%60)

    # if Build process was aborted due to SD error or Thread abort signal
    if BuildAbort :
        #ProgramSchedule.clear()
        bldmsg = "Build Aborted... Schedule not updated!"
        LSBuildStat.SetLastBuildTime(bldmsg)
        LSBuildStat.SetBuildTime("Build Aborted at " + lt)
        logging.info("LS:BLS: Build Aborted at " + lt)
        
    elif len(NewSchedule) > 0:  # SD build was successful

        ProgramSchedule.clear()          # delete prior schedule, update with new.
        ProgramSchedule = NewSchedule    # assign to new list
        #gc.collect()                    # reclaim old list memory (doesnt do much)

        # parser.parse() returns date-aware times
        # parser.parse('2022-05-11Z00:00Z') >>> datetime.datetime(2022, 5, 11, 0, 0, tzinfo=tzutc())
        # datetime.utcnow() returns naive times datetime.datetime(2022, 5, 11, 15, 8, 24, 711493)
        # datetime.now(timezone.utc) to return aware now() times
        #     datetime.datetime(2022, 5, 11, 15, 29, 8, 995368, tzinfo=datetime.timezone.utc)
        
        # d1 Total days in schedule
        t1Diff = parser.parse(ProgramSchedule[-1]['AiringTime']) - parser.parse(ProgramSchedule[0]['AiringTime'])
        d1 = t1Diff.days + 1

        # New days from Today to end of schedule
        t2Diff = parser.parse(ProgramSchedule[-1]['AiringTime']) - datetime.now(tz=timezone.utc) 
        d2 = t2Diff.days + 1
        
        logging.info("\nLS:BLS: Schedule Ends: %s  (%d / +%d days)", ProgramSchedule[-1]['AiringTime'], d2-d1,d2)        
        logging.info("LS:BLS: Calculating Program Schedule memory usage...")
        ListMBs = total_size(ProgramSchedule)/1024/1024
        logging.info("LS:BLS: ProgramSchedule[%d items] %4.1f MB", len(ProgramSchedule), ListMBs )
        
        #Total size of programs Cache
        
        logging.info("LS:BLS: Checking Programs Cache for Outdated items...")
        OutdatedSet = set()
        for pidstr in ProgramsCache :
            tDiff = LastBuiltTime - ProgramsCache[pidstr]['LastAccessed']
            if tDiff.days > 3 : OutdatedSet.add(pidstr)
                
        if len(OutdatedSet) :
            logging.info("  Purging %d Outdated Programs from Cache...", len(OutdatedSet))
            for pidstr in OutdatedSet : ProgramsCache.pop(pidstr, 0)
        
        PgmsMBs = total_size(ProgramsCache)/1024/1024
        logging.info("LS:BLS: ProgramsCache[%d items] %4.1f MB", len(ProgramsCache), PgmsMBs )

        LSBuildStat.SetLastBuildTime(lt)
        LSBuildStat.SetPrograms(str(len(ProgramSchedule)) + str.format(' ({:.1f} MB)', ListMBs))
        LSBuildStat.SetSchStart(ProgramSchedule[ 0]['AiringTime'])
        LSBuildStat.SetSchEnd(ProgramSchedule[-1]['AiringTime'])

        LSBuildStat.SetBuildTime(buildMinSecStr)

        # Save the Program Schedule and Programs Cache to Disk

        if PiGScfg['SaveSchedule'] == 2 :
            logging.info('LS:BLS: Saving Program Schedule to Disk...')
            utSaveObjectToDisk(ProgramSchedule, SCHED_FILE)

        if PiGScfg['SaveSchedule'] >= 1 : # 1 
            logging.info('LS:BLS: Saving Program Cache to Disk...')
            utSaveObjectToDisk(ProgramsCache, CACHE_FILE)

        
        alertmsg = "Build Complete %s!\r\n%s Total programs." % ( lt, len(ProgramSchedule) )
        SendEmailAlert(alertmsg, UpdateMsg=updateTxt)
        
        LSBuildStat.SetLastBuildTime(lt)
        logging.info("LS:BLS: Build Complete at %s", lt)
        
    else :  # zero length schedule
        LSBuildStat.SetLastBuildTime("No programs found.  Schedule not updated!")
        LSBuildStat.SetBuildTime(buildMinSecStr)
        
        logging.warning("LS:BLS: Build Complete! %s total programs. Something Misconfigured?", len(NewSchedule) )
        alertmsg = "Schedule Build produced 0 programs!  %s" % ('Something misconfigured? Check end of pigslogs.txt')
        SendEmailAlert(alertmsg)


    # Clear or Save the Program Cache on disk?
    # Should it be cleared to save memory?
    #ProgramsCache.clear()
    
    # compute build percent
    pct = 100
    #BuildInfo['Schedule Build Info']['Progress'] = str(int(pct)) + "%"

    logging.info("LS:BLS: Total Build time %s\n", datetime.utcnow()-BuildStartTime)

    return BuildResult




# Message type 2 - Return the a schedule when DVR+ request it  
# The http GET message looks like this:
# GET /TVlistings/v9/listings/linearschedule/20499/info?
#        apikey=v4edzwwtca4e847n72j8kesw&
#        locale=en-US&
#        duration=240&
#        inprogress=true&
#        startdate=2020-05-02T16:00:00Z&
#        format=json HTTP/1.1
#
# GetLinearSchedule() returns a DVR+ formatted json message of all listings that 
# *Start* 'in' the DVR+ requested time window :
#    (startdate) through (startdate+duration) 
# If the flag inprogress is set to true, then GLS will also return listings that
#  Start *before* the request window, but end *in* the request window
#
# Maintain a list of the DVR IP's that are making requests.  Save their ListIndex & StartTimes. 
# DVR IP will be the key, the value is a list with the List Index, and the tFirstRequest

DictOfDVRs = {}
# will look like this as various dvr's make requests to PiGS
# { 'DVR1-IP' : [ListIndex1, StartTime1, LastUpdateStr1, RequestCounter1],
#   'DVR2-IP' : [ListIndex2, StartTime2, LastUpdateStr2, RequestCounter2],
#   'DVR3-IP' : [ListIndex3, StartTime3, LastUpdateStr3, RequestCounter3],
#   ....
# }

# Index offsets to List Items - 
# SI StateIndex enum
from enum import IntEnum
class SI(IntEnum) :
    STATE_LI = 0    # List Index
    STATE_ST = 1    # Start Time
    STATE_LU = 2    # Last Update
    STATE_RC = 3    # Request Counter

ZEROLIST = [0]*len(SI)   # list of 4 zero's

# Get & Set so we can save DVR states in pigs.json
# need to convert datetime tuples to timestamps for storage in json

def LS_GetDictOfDVRs() :
    global DictOfDVRs
    dj = {}
    for k in DictOfDVRs :
        dj[k] = {}
        dj[k]['STATE_LI'] = DictOfDVRs[k][SI.STATE_LI]
        dj[k]['STATE_RC'] = DictOfDVRs[k][SI.STATE_RC]
        dj[k]['STATE_ST'] = DictOfDVRs[k][SI.STATE_ST].timestamp()
        dj[k]['STATE_LU'] = DictOfDVRs[k][SI.STATE_LU].timestamp()
    return dj

# convert json timestamps back to datetime tuples
def LS_SetDictOfDVRs(stateDict) :
    global DictOfDVRs
    DictOfDVRs = {}
    for key in stateDict :
        logging.info("  LS: DVR %s ",key)
        DictOfDVRs[key] = copy.copy(ZEROLIST)   # create list with IP key
        # If PiGS crashed mid-update, LI & RC could be out-of-date. Don't load LI (reset to 0)
        #DictOfDVRs[key][SI.STATE_LI] = copy.copy(stateDict[key]['STATE_LI'])
        DictOfDVRs[key][SI.STATE_RC] = copy.copy(stateDict[key]['STATE_RC'])
        DictOfDVRs[key][SI.STATE_ST] = copy.copy(datetime.fromtimestamp( stateDict[key]['STATE_ST'] ))
        DictOfDVRs[key][SI.STATE_LU] = copy.copy(datetime.fromtimestamp( stateDict[key]['STATE_LU'] ))

# For Status display in a browser
def LsGetDVRStatus(maxdays) :
    global DictOfDVRs
    
    # cant json datatime objects so convert list item [1] to str
    # and make a more pretty json object
    dj = {}
    for k in DictOfDVRs :
        
        # don't display DVRs that are not being updated
        td = datetime.utcnow() - DictOfDVRs[k][SI.STATE_LU]
        if maxdays and td.days > maxdays :
            continue
        
        dj[k] = {}
        dj[k]['ListIndex'] = DictOfDVRs[k][SI.STATE_LI]
        dj[k]['RequstNum'] = DictOfDVRs[k][SI.STATE_RC]
        
        ts,tz = LocalTimeStr( DictOfDVRs[k][SI.STATE_ST],True )
        dj[k]['FirstRqst'] = ts
        ts,tz = LocalTimeStr( DictOfDVRs[k][SI.STATE_LU],True )
        dj[k]['Last_Rqst'] = ts
        
        tdiff = DictOfDVRs[k][SI.STATE_LU]-DictOfDVRs[k][SI.STATE_ST]
        ts=str.format('{:02}:{:02}s', tdiff.seconds//60, tdiff.seconds%60)
        dj[k]['__Elapsed'] = ts
        
    return dj

def GetLinearSchedule(PiGScfg, qs, serviceID, DVR_IP, tzlist):

    global DictOfDVRs

    theTime=datetime.utcnow()  # time of this request

    # parse the GET params
    dur     = int(qs["duration"][0])      # duration in minutes
    bInProg = qs["inprogress"][0]=='true'
    rsts    = qs['startdate'][0]          # request start time string
    
    # Figure out DVR+ request window start and end times in Zulu
    rst     = parser.parse( rsts )
    ret     = rst + timedelta(minutes=dur)

    # Fixing Frank70's multiple dvr's at the same time bug.
    # Cache a list of DVRs making schedule requests to this server
    # keep track of their list index and start time 
    if DVR_IP not in DictOfDVRs :
        DictOfDVRs[DVR_IP] = copy.copy(ZEROLIST)    # if key does not exist, create it and initialize
        
    # Get the ListIndex for this DVR, ListIndex will be cleared below if this is the first request
    ListIndex = DictOfDVRs[DVR_IP][SI.STATE_LI]
    
    logging.info("LS:GLS DVR %s State[%d,%s]", DVR_IP, DictOfDVRs[DVR_IP][SI.STATE_LI], DictOfDVRs[DVR_IP][SI.STATE_ST] )

    # Update the Response Header
    # Most of the stuff in the header is ignored by DVR+
    #%f is 6-digit microsecs.  need mS?  setting to '000' 
    lsr["LinearScheduleResult"]["TimeStamp"] = theTime.strftime("%Y-%m-%dT%H:%M:%S.000Z")
    lsr["LinearScheduleResult"]["Schedule"]["ServiceId"]= serviceID  # always 99999

    lsr["LinearScheduleResult"]["Schedule"]["TimeZones"] = copy.deepcopy(tzlist)
    
    lsr["LinearScheduleResult"]["Schedule"]["StartDate"] = qs['startdate'][0]
    lsr["LinearScheduleResult"]["Schedule"]["Duration"]  = dur
   
    #https://www.w3schools.com/python/python_dictionaries.asp
    # get reference to the airings list in result dict object
    
    alist = lsr["LinearScheduleResult"]["Schedule"]["Airings"]
    alist.clear()  # python 3.0, clear the list
    
    # start at the beginning of the list (ListIndex=0) on the first DVR+ request (bInProg=true)
    #  on subsequent requests leave ListIndex alone to start where we left off for speed.
    # On the first request there could be shows on the schedule grid that are
    #  inProgress.  like if we are viewing at 730pm, a show from 7pm could be ongoing.
    # DVR+ needs to know about this and we can use it to print start time, reset search index, etc
    if bInProg :
        ListIndex = 0
        
        DictOfDVRs[DVR_IP][SI.STATE_ST] = datetime.utcnow()          #tFirstRequest = datetime.utcnow()
        DictOfDVRs[DVR_IP][SI.STATE_RC] = 0
    
        #logging.info("LS:GLS First Request at UTC: %s", tFirstRequest.strftime("%Y-%m-%dT%H:%M:%SZ"))
        tss, tzz = LocalTimeStr( DictOfDVRs[DVR_IP][SI.STATE_ST], showSeconds=True)
        logging.info("LS:GLS First Request at: %s", tss )

    # error condition - if PiGS was restarted somehow during a DVRs schedule request,
    # The dvr will continue on its Merry way, but the Start Time would be unitialized at 0
    
    if DictOfDVRs[DVR_IP][SI.STATE_ST] == 0 :
        DictOfDVRs[DVR_IP][SI.STATE_ST] = datetime.utcnow()          #tFirstRequest = datetime.utcnow()
        logging.warning("LS:GLS Unexpected unitialized Start Time.  Setting to UTC now!")

    # Get the start time for this particular DVR, time could get reset just above
    tFirstRequest = DictOfDVRs[DVR_IP][SI.STATE_ST]

    # Total items in the AiringList
    listlen = len(ProgramSchedule)

    # if the list is empty, its wierd, but not the end of the world.
    # we will just return an empty list and DVR+ will be happy
    if not listlen  :
        logging.warning("LS:GLS: ProgramSchedule has not been built!!")

    # As a DVR or DVR's make schedule requests, we just leave the ListIndex at its last position
    # because the DVR is going to make the next request right there anyway.  minimizes seaching.
    
    # Dont have to test for accessing beyond the end of the list because the search 
    # method below backs up at least 1 listing before accessing ProgramSchedule[]
    
    if (listlen>0) : # and (ListIndex<listlen) : 
        
        # Test for a request time before the listIndex.
        # In case a DVR does a retry we may need to back up the search index pointer ListIndex.
        # search backwards until I find a listing before the rst, or I get to the beginning of the list.
        # if this 'Backup' linear search method got too slow, some type of binary search could be implemented

        lis = ListIndex             # remember where we started for logging info
        
        while (ListIndex > 0 ) :    # search, but not before the start of the list.

            #  back up 1 listing and check listing time.
            ListIndex -= 1    # pre-decrement to ensure not accessing past the end
            
            #get program airing start time STRING like "2020-03-08T08:00:00Z"
            pss = ProgramSchedule[ListIndex]['AiringTime']
            
            # convert to a TIME - prog start time pst
            pst = parser.parse( pss )
            
            # if the Program Start time pst is before the requested start time rst, were are done
            if (pst < rst) : break      # out of the loop
            
        
        # backup of 1 is expected
        if (lis - ListIndex) > 1 :   
            logging.info('LS:GLS:**** Request before current ListIndex(%d). Rollback to ListIndex(%d).', lis, ListIndex)
 
    logging.info('LS:GLS:Searching for %s. InProgessFlag=%d...', rsts,bInProg)

    ListingCnt = 0   # Count how many we send in each request, just for Info
    
    while (ListIndex < listlen) :   # dont search past end of list, could be 0 items

        #get program airing start time string like "2020-03-08T08:00:00Z"
        pss = ProgramSchedule[ListIndex]['AiringTime']
        #prog start time pst
        pst = parser.parse( pss )
        #prog duration int
        pdur= ProgramSchedule[ListIndex]['Duration']
        #prog end time pet
        pet = pst + timedelta(minutes=pdur)

        # if the Program Start time pst is beyond requested end time ret, were are done
        if (pst >= ret) : break      # out of the loop 

        # test if the start of this program is in the req window (pst>=rst)
        # or
        # if asking for an inprogress shows, it 'starts before the window' and 'ends in the window'
        bInWindow = (pst>=rst) or \
                     bInProg and (pst<rst) and (pet > rst)
  
        # this compiles list of references to the Master list, so dont delete Master list!
        
        if (bInWindow) :
            
            ListingCnt += 1
            
            alist.append( ProgramSchedule[ListIndex] )
            
            # could deep copy the data, but we should not need to 
            # create a new Channel structure, recursive dict copy
            #alist.append( copy.deepcopy(ProgramSchedule[ListIndex]) )
            
            #debug print a list of New shows
#            bnew = ProgramSchedule[ListIndex]['AiringType'] == 'New'
#            if bnew :
#                a = ProgramSchedule[ListIndex]
#                airtimez = parser.parse( a['AiringTime'] )
#                airtimecst = airtimez + timedelta(minutes=-300)            
#                print("New", a['Channel'], airtimecst, a['Title'])
                
        ListIndex += 1;  # move on to the next show in ProgramSchedule
        
    #end of while (ListIndex < listlen) :

    t2=datetime.utcnow()
    tdiff = t2-tFirstRequest
    #ts=str.format('{:02}:{:02}.{:06}s', tdiff.seconds//60, tdiff.seconds%60,tdiff.microseconds)
    ts=str.format('{:02}:{:02}s', tdiff.seconds//60, tdiff.seconds%60)
 
    # Done serving this request, Save the ListIndex for this DVR for it's next call
    # Update this DVR's State

    DictOfDVRs[DVR_IP][SI.STATE_RC] += 1         # count number of requests
    DictOfDVRs[DVR_IP][SI.STATE_LI] = ListIndex  # 
    DictOfDVRs[DVR_IP][SI.STATE_LU] = t2         # Last Update Time in Local Time
    
    rqCnt = DictOfDVRs[DVR_IP][SI.STATE_RC]

    logging.info('LS:GLS:Request complete! %d Listings. LI=%d RequestCount=%d, Elapsed= %s\n', ListingCnt, ListIndex, rqCnt, ts)


    # If we have no more data to send to the DVR we have to send it an empty list
    # on the final request the DVR+ gets/needs an empty list airings=[]
    # but with SD we get more than 2 weeks of data, so DVR+ just quits asking after 14days (or 90 requests
    # I see this if the DaysLimit is set to something less than 16 in the config file
    # so this message doesnt always get printed - it depends on the data
    # It's been empirically determined that DVR asks for 90 4 hour blocks (15 days) and then stops asking
    
    if len(alist) == 0 :
        #tdiff = t2-tFirstRequest
        #ts=str.format('{:02}:{:02}.{:06}s', tdiff.seconds//60, tdiff.seconds%60,tdiff.microseconds)
        #logging.info('LS:GLS:That appears to be the final request for (%s)! Total time =%s\n', DVR_IP, ts)
        logging.info('LS:GLS:PiGS is out of data. Sending empty response to DVR (%s)!\n', DVR_IP)
 
    if rqCnt == 90:
        logging.info('LS:GLS:That appears to be the final request for DVR (%s)! Elapsed Time %s\n', DVR_IP, ts)
        
    #convert Python dict to json and return
    js = json.dumps(lsr)    
    
    return js


# this doesnt work, or is not tested
def GetChunkedLinearSchedule(PiGScfg, qs, serviceID, DVR_IP, socket_wfile):

    global DictOfDVRs

    theTime=datetime.utcnow()  # time of this request

    # parse the GET params
    dur     = int(qs["duration"][0])      # duration in minutes
    bInProg = qs["inprogress"][0]=='true'
    rsts    = qs['startdate'][0]          # request start time string
    
    # Figure out DVR+ request window start and end times in Zulu
    rst     = parser.parse( rsts )
#    rst     = rst + timedelta(minutes=offset)  # chunk
    ret     = rst + timedelta(minutes=dur)

    # Fixing Frank70's multiple dvr's at the same time bug.
    # Cache a list of DVRs making schedule requests to this server
    # keep track of their list index and start time 
    if DVR_IP not in DictOfDVRs :
        DictOfDVRs[DVR_IP] = copy.copy(ZEROLIST)    # if key does not exist, create it and initialize
        
    # Get the ListIndex for this DVR, ListIndex will be cleared below if this is the first request
    ListIndex = DictOfDVRs[DVR_IP][SI.STATE_LI]
    
    logging.info("LS:GLS DVR %s State[%d,%s]", DVR_IP, DictOfDVRs[DVR_IP][SI.STATE_LI], DictOfDVRs[DVR_IP][SI.STATE_ST] )

    # Update the Response Header
    # Most of the stuff in the header is ignored by DVR+
    #%f is 6-digit microsecs.  need mS?  setting to '000' 
    lsr["LinearScheduleResult"]["TimeStamp"] = theTime.strftime("%Y-%m-%dT%H:%M:%S.000Z")
    lsr["LinearScheduleResult"]["Schedule"]["ServiceId"]= serviceID  # always 99999

    lsr["LinearScheduleResult"]["Schedule"]["TimeZones"] = copy.deepcopy( PiGScfg['TimeZones'])
    
    lsr["LinearScheduleResult"]["Schedule"]["StartDate"] = qs['startdate'][0]
    lsr["LinearScheduleResult"]["Schedule"]["Duration"]  = dur
   
    #https://www.w3schools.com/python/python_dictionaries.asp
    # get reference to the airings list in result dict object
    
    alist = lsr["LinearScheduleResult"]["Schedule"]["Airings"]
    alist.clear()  # python 3.0, clear the list
    
    # start at the beginning of the list (ListIndex=0) on the first DVR+ request (bInProg=true)
    #  on subsequent requests leave ListIndex alone to start where we left off for speed.
    # On the first request there could be shows on the schedule grid that are
    #  inProgress.  like if we are viewing at 730pm, a show from 7pm could be ongoing.
    # DVR+ needs to know about this and we can use it to print start time, reset search index, etc
    if bInProg :
        ListIndex = 0
        
        DictOfDVRs[DVR_IP][SI.STATE_ST] = datetime.utcnow()          #tFirstRequest = datetime.utcnow()
        DictOfDVRs[DVR_IP][SI.STATE_RC] = 0
    
        #logging.info("LS:GLS First Request at UTC: %s", tFirstRequest.strftime("%Y-%m-%dT%H:%M:%SZ"))
        tss, tzz = LocalTimeStr( DictOfDVRs[DVR_IP][SI.STATE_ST], showSeconds=True)
        logging.info("LS:GLS First Request at: %s", tss )

    # error condition - if PiGS was restarted somehow during a DVRs schedule request,
    # The dvr will continue on its Merry way, but the Start Time would be unitialized at 0
    
    if DictOfDVRs[DVR_IP][SI.STATE_ST] == 0 :
        DictOfDVRs[DVR_IP][SI.STATE_ST] = datetime.utcnow()          #tFirstRequest = datetime.utcnow()
        logging.warning("LS:GLS Unexpected unitialized Start Time.  Setting to UTC now!")

    # Get the start time for this particular DVR, time could get reset just above
    tFirstRequest = DictOfDVRs[DVR_IP][SI.STATE_ST]

    # Total items in the AiringList
    listlen = len(ProgramSchedule)

    # if the list is empty, its wierd, but not the end of the world.
    # we will just return an empty list and DVR+ will be happy
    if not listlen  :
        logging.warning("LS:GLS: ProgramSchedule has not been built!!")

    # As a DVR or DVR's make schedule requests, we just leave the ListIndex at its last position
    # because the DVR is going to make the next request right there anyway.  minimizes seaching.
    
    # Dont have to test for accessing beyond the end of the list because the search 
    # method below backs up at least 1 listing before accessing ProgramSchedule[]
    
    if (listlen>0) : # and (ListIndex<listlen) : 
        
        # Test for a request time before the listIndex.
        # In case a DVR does a retry we may need to back up the search index pointer ListIndex.
        # search backwards until I find a listing before the rst, or I get to the beginning of the list.
        # if this 'Backup' linear search method got too slow, some type of binary search could be implemented

        lis = ListIndex             # remember where we started for logging info
        
        while (ListIndex > 0 ) :    # search, but not before the start of the list.

            #  back up 1 listing and check listing time.
            ListIndex -= 1    # pre-decrement to ensure not accessing past the end
            
            #get program airing start time STRING like "2020-03-08T08:00:00Z"
            pss = ProgramSchedule[ListIndex]['AiringTime']
            
            # convert to a TIME - prog start time pst
            pst = parser.parse( pss )
            
            # if the Program Start time pst is before the requested start time rst, were are done
            if (pst < rst) : break      # out of the loop
            
        
        # backup of 1 is expected
        if (lis - ListIndex) > 1 :   
            logging.info('LS:GLS:Request before current ListIndex(%d). Rollback to ListIndex(%d).', lis, ListIndex)
 
#    logging.info('LS:GLS:Searching for %s. InProgessFlag=%d...', rsts,bInProg)

    ListingCnt = 0   # Count how many we send in each request, just for Info
    
    while (ListIndex < listlen) :   # dont search past end of list, could be 0 items

        #get program airing start time string like "2020-03-08T08:00:00Z"
        pss = ProgramSchedule[ListIndex]['AiringTime']
        #prog start time pst
        pst = parser.parse( pss )
        #prog duration int
        pdur= ProgramSchedule[ListIndex]['Duration']
        #prog end time pet
        pet = pst + timedelta(minutes=pdur)

        # if the Program Start time pst is beyond requested end time ret, were are done
        if (pst >= ret) : break      # out of the loop 

        # test if the start of this program is in the req window (pst>=rst)
        #   or
        # if asking for an inprogress shows, it 'starts before the window' and 'ends in the window'
        bInWindow = (pst>=rst) or \
                     bInProg and (pst<rst) and (pet > rst)
  
        # this compiles list of references to the Master list, so dont delete Master list!
        
        if (bInWindow) :
            
            ListingCnt += 1
            
            alist.append( ProgramSchedule[ListIndex] )
            
            # could deep copy the data, but we should not need to 
            # create a new Channel structure, recursive dict copy
            #alist.append( copy.deepcopy(ProgramSchedule[ListIndex]) )
            
        ListIndex += 1;  # move on to the next show in ProgramSchedule
        
    #end of while (ListIndex < listlen) :

    t2=datetime.utcnow()
    tdiff = t2-tFirstRequest
    #ts=str.format('{:02}:{:02}.{:06}s', tdiff.seconds//60, tdiff.seconds%60,tdiff.microseconds)
    ts=str.format('{:02}:{:02}s', tdiff.seconds//60, tdiff.seconds%60)
 
    # Done serving this request, Save the ListIndex for this DVR for it's next call
    # Update this DVR's State

    DictOfDVRs[DVR_IP][SI.STATE_RC] += 1         # count number of requests
    DictOfDVRs[DVR_IP][SI.STATE_LI] = ListIndex  # 
    DictOfDVRs[DVR_IP][SI.STATE_LU] = t2         # Last Update Time in Local Time
    
    rqCnt = DictOfDVRs[DVR_IP][SI.STATE_RC]

#    logging.info('LS:GLS:Request complete! %d Listings. LI=%d RequestCount=%d, Elapsed= %s\n', ListingCnt, ListIndex, rqCnt, ts)


    # If we have no more data to send to the DVR we have to send it an empty list
    # on the final request the DVR+ gets/needs an empty list airings=[]
    # but with SD we get more than 2 weeks of data, so DVR+ just quits asking after 14days (or 90 requests
    # I see this if the DaysLimit is set to something less than 16 in the config file
    # so this message doesnt always get printed - it depends on the data
    # It's been empirically determined that the DVR asks for 90 4 hour blocks and then stops asking
    
    if len(alist) == 0 :
        logging.info('LS:GLS:PiGS is out of data. Sending empty response to DVR (%s)!\n', DVR_IP)
 
    if rqCnt == 90:
        logging.info('LS:GLS:That appears to be the final request for DVR (%s)! Elapsed Time %s\n', DVR_IP, ts)
        
    #convert Python dict to json and return
    if FirstChunk :
        js = json.dumps(lsr)
    # chunked - just dump alist)
    else :
        js = json.dumps(alist)
    
    return js
