# pigs.py
#
# PiGS Channel Master CM-7500 DVR+ Guide Server Replacement
#
#---------------------------------------------------------------------------------------
# Copyright 2021,2022 Timothy Goihl
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# See GNU-GPL-V3.txt License File
#
#---------------------------------------------------------------------------------------
# Files in the PiGS Guide Server project
#  pigs.py The main application / server
#  ls.py   PiGS LinearSchedule & Build functions
#  sv.py   PiGS Service Details functions
#  cm.py   Channel Master / Rovi APIs (access CM/Rovi servers) to get image data
#  sd.py   Schedules Direct.org API calls
#  ut.py   Utility functions that may be used by more than one module
#  se.py   Searching schedule from url
#  em.py   V1.24 added email sending
#  find.htm    Simple search UI for se.py
#  index.htm   Configuration UI for pigs static IP e.g. //192.168.1.200
#  pigsdns.py  DNS server for DVR+ added in v1.02a
#
#  5 image files favicon.ico, pinkpig.png, 92x36box.png, misspiggy40x30.png, pigoutline92x36.png
#  pigs.service   a file that can be used to start pigs at boot (ie Auto-Run)
#  pigs.desktop   a file that can be placed on the Pi desktop to manually start PiGS
#  GNU-GPL-V3.txt License File
#  ReleaseNotes.txt - verbose history of version changes
#  version.txt   a file place on the pigs download server so Pigs can check for new updates
#   invoke nano with -L to prevent it from adding newline <LF> on save
#
#  scan.py Test functions to CM/Rovi servers, also a Simple file server.
#     scan.py needs cm.py to run, but that is it.
#     see instructions at the top of scan.py
#
#---------------------------------------------------------------------------------------
# To run pigs.py from a Terminal window, type:
#    sudo python3 pigs.py 
# Ctrl-C to exit/stop server
#
# Companion Spreadsheet with install and development Notes.
# https://docs.google.com/spreadsheets/d/1fg2dEN390CaZ0fP0DSSJjdO8EyvbOL_oaIsbLogmE8I/edit?usp=sharing
# wordpress site for pigs updates: pigs.goihl.com   
# AVS Forum for PiGS: www.avsforum.com/threads/pigs-the-channel-master-dvr-ep-guide-server-replacement.3155354/
#---------------------------------------------------------------------------------------

PIGS_VERSION = "V1.38 2022-May-14"   # must be this format for version update check
  
# See ReleaseNotes.txt for detailed version history
#
# ToDo's
# use import systemd.daemon to setup Auto-Start for user?
# Make a user.json file with setting that user can change
# make SD calls return  a dict of info instead of 3 params
# Aug 21, SD requires a token as a URL parameter to fetch show images (not chan logos)
# Test if Builder thread is alive on web request from PC or DVR, send email, restart?
# Expand 'Mark as new' concept...  keep a list of recorded episodes, set as new if not already recorded
#  Maybe call it RecordSeries
#  eg America's Test Kitchen repeats episodes multiple time a week/day like 4x or 5x
#    EP005502520424. "Elegant Dinner Party." S19, E7.
#  we can use the EPxxx number to mark it as new just 1x
# On SaveSetting do a json diff to see if only DNS has changed, no build if so
#   https://towardsdev.com/how-to-compare-and-find-difference-between-two-json-object-in-python-22cf86b7da3b
#   pip3 install deepdiff
#   from deepdiff import DeepDiff
#   result = DeepDiff(js1, js2)
#   'dictionary_item_added': [], dictionary_item_removed': [], dictionary_item_changed': [], 
# Maybe scan saved dvr list to set 1st build to 1hr before
# SD servers get very Busy as midnight moves across the USA.  Maybe have a build exclusion time from
#   11pm EST for 6 hrs?  Can still 'force' a build from UI SaveSettings if user wants it.
#   could/should disable dns until delayed build occurs
#   BuildExclude = { 'Enable'=True, 'Start'='time', 'Hours'=6, 'DisableDNS'=True}
#
# ***** Save to Disk on Each build, make pkl save to disk each time to cache is current
#
# V1.38
#  Program Schedule and Cache Saved to disk on good build in xxx.pkl
#    Needed way to keep Cache while making code changes 
#    Also ensures a Schedule if there's an SD outage on PiGS boot
#    Added key "SaveSchedule" : 2,
#       0=Dont save, 1=Save Cache, 2=Save Both Schedule & Cache
#    Saved files are loaded if they exist.  Can be manually deleted w/o harm
#    Cmd line flag -n -noload prevents Both from loading
#  Make sure md5 Stale Cache items are in ToSaveSet() so Cache current
#  DNS was left disabled after a failed build (during 20min retry) even
#    if a sched was available (yesterday or loaded)
#    Now DNS is enabled after a build if a schedule is available
#  Made Test_Webserver work when port<>80, allows multiple PiGS instances (for dev)
#  Build days displays (-1/+17 days) which means 1 prior day, 17 fwd counting today
#
# V1.37
#  Change 'Removing stale program[]...' Cache message to .debug, then display a final count of removed
#  More build logging, Do SD Post retries on SD 50x errors as they can be quite transient
#  Remove Unused 'Chanfilter' key
#  Yikes, in 1.35/1.36 I had a default value of '07:00' in FixedBuildTime, so NEW installs will build every day at 7am
#    so in 1.37 I remove FixedBuildTime and Change to StaticBuildTime to fix those few installs
#  Create UpgradeConfigFile() near Config keys to group config update near definition
#  Add "LogFileBytes" : 250000, "LogFileCount" : 3   to config file
#
# V1.36 2022-May-01
#  Add time a Cache item was last accessed & purge old items, otherwise it could grow indefinitely
#  Add some build info logging
#  Test PiGS version from pigs.goihl.com/version.txt, notify user if update needed
#  index.html, pop up message if Update needed
#
# V1.35
#  Remove LastSchdBuild & LastSvcBuild keys from config dictionary.
#  Chg timer.SetName(), .SetDaemon() due to deprication
#  remove geteuid() test for root, it only works on Linux.
#
# V1.34 2022-Apr-24
#  Fix boundary condition of ls.py when everything is in cache, nothing to look up at SD
#
# V1.33 2022-Apr-22
#  Change logging of SD Api
#  Change Webserver Testing to a HTTP HEAD with a timeout of 10 seconds
#  Use an event to sync WS Reset to Builder so messages are not interleaved
#
# V1.32 2022-Apr-21
#  Bug in Get token when no UN/PW, Tie Program Cache Clear to Build mode
#
# V1.30 2022-Apr-20
#  SD's is(will be) restricting repeat Program ID's requests so duplicate
#  PID's need to be Cached locally
#    BuildLinearSchedule() restructured to use Python Sets & memory Cache
#  Changed SD http API requests to HTTPS because thats what SD wants
#
#  PiGS requires Python 3.7 to use the threading http server
#    Added a hard exit and Help Message if Python 3.7 is not installed
#  Franks Pi zero was having issues with 1.24
#    sd.py Increased connection timeout, added 5 retries to SD API's
#  1.24's MarkAsNew code did not work...
#    I broke it with too much optimization.  Now fixed.
#  Add Fixed build time - eg Build at 7:23am ever day  
#    PiGS can build at a fixed time each day by setting "FixedBuildTime": "07:00"in pigs.json
#    This is not recommended since a DVR could be in the middle of an update and pigs would just
#    disable dns and do its build possibly causing dvr to timeout with incomplete schedule update
#  Added key 'ThreadedHTTP' to choose normal or threaded http server.
#    I noticed some missing sched data when multiple dvrs requesting concurrently.
#    appeared that two sched searches were occurring at same time, maybe confilicting?
#    ls.py or http handler may not be thread safe???
#    So reversing the 1.24 threadedHttp in lieu testing port 80 and resetting server if needed.
#   https://bugs.python.org/issue31639
#   I can simulate this problem using the Python interactive GUI
#     >>> import socket
#     >>> s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#     >>> s.connect(('192.168.1.200', 80))
#   If I now go to a browser and try 192.168.1.200/pigstatus it hangs
#   As soon as I execute :
#     >>> s.close() 
#   the webpage immediately loads
#   the blocking seems to happen in http.server
#   https://github.com/python/cpython/blob/3.10/Lib/http/server.py
#       handle_one_request(self):
#           self.raw_requestline = self.rfile.readline(65537)
#
#
#---------------------------------------------------------------------------------------
# ToDo list
#
# scan code for ToDo's and fix or remove
#
# Could this code base be used with another guide provider?
# Probably.  The key tasks would be to :
#     1) Build the list of channel services BuildServicesListNEW() using their data
#     2) Build the Linear schedule list BuildLinearSchedule() using their data
#     3) Plus a lot of UI stuff, like out how to login, configure the lineups(UI changes)
# The sequencing in this app was built based on the api provided by Schedules Direct
# If you were able to emulate the functions in SD.py (ie reformat everything there)
# using another providers data to provide equivalent results, it might go pretty 'smoothly'.      
#---------------------------------------------------------------------------------------

PIGS_LOG_FILE     = "pigslog.txt"
MAX_LOGFILE_BYTES = 250000  # 150K then rotate log file to next
LOGFILE_BACKUPS   = 3       # Current file and 3 backups

#---------------------------------------------------------------------------------------
# Json format configuration file

# PiGS can act as a DNS server.  This dict definition acts like a hosts file
# Pigs dns will translate name found here, but pigs dns uses gethostbyname()
# so entries in /etc/hosts will be checked if an entry is not found here
#   HostDict is checked, if not found, then
#   /etc/hosts is checked, if not found, then
#   use the configured primary dns server (probaby router gateway)

HostsDict = {                               # PiGS DNS will replace '*' with the discovered IP or Server_IP 
    'epg.channelmastertv.com' : '*',        # fw 1.35 guide requests go here
    'roproxy-guide.echostarcdn.com' : '*',  # fw 1.34 and before guide requests go here
    "stb.channelmastertv.com": "0.0.0.0",   # This one blocks requests to dead Channel Master TV service
                                            # These two block firmware updates 
    'tr50.dishaccess.tv': '0.0.0.0',        # A tr50.dishaccess.tv CNAME stbfw.echodata.tv A 67.148.153.225
    'stbfw.echodata.tv' : '0.0.0.0',        # firmware update url
    
    # These are here for documentation, leading x makes name not match
    # They will all be forwarded by PiGS DNS, but could be redirected here if needed
    
    'xpool.ntp.org' : '162.159.200.123',    # NTP server requests,  forwarded
    
                                            # DVR+ uses these next 4 to determine if it's Internet connected
    'xechostar.com' :  '69.164.198.192',    # with the leading x dns is forwarding these IP's to a real DNS server.
    'xmicrosoft.com': '173.255.215.209',    # These are just some ntp server IPs addresses I found
    'xamazon.com'   : '104.236.116.147',    # DVR+ doesnt like it if I return IP's on the same network
    'xgoogle.com'   : '138.236.128.112'     # DVR+ doesnt access these addresses, it just wants a DNS response.
}

CONFIG_FILE = 'pigs.json'

# Config json file is read into this global config dictionary at boot and on UI Save

dPiGS = None    # global config dictionary, empty until pigs.json is loaded

httpd = None  # webserver handle

  
# This is the layout of the pigs.json master config file
# IF NEW KEYS ARE ADDED TO THE DICTIONARY, THEN ADD AN INIT LINE IN UpgradeConfigFile() below 

PigsConfigDict = {
                          # Required keys
  "sdusername"   : "",    # your SD username
  "sdsha1pwd"    : "",    # your SD password, sha1 encoded
  "sdApiVersion" : "",    # Allows for testing different SD API's
  "Server_IP"    : "0.0.0.0",   # 0.0.0.0 means pigs will monitor port 80 on all IP's (wire & wired interfaces)
  "ThreadedHTTP" : False, # Use threaded HTTP.server.  not sure all code is thread safe.
  "IDE_Port"     : 0,     # http port over-ride for development, typ 8000.  0 means do not override
  "RebuildSecs"  : 82800, # 82800 = 23 hours = 23h*60*60 Next Build Time in seconds
                          # ????If Rebuild secs are set to 0, Pigs will only Auto-Build at same time each day.
  "StaticBuildTime": "",  # Build hour and minute in 24 Hr time  07:00 or 3:30pm = 15:30
                          # FixedBuildTime In 1.35/1.36 I had a 07:00 build time by mistake
                          # 
  "FirstBuildDelay": 30,  # First build occurs 30s after boot to allow platform to stabilize. 0 disables build
  "BuildModeDelete": False, # False= Update Mode - Keep last build until todays is sucessful.
                            # True = Delete sched before build(saves memory for PiZero?)
  "NetworkWait"  : 60,    # Wait for up to 60s for Network to become ready
  "AutoSaveConfig": 2,    # Save config file every N Builds, 0 will disable auto-save
  "IconType"     : 1,     # 0=No Icons, 1=Sched Direct, 2=CM/Rovi Icons, 3=CutePigs???
  "CacheCMIcons" : True,  # keeps a local copy of the CM / Rovi Icons
  "CacheSDIcons" : True,  # keeps a local copy of the resized SD Icons
  "SaveFsSDIcons": False, # True Saves a copy of the FullSize SD icons in ./sdicons for external manipulation
  "SDImSize"     : [72,54], # 360w x 270 / 5 = 72 x 54 (json cant do Tuple, so will use a list)
  "SDImResample" : 3,     # Default to type 3.   Pillow Lib .resize() Resample type 0..5
  "SDImReduceGap": None,  # Pillow 7.0 Reducing Gap 
  "SdIconTheme"  : "dark", # choices are dark, gray, light, white
  "UpdateIconCache" : False, # set true when icon caches need to be updated
  "DefaultIcon"  : "pinkpig.png",  # if no guide icon is found... another could be "92x36box.png"   
  "PrependEpisode" : True,  # Prepend Episode Title, SeasonEpisode #'s, FirstAirDate
  "AppendCast"   : True,  # Append cast info to description
  "LoggingLevel" : 20,    # default logging level. DEBUG 10,INFO 20,WARNING 30, ERROR 40,CRITICAL 50
  "LoggingFile"  : PIGS_LOG_FILE,     # filename of log file, like pigs.logs.
  "LogFileBytes" : MAX_LOGFILE_BYTES, # 150K then rotate log file to next
  "LogFileCount" : LOGFILE_BACKUPS,   # Current file and 3 backups
  "AlertEmails"  : "",    # csv emails to receive alerts "Name1<email1@abc.com>,Name2<email2@xyz.com"
  "SmtpUser"     : "",    # gmail address  "someaddr@gmail.com"
  "SmtpPass"     : "",    # gmail password, make a new account just for pigs
  "PiGS_Version" : PIGS_VERSION, # Version string so UI index.htm can display
  "PiGS_NewVer"  : 0.0,   # version from http://pigs.goihl.com/zips/version.txt to print update msg 1.36
  "CheckVersion" : True,  # check pigs.goihl.com/zips/version.txt for new version of PiGS
  "SaveSchedule" : 2,     # 0=Dont save, 1=Save Cache, 2=Save Program Schedule & Cache 
  "PiGS_IP"      : '0.0.0.0', # Read-Only. IP Addr string so UI index.htm can display
  "PinkPigPos"   : 0,     # display 1 pink pig icon at pos N on the guide or on Chan x.y, use 0 to disable
  "AllowAnyZip"  : True,  # allow any text in Zip box.  False = USA 5 digit, numbers only
  "PigsDoesDNS"  : True,  # PiGS acts as a DNS Server for DVR+.  Forward requests for msft,google,amzn, etc
  "PigsDelayDNS" : True,  # PiGS starts DNS server thread after Schedule is built  
                          # Development keys
  "ChunksEnabled": False, # Enable Http/1.1 chunks for schedule send
  "MGRnewSched"  : True,  # Manual Guide Refresh forces Rebuild of SD Schedule
  "debugToken"   : "",    # if this key exists, it over-rides the SD token request
  "DaysLimit"    : 0,     # Num days fetched from SD (DVR+ can only use 15), 0 for Max Days, less than 17 yields 403 errors
                          # DaysLimit was used to limit day to 2 or 3 for fast builds during development.
  "DisplayDVR"   : 7,     # pigstatus display dvrs updated within last N days, 0 for all 
  "LogHttpReq"   : False, # debug enable annoying http logs on each send
  "SaveJson"     : False, # save the PiGS Channel Lineup sent to DVR+ in DvrLineup.json for troubleshooting
  "TimeZones"    : [      # This gets copied into the DVR+ message response headers
      {
        "StartDateTime" : "2020-03-08T08:00:00Z",
        "EndDateTime"   : "2020-11-01T07:00:00Z",
        "Offset" : "???"  # ??? will be initialized with users local as found in their system
      }
    ],
  "DnsHosts" : HostsDict, # DNS Host dictionary for PiGS DNS server (functions like /etc/hosts )
  
  "DVR_States" : {},  # States of dvr's using pigs stored here 
  
  "MarkAsNew": [   # 
         # [ "America's Test", "S21", ForceNotNew],   # Example mark all Season 21's of America's Test Kitchen as New
         [ "Show1", "Season S3 or other Copy to match", False],
         [ "Show2", "Season or other Copy to match", False],
         [ "Show3", "NoComma after Last item in list", False]
      ],
                          # Finally, list of users lineups, limited to 4 by Schedules Direct
  "MyLineups"    : []     # List of LineupDef objs with users favorite stations flagged
                          # Multiple Lineups allow user to pull schedule data from far away stations.
                          # e.g. their local zipcode might not include all the stations they can receive.
                          # MyLineups will be used to build the ServiceDetails station List DVR+ requests  
}


def UpgradeConfigFile() :   # Auto-Upgrade pigs.json

    if 'SaveSchedule' not in dPiGS :      # 1.38 keys
        dPiGS["SaveSchedule"] = 2         # pickle sched and cache to disk 

    if 'LogFileBytes' not in dPiGS :      # 1.37 keys
        dPiGS["LogFileBytes"] = MAX_LOGFILE_BYTES  # 250K then rotate log file to next
    if 'LogFileCount' not in dPiGS :      # 1.37 keys
        dPiGS["LogFileCount"] = LOGFILE_BACKUPS  # Current file and 3 backups

    if 'ChanFilter' in dPiGS :           # Remove unused key
        dPiGS.pop("ChanFilter",0)

        # version from http://pigs.goihl.com/zips/version.txt to print update msg 1.36
    if 'PiGS_NewVer' not in dPiGS :      # 1.36 keys
        dPiGS["PiGS_NewVer"] = 0.0
    if 'CheckVersion' not in dPiGS :      
        dPiGS["CheckVersion"] = True

    if 'LastSvcBuild' in dPiGS :         # Remove unused keys in 1.35
        dPiGS.pop("LastSvcBuild",0)
    if 'LastSchdBuild' in dPiGS :        # Remove unused key
        dPiGS.pop("LastSchdBuild",0)

    if 'ThreadedHTTP' not in dPiGS :     # 1.30 Use threaded HTTP.server. not sure all code is thread safe
      dPiGS["ThreadedHTTP"] = False  

    if 'StaticBuildTime' not in dPiGS :  # 1.30/1.37 Clock scheduled build time
      dPiGS["StaticBuildTime"] = ""      # Build H,Min 24 Hr time  7:00AM "07:00"  3:30pm = "15:30"
    # in 1.30-1.36, I had an accidental default time of '07:00' in 'FixedBuildTime' delete so DVR auto schedules
    if 'FixedBuildTime' in dPiGS :
      dPiGS.pop("FixedBuildTime",0)      #  Remove FixedBuildTime key, now its called StaticBuildTime
     

    if 'sdApiVersion' not in dPiGS :     # 1.24 Allows for testing different AD API's
      dPiGS["sdApiVersion"] = ""         # Empty means no version over-ride
    if 'BuildModeDelete' not in dPiGS :  # 1.24 
      dPiGS["BuildModeDelete"] = False   # True = delete current schedule before starting build

    if 'AlertEmails' not in dPiGS :      # 1.24 CSV email list to receive alerts 
      dPiGS["AlertEmails"] = ""          # 
    if 'SmtpUser' not in dPiGS :         # 1.24 smtp email/gmail address  "someaddr@gmail.com"
      dPiGS["SmtpUser"] = ""             # 
    if 'SmtpPass' not in dPiGS :         #  1.24 smtp email/gmail password, make a new account just for pigs
      dPiGS["SmtpPass"] = ""             # 

    if 'AutoSaveConfig' not in dPiGS :   # 1.24 Save config file ever N days
      dPiGS["AutoSaveConfig"] = 2    # 

    if 'MarkAsNew' not in dPiGS :        # 1.24 mark as new shows
      dPiGS["MarkAsNew"] = []    #
      dPiGS["MarkAsNew"].append( ["Show1", "Season or other Copy to match", False] )
      dPiGS["MarkAsNew"].append( ["Show2", "Season or other Copy to match", False] )
      dPiGS["MarkAsNew"].append( ["Show3", "NoComma after Last item in list", False] )
      
    if 'NetworkWait' not in dPiGS :   # First build delay is 30s after reboot/start 
      dPiGS["NetworkWait"] = 60   # Wait for up to 60s for Network to become ready

    if 'UpdateIconCache' not in dPiGS :   # 1.24 icon cache update when changed
      dPiGS["UpdateIconCache"] = False    # 

    if 'FirstBuildDelay' not in dPiGS :   # First build delay is 30s after reboot/start 
      dPiGS["FirstBuildDelay"] = 2   # Upgrades get 2s so as not to mess with existing installs

    if 'IDE_Port' not in dPiGS :     # change default http port for working in IDE
        dPiGS["IDE_Port"] = 0        # Default new key to 0
    if 'Server_Port' in dPiGS :      # Remove unused key
        dPiGS.pop("Server_Port",0)
        
    if 'ChunksEnabled' not in dPiGS :     # Enable Http/1.1 chunks for schedule send
        dPiGS["ChunksEnabled"] = False    # Default new key to false, Chunking requires HTTP/1.1

    if 'DnsHosts' not in dPiGS :        # PiGS acts as a DNS Server for DVR+.  Forwards some requests.
        dPiGS["DnsHosts"] = HostsDict   # Add HostsDict if it does not exist
    if 'stb.channelmastertv.com' not in dPiGS['DnsHosts']:  # add host I forgot
        dPiGS['DnsHosts']["stb.channelmastertv.com"] = "0.0.0.0"

    if 'DisplayDVR' not in dPiGS :      # pigstatus display dvrs updated within last N days, 0 for all
        dPiGS["DisplayDVR"] = 7

    if 'IconType' not in dPiGS :        # Icon type 1=SD, 2=CM
        dPiGS["IconType"] = 2 if dPiGS["UseCMImages"] else 1
        dPiGS.pop('UseCMImages',0)      # remove old key nolonger used default 0 prevent exception
        
    if 'PigsDoesDNS' not in dPiGS :     # PiGS acts as a DNS Server for DVR+.  Forwards some requests.
        dPiGS["PigsDoesDNS"] = False    # Default new key to false to not mess up current users
    if 'PigsDelayDNS' not in dPiGS :    # PiGS acts as a DNS Server for DVR+.  Forwards some requests.
        dPiGS["PigsDelayDNS"] = True    # Default new key to True, start DNS after Schedule is built
    if 'Server_IP' not in dPiGS :
        dPiGS["Server_IP"] = '0.0.0.0'  # defalt "0.0.0.0"

    if 'RebuildSecs' not in dPiGS :
        dPiGS["RebuildSecs"] = 82800  # 23 hours = 23h*60*60 Next Build Time in seconds
    if 'MGRnewSched' not in dPiGS :
        dPiGS["MGRnewSched"] = True    # Manual Guide Refresh forces Rebuild of SD Schedule
    if 'AllowAnyZip' not in dPiGS :
        dPiGS["AllowAnyZip"] = True    # allow any text in Zip box.  False = USA 5 digit, numbers only
    if 'SaveFsSDIcons' not in dPiGS :
        dPiGS["SaveFsSDIcons"] = False # keeps a local copy of the resized SD Icons
    if 'SDImSize' not in dPiGS :
        dPiGS["SDImSize"] = [72,54]    # 360w x 270 / 5   = 72 x 54
    if 'SDImResample' not in dPiGS :
        dPiGS["SDImResample"] = 3      # Deault to BICUBIC.   Pillow Lib .resize() Resample type 0..5
    if 'SDImReduceGap' not in dPiGS :
        dPiGS["SDImReduceGap"] = None  # Only in Pillow 7.0 - Reducing Gap
    if 'CacheSDIcons' not in dPiGS :
        dPiGS["CacheSDIcons"] = True  # keeps a local copy of the resized SD Icons
    if 'CacheCMIcons' not in dPiGS :
        dPiGS["CacheCMIcons"] = True   # keeps a local copy of the CM / Rovi Icons
    if 'DefaultIcon' not in dPiGS :
        dPiGS["DefaultIcon"] = "pinkpig.png"
    if 'SdIconTheme' not in dPiGS :
        dPiGS["SdIconTheme"] = "dark" # choices are dark, gray, light, white
    if 'TimeZones' not in dPiGS :
        dPiGS['TimeZones'] = copy.deepcopy(PigsConfigDict['TimeZones'])
            
    if dPiGS['TimeZones'][0]['Offset'] == "???" :  # ??? is flag to tell us to init
        # Set the timezone object to the current time zone on each boot
        # ToDo what happens if user changes timezone while pigs is running
        # ToDo how do we update this for 2021?????
        s,offset = LocalTimeStr( datetime.utcnow() )
        dPiGS['TimeZones'][0]['Offset'] = str(offset)


#---------------------------------------------------------------------------------------
# UI Settings of config
# A full copy of the PigsConfigDict is sent to the UI index.htm for manipulation
# Some is for display only, but some is for getting the users settings.
#
# The following PigsPreserveSettings dict is a list of keys that will be preserved
# when the user hits Save Settings.
#
# Anything in this list becomes essentially READ-ONLY - UI cant change it.
# E.g, if a guide update occurred while the user had the UI open and the
# LastSchdBuild time had been updated, we want to preserve that key value since
# it occurred after the UI was opened.
# Another example is the sha1 password.  Initally it does not exist in the config file.
# User presses verify login - UN & PW is sent to the server and the  sha1 is created
# and saved in the server conig, but not sent back to the UI so the UI data is 'out of date,'
# so the server side sha1 is 'Preserved' when the user clicks save settings.
# The key names here must match those in PigsConfigDict exactly
# The values in PigsPreserveSettings do nothing - Just place holders.
# I suppose I could just make a straight List [] object of key names instead of a dict object.

PigsPreserveSettings = {
  "sdsha1pwd"    : "",  # your SD password, sha1 encoded
  "PiGS_Version" : "",  # Version string for loggin
  "PiGS_IP"      : "",  # IP string for loggin
  "debugToken"   : "",  # this key over-rides the token request
  "DaysLimit"    : "",  # limit number of schedule days fetched from SD
  "LogHttpReq"   : "",  # debug enable annoying http logs on each send
  "SaveJson"     : ""   # save DVR+ json request body
}

# A lineupDef is created when the User clicks 'Preview Lineup' with a valid ZIP
# 
# It will be created by PiGS and sent to the UI but exist only on the UI side until the user
# clicks 'Add This Channel Lineup' (added to MyLineups[]) followed by Save Settings (send to server)
# 'Add..." action will add the 'Preview" to the UI copy of config MyLineups[].
# So the MyLineups array can/will change size from what was sent to the UI
# 
# To commit UI changes to the Server, they must click 'Save Settings' to send to PiGS
#
# SD allows for up to 4 lineups
# LineupDef dict is one user lineup: the LineupName for UI display, and list of favorite chans.
# MyLineups[] is a list of LineupDef's in pigs.json

LineupDef = {
  "LineupName" : "LU-Name",   # SD Name for the lineup
  "ChannelList": []           # List of ChannelDef objs. It will include every channel in the lineup
  }

# The ChannelDef object below is for reference only.  The first 4 keys come from SD preview call.
# The SD preview will send the channel,name,callsign,and affiliate keys, I will add the favorite key
# after getting the results from the SD command.

# When the user saves the lineup, the DVR+ services list will be created using the
# channels marked as favorites and info obtained from SD station Map command

# These are the key names sent from the SD lineup preview results, with the addition of
# my favorites key.  we will loop through and add the favorite key after getting the data
# from SD so the favorite key exists for the UI and has an initial value

ChannelDef = {
      "channel"  : "3.1",    # with example values.
      "name"     : "WISCDT (WISC-DT)",
      "callsign" : "WISCDT",
      "affiliate": "CBS",
      "favorite" :  True      # by default.
}

#---------------------------------------------------------------------------------------
# References for HttpServer object
#https://github.com/python/cpython/blob/3.8/Lib/http/server.py

from http.server import BaseHTTPRequestHandler, SimpleHTTPRequestHandler

from io import BytesIO
from urllib.parse import parse_qs, urlparse

# CM/DVR+ bases channel lineups on Service ID's  Multiple ZIP's might map to one service ID
# PiGS will be configured up front using Schedule Direct lineups 
# Going to use a 'dummy' service ID of 99999 everywhere.
# Every DVR+ zip request will get a service id of 99999
# The 99999 number is important b/c it must be consistent across all DVR+ service calls,
# listing calls, and program airings or DVR+ wont display it.
# The actual channel lineups will be configured in the UI / via Schedules Direct

PiGS_SERVICEID = 99999

#running starting a program at bootup on a Pi
#https://www.dexterindustries.com/howto/run-a-program-on-your-raspberry-pi-at-startup/

#https://docs.python.org/3.7/library/logging.html?highlight=logging#module-logging
#https://www.blog.pythonlibrary.org/2014/02/11/python-how-to-create-rotating-logs/

import logging
import logging.handlers
import time
from   datetime import datetime, timedelta
from   dateutil import parser
import json
import gzip, zlib  # jsgz = gzip.compress(js), zlib does stream
import sched
import sys
from   copy import copy
import copy

# Functions to send data to DVR+
from cm import CMInitAltServer
from sd import GetSha1Password, SDGetToken, SDGetLineupPreview, SDAddDeleteLineups, SDSetAPIVersion
from sv import GetServiceIdFromZip, GetServiceDetails, BuildServicesListNEW, GetServiceList, GetSVCBuildInfo
from ls import BuildLinearSchedule, GetLinearSchedule, GetChunkedLinearSchedule
#LSClearSchedule, 
from ls import LsGetDVRStatus, LS_GetDictOfDVRs, LS_SetDictOfDVRs
from ls import LSGetProgramSchedule, lsr, LSBuildStat
from se import SearchSchedule, DisplayLineup

from pigsdns import PigsDnsStart, PigsDnsStop, PigsDnsAlive, PigsDnsStatus

# utility functions
from ut import LocalTimeStr
from ut import VersionStringToFloat
from ut import SendEmailAlert
from ut import EM_Initialize
#from ut import test_serverport
from ut import test_WebServer, utGetPigsVersion

bLogHttp = False  # a flag to stop some annoyingly large httplib debug messages

#https://docs.python.org/2/library/os.path.html
import shutil
from os import path
import os
import signal

#=========================================================================
# Save config file, keep a history of 5 backups
# https://docs.python.org/3/library/shutil.html
# https://www.w3schools.com/python/python_file_remove.asp
# https://docs.python.org/3/library/shutil.html
# https://docs.python.org/3/library/os.path.html
# https://docs.python.org/3/library/os.html#module-os
# I wonder if there is some type of Rotation Save function in Python

CFG_HISTORY = 5

def SaveSettingsFile(configDict, configFile=CONFIG_FILE) :
    
    # separate file path/name from its extention
    root, ext = path.splitext(configFile)

    # delete the oldest log file #5
    dst = root+str(CFG_HISTORY)+ext
    if os.path.exists(dst):
        os.remove(dst)
    
    # then rename files to keep history
    # ren 4 to 5, 3 to 4, 2 to 3, etc
    for i in range(CFG_HISTORY-1,-1,-1) : # start, stop 1 before, step
        src = root+str(i)+ext if (i>0) else root+ext  # strip '0' for current file 
        dst = root+str(i+1)+ext # to 5
        if os.path.exists(src):
            os.rename(src, dst)  
        #src = path.realpath("guru99.txt");
        #os.replace(src, dst, *, src_dir_fd=None, dst_dir_fd=None)

    # and write config file to storage 
    # pretty json with indent=2.  sorted is ugly and ungrouped
    with open(configFile, "w") as outfile:    
        json.dump(configDict, outfile, sort_keys=False, indent=2) 
    
    outfile.close()

    #TodDo  # seems to be only 644 so user cant delete file. do i want that???

#---------------------------------------------------------------------------------------
    
def CheckForPigsUpdate() :
    
    global dPiGS    
    pver = 0.0
    msg = ""

    # get current ver from pigs.goihl.com/zips/version.txt
    if dPiGS["CheckVersion"] :
        logging.info("MAIN: Checking for a new version of PiGS...")
        pver = utGetPigsVersion()  

    dPiGS["PiGS_NewVer"] = pver  # save in config file for UI Alert Msg
    
    # Strip version from string "V1.35 2022...."
    if pver > float(dPiGS["PiGS_Version"][1:6]) : 
        
        msg = 'There is a new version(%s) of PiGS at http://pigs.goihl.com' % (pver)
        logging.warning("*****************************************************************")
        logging.warning("* " + msg + " *")
        logging.warning("*****************************************************************")
        
    return msg

#---------------------------------------------------------------------------------------
# The schedule build can take a long time, up to a minutue or two (or even more).
# To allow the web server to respond to multiple DVR requests, UI commands, and
# web pigstatus requests in real-time, the build will be run in a separate thread.  
# If a DVR request comes in while the Build is in progress, the webserver will report
# a http 503 server busy error.
# This will allow the webserver to respond to all requests instead of ignoring requests
# when the build is in progess.
# It appears requests were stacking up, timing out, confusing the http.server resulting in
# low level "broken pipe" errors.
#
# A Build Timer thread is used to build the schedule about 1 hour before a DVR typically requests it.


import queue
import threading

BuilderThread   = None               # Thread Handle
BuilderQueue    = queue.Queue()      # Queue to send Buildxx commands to Builder
BuildAbortEvent = threading.Event()  # Event used to abort Build if needed.  isSet()  set()  clear()  wait(timeout)
SDResourceLock  = threading.Lock()   # Schedules Direct Resource Lock, only 1 command can use it at a time

WebserverResetEvent = threading.Event() 
WebserverResetEvent.clear()


BuildLogin     = "Login"
BuildServices  = "Services"    
BuildSchedule  = "Schedule"
BuildTerminate = "Terminate"
BuildAddRemove = "AddRemove"
BuildPreview   = "Preview"
BuildStartDNS  = "DnsStart"
BuildWebserver = "WebServer"
BuildDelay     = "Delay"

# Initialization can fail for various reasons - DNS errors, SD/CM server errors,
# Keep track of where we are so we know what to build

from enum import IntEnum, auto

class InitState(IntEnum) :
    NOT_READY   = 0    # Not Initialized
    CM_ALT_URL  = 1    # Init the CNAME url for CM
    SD_LOGIN    = 2    # Test SD Login Credentials
    SD_LINEUPS  = 3    # Verify PIGS matches SD Lineups
    SERVICE_RDY = 4    # DVR Service List Built
    SCHEDL_NOTRDY=5    # Schedule has been cleared pending build
    SCHEDL_RDY  = 6    # Schedule has been Built
    #INIT_COMPLETE = auto()  # SD verified, Schedule built
    
ReadyState = InitState.NOT_READY

AutoSaveCounter = 0  # config file auto-save counter

def ResetReadyState() :
    global ReadyState
    ReadyState = InitState.NOT_READY

# Commands to the Builder are sent via a dict structure. 'Command' key is required.
# The remaining keys are defined by the command
BldCmd = {
    'Command' : None,    # one of the Above Buildxx commands
    'Params'  : None,    # Params to pass in.  See BuildAddRemove for example
    'Event'   : None,    # If the caller needs to be signaled when result is ready. See BuildPreview for example
    'Result'  : None     # Results to send back to caller. A dict works good.  See BuildPreview for example
  }


# Handles to 2 build timers - One started by DVR, and one as a Watchdog if DVR never requests data
WDogBuildTimer  = threading.Timer(1,None)   # watchdog build timer 
DvrBuildTimer   = threading.Timer(1,None)
NextBuildTimeStr= ""
AutoBuildTimeStr= ""

# This is a watchdog timer - it ensures that the schedule gets built at lease every 24 hrs
# This timer is activated right after the DVR build timer fires in case a DVR never contacts pigs to reset timer

def DoWDogBuildTimer():
    
    if ( dPiGS["StaticBuildTime"] ) :
        logging.warning("THRD:DoWDogBuildTimer: Performing Daily Auto-Build")
    else :
        logging.warning("THRD:DoWDogBuildTimer: has fired. Where are the DVR's?")
        
    SetDvrBuildTimer(0.1)       # Trigger the normal Timer build process 0.1 second from now
    # Kicking off the Build timer will also reset this Auto build timer.

# DoTimerBuild is called when the DvrBuildTimer thread fires
# The Timer thread is dead after the Build command is Queued
# The build timer will be reset when a dvr makes a schedule request.
# The Timer build is not scheduled if the username/pwd has not been set (like on first run)
# 

def DoTimerBuild():
    global WDogBuildTimer
    global AutoBuildTimeStr
    global ReadyState
    global httpd
    global RestartServerRequest
    global AutoBuildTime
    global port # webserver port
    
    logging.info("THRD:DoTimerBuild: has fired. Waiting for SDResourceLock Lock...")

    DoInit = True

    # Do a hard wait for the SD Lock here.  This waits until the Build Queue is empty
    if SDResourceLock.acquire(True) :
        
        logging.info("THRD:DoTimerBuild: Acquired SDResourceLock.")
        
        # Todo - SDResourceLock there could be race conditions if commands get processed faster than they are

        updateMsg = CheckForPigsUpdate()  # check for newer PiGS version

        # Verify that PiGS Webserver (typ port 80) is responding to http requests
        # this might break if PiGS started with a port other than 80. not after 1.38

        wsAddr = dPiGS['PiGS_IP'] + ":" + str(port)
        
        logging.info("THRD:BUILDER: Testing PiGS Webserver at %s", wsAddr)
        
        if test_WebServer( wsAddr ) == 0 :
            
            logging.info("THRD:DoTimerBuild: WebServer is Responding")
            
        else :
            logging.warning("*******************************************************************")
            logging.warning("****** THRD:DoTimerBuild: PiGS WebServer is not responding!  ******")
            logging.warning("*******************************************************************")
            
            logging.warning("THRD:DoTimerBuild: Queuing WebServer Restart.")
            BuilderQueue.put( {'Command':BuildWebserver} )
            
            # From https://docs.python.org/3/library/socketserver.html#socketserver.TCPServer
            # httpd.shutdown()        # This will like a Ctrl-c to break main loop(doesnt work)
            # Tell the serve_forever() loop to stop and wait until it does. shutdown() must be called
            # while serve_forever() is running in a ***different thread*** otherwise it will deadlock.
            # signal main loop to restart WebServer
            #RestartServerRequest = True
            # Setting private variable __shutdown_request = True will break the serve_forever() loop
            #httpd._BaseServer__shutdown_request = True
            
        # Added ReadyState Initialization in V1.22.
        # If Startup Initialization did not complete due to some error,  complete it here.
        # If after Build Schedule, its not built, the WatchDogtimer will be set for 20-minutes
        
        if ReadyState <  InitState.CM_ALT_URL :
            logging.info("THRD:DoTimerBuild:INIT: Setting Default CM Server.")
            cmAltURL = CMInitAltServer()
            ReadyState = InitState.CM_ALT_URL  # not going to fail if cant get Alt URL
            #if CMInitAltServer() :  # we can do this right here.  
            #    ReadyState = InitState.CM_ALT_URL
        
        # Test SD Connection
        if ReadyState <  InitState.SD_LOGIN :
            logging.info("THRD:DoTimerBuild:INIT: Verifying SD connection(GetToken)")
            un = dPiGS['sdusername']     # default: use username and pwd from pigs.json
            pw = dPiGS['sdsha1pwd']
            code, tokencode, errj = SDGetToken( un, pw, dPiGS.get('debugToken'))
            if (code == 200 and tokencode == 0) :  # Token get successful -
                ReadyState = InitState.SD_LOGIN
            else :
                logging.warning("THRD:DoTimerBuild:INIT: SD_LOGIN Error(%d) %s", code+tokencode, errj)
                DoInit = False
                # if we could not get a token for whatever reason, abort the build commands 
                # BuildAbortEvent.set()   #signal Linear Schedule Builder to abort right away

        if DoInit and ReadyState <  InitState.SD_LINEUPS :
            logging.info("THRD:DoTimerBuild:INIT: Verifying SD Lineup (BuildAddRemove)")
            newLineups = []
            for L in dPiGS['MyLineups'] :
                newLineups.append( L['LineupName'] )
            BuilderQueue.put( {'Command':BuildAddRemove, 'Params':newLineups} )
        
        if DoInit and ReadyState <  InitState.SERVICE_RDY :
            logging.info("THRD:DoTimerBuild:INIT: Queuing BuildServices")
            BuilderQueue.put( {'Command':BuildServices} )
            
        # Schedule is always queued when build timer fires so build
        # can be recheduled if needed. also Build needs to run to release SDResourceLock
        
        logging.info("THRD:DoTimerBuild: Queuing BuildSchedule.")
        BuilderQueue.put( {'Command':BuildSchedule, 'Params':updateMsg } )

    return 

def SetWDogBuildTimer(delaySeconds) :

    global WDogBuildTimer, AutoBuildTimeStr
    
    at  = datetime.utcnow()                # return the local clock time when the timer will fire
    # print('at', at, delaySeconds)
    
    # If User has specified a fixed build time, Calc secs to build at that time tomorrow
    
    if dPiGS["StaticBuildTime"] : # in Local time like "07:30" or "15:40"
        UserBuildTime = parser.parse( dPiGS["StaticBuildTime"] )  #User Hour:Min local time
        tzmins = int(time.localtime().tm_gmtoff/60) # get gmt offset minutes
        lt  = at + timedelta(minutes=tzmins) # adjust at to get current local time
        fbt = lt + timedelta(seconds=delaySeconds)  # Advance to next day
        # but reset H:M to a user requested build time so no slipage occurs
        fbt  = fbt.replace(hour=UserBuildTime.hour, minute=UserBuildTime.minute, second=0)
        dif  = fbt - lt   # a timedelta object calc exact secons needed from now
        delaySeconds = dif.seconds  # set delay to keep time fixed
        
    # When DoTimerBuild() ends, the DvrBuildTimer thread is dead.
    # We expect a dvr to come along within the next hour & reset this timer, but if it does not,
    # If the dvrs are offline, eg the network is down, the sched could get really stale.
    # The Build timer was set for 23 hrs after the a dvr made a schedule request, we now expect
    # that dvr to make a request within an hour.  We will set a 24 hour timer 'watchdog' timer
    # so that if dvr doesnt ask in the expected time, we will do a build anyway.
    # The WDogTimer will get reset right here each time it fires.

    WDogBuildTimer = threading.Timer(delaySeconds, DoWDogBuildTimer)
    WDogBuildTimer.name = 'Auto Build Timer'
    WDogBuildTimer.daemon = True         # thread is automatically killed when program exits
    WDogBuildTimer.start()
    
    at += timedelta(seconds=delaySeconds)
    AutoBuildTimeStr,tz = LocalTimeStr(at, showSeconds=True)

    logging.info("THRD:SetWDogBuildTimer: Auto Build time set for %s.", AutoBuildTimeStr)

    return AutoBuildTimeStr


def SetDvrBuildTimer(delaysecs) :
    global DvrBuildTimer
    global NextBuildTimeStr
    global WDogBuildTimer

    # If the auto build timer is set, cancel it since we are scheduling a build
    try :
        if WDogBuildTimer.is_alive() :
            WDogBuildTimer.cancel()
    except :
        logging.warning("THRD:SetDvrBuildTimer: Exception canceling WDogBuildTimer.")

    try :  # if DvrBuildTimer is currently active, kill it and restart with a new time
        if DvrBuildTimer.is_alive() :
            DvrBuildTimer.cancel()
    except :
        logging.warning("THRD:SetDvrBuildTimer: Exception canceling DvrBuildTimer.")
    
    # have to create a new Timer thread each time, cant reuse or reset thread timers
    DvrBuildTimer = threading.Timer(delaysecs, DoTimerBuild)
    DvrBuildTimer.name='DVR Build Timer'
    DvrBuildTimer.daemon=True             # thread is automatically killed when program exits
    DvrBuildTimer.start()
    
    bt  = datetime.utcnow()                # return the local clock time when the timer will fire
    bt += timedelta(seconds=delaysecs)
    NextBuildTimeStr,tz = LocalTimeStr(bt, showSeconds=True)
    
    return NextBuildTimeStr

# Builder Thread Functions will generally use global variables, but some pass in parameters and return results

def Builder() :
    global ReadyState
    global BuilderQueue
    global BuildAbortEvent
    global SDResourceLock
    global AutoSaveCounter
    global RestartServerRequest
    global httpd

    Terminate = False
    BldStat = 0
    CmdToken = None
    
    while not Terminate :  # Builder continues until terminated (nothing terminates it right now)
        
        # Check if someone is requesting a Builder abort (UI save settings can do this)
        if BuildAbortEvent.is_set() :
            logging.info("THRD:BUILDER: Abort Request. Flushing Command Queue.")
            BuildAbortEvent.clear()
            while not BuilderQueue.empty() :
                BuilderQueue.get()
                BuilderQueue.task_done()
        
        # Builder keeps the SDResourceLock until all commands have been processed
        # If Queue Len is 0 and Builder has the lock, then release the SD Resource Lock
        if BuilderQueue.empty() and SDResourceLock.locked() :
            logging.info("THRD:BUILDER: Queue is Empty, releasing SDResourceLock\n")
            SDResourceLock.release()

        # Before placing a command in the Queue, the initiator requested the SDResourceLock
        # The caller could have submitted multiple build commands
        # Get a build command from Queue.  this Thread will wait/block here for a command
        
        dBuildCmd = BuilderQueue.get()  # dBuildCmd is Dictionary with 1 or more keys
        buildCmd = dBuildCmd['Command']
        logging.info("THRD:BUILDER: Command is %s",buildCmd)

        # First handle non-SD commands that dont need a Token
        if buildCmd == BuildTerminate :
            
            logging.info("THRD:BUILDER: Builder Thread Terminating...")
            Terminate = True
            
        elif buildCmd == BuildDelay:
            secs = dBuildCmd['Params']
            logging.info("THRD:BUILDER: Sleeping for %ss",secs)
            time.sleep(secs)
            
        elif buildCmd == BuildWebserver:
            
            logging.info("THRD:BUILDER: Restarting Webserver...")
            
            # restart will happen concurrent with Build Schedule
            
            # signal main loop we want to restart the WebServer
            RestartServerRequest = True
            
            # https://github.com/python/cpython/blob/3.10/Lib/socketserver.py
            # shutdown() seems to do a hard wait locking up PiGS
            # httpd._BaseServer__is_shut_down.set()
            # httpd.shutdown()        # This acts like a Ctrl-c to break main loop

            # so instead set the flags needed to shutdown the HTTP server manually
            #https://www.netjstech.com/2019/04/name-mangling-in-python.html
            httpd._BaseServer__shutdown_request = True

            # Clear event flag to sync restart with this builder - MAIN will set flag when WS restarts.
            WebserverResetEvent.clear()
            
            #pid = os.getpid()
            #signal.pidfd_send_signal(pid, signal.SIGINT , siginfo=None, flags=0) #python 3.9
            #signal.raise_signal( signal.SIGINT ) # send a ctrl-C, python 3.8
            #os.kill(pid, signal.CTRL_C_EVENT)  # 3.2 windows only?
            
            os.kill(os.getpid(), signal.SIGINT)  # send a ctrl-C

            # wait for up to 10 seconds for Main to restart webserver...
            logging.info("THRD:BUILDER: Waitingfor Webserver to Restart" )
            WebServerReady = WebserverResetEvent.wait(10) 
            
                   
        elif buildCmd == BuildStartDNS:
            
            logging.info("THRD:BUILDER: Starting DNS...")
            
            # ensure we have a PiGS_IP in config to display on UI and for DNS
            if not dPiGS['PiGS_IP'] :  
                IPaddr = WaitForNetwork(1)
                logging.info("THRD:BUILDER:DNS: PiGS Discovered IP = %s", IPaddr )

                if not IPaddr :  # PiGS did not detect an IP address - no Internet Connection?
                    logging.warning("THRD:BUILDER:DNS: Computer does not appear to have an Ethernet connection, cant start DNS" )
                    
                dPiGS['PiGS_IP'] = IPaddr  # save PiGS_IP in config to display on UI
                
            if dPiGS['PiGS_IP'] :  # ensure we have a PiGS_IP to start DNS
                PigsDnsStart(dPiGS['Server_IP'], dPiGS['PiGS_IP'], dPiGS['DnsHosts'])
            else :
                logging.warning("THRD:BUILDER:DNS: Computer does not appear to have an Ethernet connection, cant start DNS" )
                
        else : # its an SD command that needs a token
            
            un = dPiGS['sdusername']     # default: use username and pwd from pigs.json
            pw = dPiGS['sdsha1pwd']
        
            # special case for initial login, un/pw not in dPiGS yet, use params in Command
            if buildCmd == BuildLogin :
                un = dBuildCmd['Params']['un']
                pw = dBuildCmd['Params']['pw']

            # Get a Token
            code, tokencode, CmdToken = SDGetToken( un, pw, dPiGS.get('debugToken'))
                
            if buildCmd == BuildLogin :  # special case for initial login, return token status
                
                # send the result back as a small dict object
                dBuildCmd['Result'] = {'code':code+tokencode, 'errj':CmdToken}
                dBuildCmd['Event'].set()  #signal waiting process results are ready
            
            # GTR IF  ---------
            if (code == 200 and tokencode == 0) :  # GTR Token get successful -
                
                if buildCmd == BuildLogin :  # nothing to process
                    pass
                
                elif buildCmd == BuildAddRemove :
                    try :
                        lineups = dBuildCmd['Params']
                        SDAddDeleteLineups(lineups)
                    except :
                        # failed, build timer will be reset below
                        logging.warning("THRD:BUILDER: BuildAddRemove Lineup Exception occurred! ", exc_info=True)
                    else : 
                        ReadyState = InitState.SD_LINEUPS

                elif buildCmd == BuildPreview :
                    try :
                        code, lujson, luname = SDGetLineupPreview( dBuildCmd['Params'] )  #
                        # send the result back as a small dict object
                        dBuildCmd['Result'] = {'code':code, 'lujson':lujson,'luname':luname}
                        logging.info("THRD:BUILDER:Preview: Setting callers Event flag.")
                        dBuildCmd['Event'].set()  #signal results are ready
                    except :
                        #e = sys.exc_info()
                        logging.warning("THRD:BUILDER: Build Preview Exception occurred! ", exc_info=True)

                elif buildCmd == BuildServices :
                    
                    if ReadyState >= InitState.SD_LINEUPS :  # then we can build service list

                        try :
                            BuildServicesListNEW(dPiGS, PiGS_SERVICEID, CmdToken)

                            # for debug, save the json (large) message sent to dvr+
                            if dPiGS.get('SaveJson') :
                                logging.info('THRD:BUILDER: Saving ServiceDetailsMsg to '"'DvrLineup.json'"' ')
                                sdjson = GetServiceDetails(dPiGS, PiGS_SERVICEID, dPiGS['TimeZones'])
                                file = open('DvrLineup.json', "w")
                                file.write(sdjson)
                                file.close()
                                os.chmod('DvrLineup.json', mode=0o777)

                        except :
                            logging.warning("THRD:BUILDER: Build Service Lineup Exception occurred!", exc_info=True)
                        else : 
                            ReadyState = InitState.SERVICE_RDY
                            
                            if dPiGS["UpdateIconCache"] : # if we did an icon cache update, clear the flag
                                logging.info('THRD:BUILDER: Clearing [UpdateIconCache] flag and saving settings...' )

                                dPiGS["UpdateIconCache"] = False
                                # This saves 5 backps as <name>N.json, where N = 1..5
                                SaveSettingsFile(dPiGS)                            
                    else :
                        logging.warning("THRD:BUILDER: Can't Build Service Lineup! ReadyState(%s)", ReadyState)
                            
        
                elif buildCmd == BuildSchedule :
                    
                    # if services was not built, Schedule cant be built
                    
                    if ReadyState >= InitState.SERVICE_RDY :
                        
                        # Disable DNS during schedule builds
                        if PigsDnsAlive() :
                            logging.info("THRD:BUILDER: Suspending PiGS DNS during Schedule Build...")
                            dnsThread = PigsDnsStop(message='THRD:BUILDER:PiGS DNS disabled during Schedule Build...')
                            dnsThread.join(2.0)  # wait up to 2s for DNS to Stop (should stop in less than 1 sec)

                        try :
                            # Set ReadyState state to SCHEDL_NOTRDY 
                            ReadyState = InitState.SCHEDL_NOTRDY  
                            BldStat = BuildLinearSchedule(dPiGS, PiGS_SERVICEID, BuildAbortEvent, dBuildCmd['Params'])
                            # return 0 means schedule was built 
                            # != 0 not built, but yesterdays could still be there
                        except : 
                            logging.warning("THRD:BUILDER: Build Schedule Exception occurred! Schedule Not Updated.", exc_info=True)
                            BldStat = 100  # unexpected exception
                            #LSClearSchedule() 
                        else :
                            if BldStat :
                                logging.warning("THRD:BUILDER: Build did not Complete. Error(%s)! )", BldStat)
                                
                            if LSGetProgramSchedule() :  # if there is a schedule, even from yesterday, something is ready
                                ReadyState = InitState.SCHEDL_RDY
                    else :
                        logging.warning("THRD:BUILDER: Can't Build Schedule! ReadyState(%s)", ReadyState)
                    
            # GTR IF ELSE ---------
            else :  # GTR Token Request failed
                # this might happen if user has not set their SD UN/PW
                logging.warning("THRD:BUILDER:cmd(%s) GetToken Error(%d) %s", buildCmd, code+tokencode, CmdToken)
                BldStat = 200 # no token
                
            # If PiGS was not able to build the Schedule, schedule a build 20 minutes from now
            # dns resolution failure or SD timeouts may cause build exceptions in SD call requests
                            
            if buildCmd == BuildSchedule :
                
                #if ReadyState == InitState.SCHEDL_RDY :

                # do I have to test if LSGetProgramSchedule()
                
                if LSGetProgramSchedule() and BldStat == 0 :
                    
                    # if build success, schedule a tentative build 24 hours from now
                    SetWDogBuildTimer(24*60*60)
                    
                    # V1.24 AutoSave Config file for changing DNS states
                    if dPiGS["AutoSaveConfig"] : # if a value other than 0 is set....
                        AutoSaveCounter += 1
                        if AutoSaveCounter >= dPiGS["AutoSaveConfig"] :
                            logging.info('THRD:BUILDER: Auto-Saving config settings...' )
                            AutoSaveCounter = 0
                            # This saves 5 backps as <name>N.json, where N = 1..5
                            SaveSettingsFile(dPiGS)                            
                    
                else :  # Build not completed.  Try again in 20 minutes
                    # 20 mins from now (will give 2-3 build chances before dvr needs)
                    # Once the Timer fires and the Build command is Queued, the DvrBuildTimer thread will be dead
                    bts = SetDvrBuildTimer(20*60)
                    logging.info("THRD:BUILDER: BUILD FAILED. Retrying at %s", bts)
                    
                    # Send Email Alert
                    alertmsg = "Build Did Not Complete(%s)!  Retrying at %s.\r\n ." % ( BldStat, bts)
                    SendEmailAlert(alertmsg)
                        
                # Start DNS after build attempt so DVR's dont make requests before Sched is ready
                
                # If we have a schedule, even yesterdays, we can enable DNS
                # This should also enable DNS for a disk loaded schedule if first build fails
                if LSGetProgramSchedule() and not BuildAbortEvent.is_set() :
                    if dPiGS['PigsDoesDNS'] and not PigsDnsAlive() :
                        BuilderQueue.put( {'Command':BuildStartDNS} )
                    
                    
        logging.info("THRD:BUILDER: ReadyState(%s)", ReadyState)
        BuilderQueue.task_done()
        
        # take a 200mS nap. Other code sometimes submits multiple SD commands.
        # ensure those threads had time to submit all the commands it wanted to
        
        time.sleep(0.500)
        
    logging.info("THRD:BUILDER: Builder Thread Terminated!")
    
    return


#---------------------------------------------------------------------------------------
# References
# https://blog.anvileight.com/posts/simple-python-http-server/
# This reference also discusses SSL support, althought DVR+ does not do SSL
# Subclass because I need to serve some files (index.htm, log file, images) but also
# need to handle the GET and POST requests BEFORE passing on the Simple handler

#Server_error_responses https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
ServerErrorCode = 503  #
gWebServerRestartTime = "Not Started."

class PiGS_HTTPRequestHandler( SimpleHTTPRequestHandler ):

    global bLogHttp
    global dPiGS
    global PigsStartTime
    global gWebServerRestartTime
    global RestartServerRequest


    # Scheudle Builder communication
    global BuilderQueue
    global BuildAbortEvent
    global SDResourceLock
    global BuilderThread
    global ReadyState

    # Scheudle Builder Commands
    global BuildDelay    
    global BuildServices  
    global BuildSchedule 
    global BuildTerminate
 
    # protocol_version defaults to HTTP/1.0
    # One big difference is that in HTTP/1.1 keep-alive is assumed (connection re-use)
    # This seems to have big issues with this code, esp the Pi-Zero. got crashes and lockups
    
#    protocol_version = 'HTTP/1.1'  # set to version 1.21 to support chunking
    # Over-ride to send 'Connection : Close' when serving a file to disable keep-alive
    _close_connection = False

    #override/extend StreamRequestHandler/BaseRequestHandler of socketServer
    def XXhandle(self):  #remove XX to use
        print('httpServer.handle() doing one request')
        SimpleHTTPRequestHandler.handle(self)
        print('httpServer.handle() done with request')


    #-------------------------------------------------------------------
    # over-ride log_message to silence the logging messages sent from
    #      self.send_response()  log_message(format, ...) log_error(format, ...)
    # https://docs.python.org/2/library/basehttpserver.html
    # https://stackoverflow.com/questions/3389305/how-to-silent-quiet-httpserver-and-basichttprequesthandlers-stderr-output

    def log_message(self, format, *args):
        if bLogHttp :                    # a key allows me to block or show these messages
            logging.info(format, *args)  # comment this out to completely silence http:lib
            #logging.warning(format, *args)
            pass
        return

    #-------------------------------------------------------------------
    # Over-ride end_headers() to insert/add 'Connection : Close' when serving a file to disable http/1.1 keep-alive
    # we can flag the Simple file server to close the connection after serving a file
    # internally when a file is sent, we will send the Close header right before the flush
    # see source https://github.com/python/cpython/blob/3.7/Lib/http/server.py
    # the persistent connections were causing all sorts of problems with http.server
    
    def end_headers(self):
        if self._close_connection :
            self.send_header("Connection", "Close")
            #logging.info("MAIN:GET: Adding (Connection : Close) header to response")
            self._close_connection = False   # just do it for this one request
            
        SimpleHTTPRequestHandler.end_headers(self)  # do normal end and flush


    #https://realpython.com/intro-to-python-threading/
    #def __init__(self) :
    #    SimpleHTTPRequestHandler.__init__(self)
    #    super(PiGS_HTTPRequestHandler, self).do_GET()
        
    # self.client_address .server .command .path .request_version .headers)
    
    # https://www.geeksforgeeks.org/http-headers-content-type/
    def SendCommonHeaders(self):
        # HTTP/1.1 200 OK\r\n
        # Headers sent by CM servers to DVR+, We will emulate
        #SimpleHTTP sends these two automatically
        #self.send_header("Date:", "Sat, 02 May 2020 16:04:00 GMT") #Sat, 02 May 2020 16:04:00 GMT\r\n
        #self.send_header("Server:", "nginx/1.10.3 (Ubuntu)")
        self.send_header("Cache-Control", "no-cache,no-store,must-revalidate")
        self.send_header("Pragma", "no-cache")
        self.send_header("Expires", "0")
        self.send_header("Vary", "Accept-Encoding")
#        self.send_header("Connection", "keep-alive")
#        self.send_header("Content-Type", "application/json")
        #These will be set by individual commands
        #self.send_header("Content-Length:", "359")  # size of payload, gzipped data
        #self.send_header("Content-Encoding:", "gzip")

    def do_GET(self):   # I intercept GET's before passing up to parent 'fileserver'

        global dPiGS    # uses the global dictionary
        global DvrBuildTimer
        global NextBuildTimeStr
        global AutoBuildTimeStr
        global RestartServerRequest
                                # https://docs.python.org/3.7/library/urllib.parse.html
        o=urlparse(self.path)   # break url into scheme,netloc,path,params, query,fraqment
        qs=parse_qs(o.query)    # parse the query parameters into a dictionary
        
        #logging.debug("GET: Path: %s\nHeaders:%s", str(self.path), str(self.headers))
        #logging.debug("GET: Path: %s", str(self.path))#, str(self.headers))

        # Handle DVR+ of Get service code from Postal Code
        # The DVR+ zip is going to be ignored - we will return 99999 as the Service Code
        
        # If we wanted to send commands to the server from DVR+, we could 
        # use the zip codes as 'commands'.  eg 5000 could be force SD refresh.
        # For example, set the zip to 50000, then do a guide refresh.
        # The 50000 would be interpretted here and start the download
        # Could require one of these to set the daily download time.
        # after the command, the zip should be set to 99999 on dvr+
        # if user did not reset zip to 99999 the cmd would execute each time dvr did a zip request.
        # We'd need to have a Latch flag (did it), and a reset (another zipcode) for this to work
        
        # Get time of request, convert to local time - easier to match up with what we see on dvr stbhealth page
        doGet_time = datetime.utcnow()
        lts, tz = LocalTimeStr(doGet_time, showSeconds=True)
        
        if ("toggledns" in o.path): # This may not play well with the UI
            
            #logging.info("MAIN: Queuing WebServer Restart.")
            #BuilderQueue.put( {'Command':BuildWebserver} )
            
            # If not running, Start it
            if not PigsDnsAlive() :  # do we need to start DNS?
                
                dPiGS['PigsDoesDNS'] = True
                logging.info("MAIN:toggledns: Starting DNS...")
                PigsDnsStart(dPiGS['Server_IP'], dPiGS['PiGS_IP'], dPiGS['DnsHosts'])
                
            else :   # otherwise Stop DNS
                dPiGS['PigsDoesDNS'] = False
                logging.info("MAIN:toggledns: Stopping DNS - waiting for thread to end...")
                dnsThread = PigsDnsStop()
                dnsThread.join(2.0)  # wait up to 2 seconds for DNS to Stop (should stop in less than 1 sec)
                logging.info("MAIN:UIsettings: DNS Stopped... Alive=%s", dnsThread.is_alive())
                
            #https://www.netjstech.com/2019/04/name-mangling-in-python.html
            #httpd._BaseServer__shutdown_request = True
            #RestartServerRequest = True
            self.send_response(200)
            self.end_headers()
        
        elif ("postalcode" in o.path): # DVR+ asking for GET of service ID
            
            logging.info("MAIN:**** Postalcode request from %s at %s.", self.client_address, lts )
            
            # Postalcode only happens on a Manual Guide refresh, or when user does some things on the DVR
            # We will cancel the scheduled build time so that the build time syncs with the next dvr???
            
            self.send_response(200)
            self.SendCommonHeaders()
            self.send_header("Content-Type", "application/json")
 
            # We could parse out the postalcode from the URL sent by DRV+ but we are not using it
            # when DVR+ sends a zipcode, we always return a service code of 99999
            # Send the current TimeZone info for incorporation to DVR+
            # Get a json response object to send to DVR+
            
            srjson = GetServiceIdFromZip("", PiGS_SERVICEID, dPiGS['TimeZones'])
            
             # size of payload
            self.send_header("Content-Length", str(len(srjson))) 
            self.send_header("Connection","Close")
            self.end_headers()                      
            self.wfile.write( srjson.encode() )
           
        elif ("servicedetails" in o.path): #DVR+ asking for GET of service ID
            
            # this message can't happen without DVR+ first calling postalcode, so we dont
            # have to check for the build state.
            # Todo, well maybe it could from another DVR????

            logging.info("MAIN:**** Service Details request from %s at %s.", self.client_address, lts )

            # If PiGS is not ready when DVR contacts...
            # if network was not ready at boot, the service list may not have been built
            # Could for SERVICE_RDY, but just want the whole sched ready 
            
            if ReadyState != InitState.SCHEDL_RDY :
                
                if SDResourceLock.acquire(False) :  # if we got the lock, start the build                
                    logging.warning("MAIN:SERVICEDETAILS: No service list, Initiating build process...")
                    
                    # We have the Lock, we can make SD calls, or Queue commands
                    SDResourceLock.release()  # Build init process will get it's own lock
                    SetDvrBuildTimer(1.0)
                else :
                    # seems like this could never happen - cant be building anything if there is no service list
                    logging.warning("MAIN:SERVICEDETAILS: PiGS is busy rebuilding schedule! (%S)", ServerErrorCode)
                    
                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule")
                return

            # received headers Accept: */*  Accept-Encoding: gzip,deflate
            self.send_response(200)
            self.SendCommonHeaders()
            
            # Send the Channel lineup list of users Favorite Stations/Channels
            # these have been preconfigured in the UI and are in pigs.json       

            # Get service details and Send to DVR+
            # The ServiceDetails list was built at boot and when user saved settings
            
            # Send DVR+ a service list that was built from SD information
            # This contain the channel numbers, image links, station ID's, etc
            # DVR+ keeps a local copy of this
            
            sdjson = GetServiceDetails(dPiGS, PiGS_SERVICEID, dPiGS['TimeZones'])
            
            # gzip encocded response
            sdgz = gzip.compress( sdjson.encode() )  #to utf-8
            #logging.debug("sdlen(%d)  gzlen(%d)\n", len(sdjson), len(sdgz))

            #gzip encoded response
            self.send_header("Content-Type", "application/json")
            self.send_header("Content-Encoding", "gzip")
            self.send_header("Content-Length", str(len(sdgz))) 
            self.send_header("Connection","Close")
            self.end_headers()
            self.wfile.write( sdgz )

        elif ("linearschedule" in o.path): # DVR+ is asking for a GET of listings

            logging.info("MAIN:LINSCHED **** Program Sched Req from %s at %s", self.client_address, lts )
            
            # if network was not ready at boot, this list may not have been built
            # minor issue on 1st boot.  If first build fails, a schedule can still exist if loaded
            
            #if ReadyState != InitState.SCHEDL_RDY :    # INIT_COMPLETE :
            if not LSGetProgramSchedule() : # if no schedule built or loaded from disk, then

                if SDResourceLock.acquire(False) :  # if we got the lock, start the build                
                    logging.warning("MAIN:LINSCHED: No Schedule built, Initiating scheudule build process...")
                    # We have the Lock, we can make SD calls, or Queue commands
                    SDResourceLock.release()  # Build init process will get it's own lock
                    SetDvrBuildTimer(1.0)
                else :
                    logging.warning("MAIN:LINSCHED: PiGS is busy rebuilding schedule! http(%s)", ServerErrorCode)

                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule", )
                return
            
            # could happen if another dvr made a request just after the Build Timer fired
            if SDResourceLock.locked() :  # test the SD lockif                
                logging.warning("MAIN:LINSCHED: PiGS is busy rebuilding schedule! http(%s)", ServerErrorCode)
                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule" )
                return

            # In normal daily updates DVR+ just jumps right in with a schedule request since
            #  it already knows its service details

            # This is how the schedule is pre-built on the fly...
            #   Schedule the next build 23 hrs from now, just before this dvr needs it tomorrow
            # The other dvrs will just use the existing schedule b/c its already been built
            # Use the first request from a dvr to schedule the next build time
            
            # If a StaticBuildTime is specified, dont schedule the dynamic 'Prebuild'
            # if this is a DVR's first request of the day, it sets the inprogress flag true
            
            if qs["inprogress"][0]=='true' and not dPiGS["StaticBuildTime"] :   
                if not DvrBuildTimer.is_alive() :  # and the Build timer is not already set (by another DVR)
                    bts = SetDvrBuildTimer(dPiGS["RebuildSecs"]) # then set it for 23 hrs in seconds from now
                    logging.info("MAIN:LINSCHED: Next Build time is scheduled for %s", bts)
                    # Once the Timer fires and the Build command is Queued, the DvrBuildTimer thread will be dead
                    
            
            if not dPiGS['ChunksEnabled'] :
                
                self.send_response(200)
                self.SendCommonHeaders()
                
                # Send Linear Schedule to DVR+  qs = has the request params sent by DVR+
                
                lsjson = GetLinearSchedule(dPiGS, qs, PiGS_SERVICEID, \
                                           self.client_address[0], dPiGS['TimeZones'] )
                
                # time.sleep(35)  # testing broken pipe error.  DVR closes connection after 30s

                # This is a very large response - typically 4hours x many channels of data
                self.send_header("Content-Type", "application/json")
                self.send_header("Content-Encoding", "gzip")
                lsgz = gzip.compress( lsjson.encode() )  #to utf-8
                #logging.debug("MAIN:LS: lslen(%d)  gzlen(%d)\n", len(lsjson), len(lsgz) )

                self.send_header("Connection","Close")
                self.send_header("Content-Length", str( len(lsgz) )) 
                self.end_headers()
                
                d = datetime.utcnow() - doGet_time
                if d.seconds > 28 :  # DVR closes connection after 30s
                    logging.warning("MAIN:LINSCHED ***** Slow Response getting Request! (%s secs) ***** \n", d.seconds)
                
                try :
                    self.wfile.write( lsgz )
                except BaseException as e:  # need BaseException to catch broken pipe error
                    d = datetime.utcnow() - doGet_time
                    logging.warning("MAIN:LINSCHED Socket write Exception (%s secs). %s\n", d.seconds, e)
                    
            else :  # DONT USE!!!  do Chunked encoding so DVR is not left waiting for a long time
                
                # MUST have protocol_version = 'HTTP/1.1'  at class level for chunked to work
                
                self.send_response(200)
                self.SendCommonHeaders()
                self.send_header("Content-Type", "application/json")
                #self.send_header("Connection", "keep-alive")
                self.send_header("transfer-encoding", "chunked")
                self.send_header("Content-Encoding", "gzip")
                self.end_headers()

#                   Future, do chunking in dedicated function
#                lsjson = GetChunkedLinearSchedule(dPiGS, qs, PiGS_SERVICEID, \
#                                 self.client_address[0], self.wfile )

                first = True
                ChunkCnt = 1
                Done = False

                #compressobj(level=-1, method=DEFLATED, wbits=15, memLevel=8, strategy=Z_DEFAULT_STRATEGY[, zdict])
                #https://docs.python.org/3.5/library/zlib.html#module-zlib
                #The wbits parameter controls the size of the history buffer (or “window size”),
                #  and what header and trailer format is expected.
                #+25 to +31 = 16 + (9 to 15): Uses the low 4 bits of the value as the window size
                #   logarithm, while including a basic gzip header and trailing checksum in the output.
                
                co = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION, zlib.DEFLATED, 16+11)
                
                totaldur = int(qs["duration"][0])      # Total duration requested in minutes
                offset = 0
                chunkDur = 30
                if totaldur < chunkDur : chunkDur = totaldur

                #time.sleep(28)  # testing broken pipe error

                while not Done :
                
                    # todo modify qs
                    lsjson = GetChunkedLinearSchedule(dPiGS, qs, offset, chunkDur, PiGS_SERVICEID, \
                                 self.client_address[0], FirstChunk=first )
                    
                    offset += chunkDur
                    if offset > totaldur : offset = totaldur
                    
                    # done if we processed all teh minutes
                    Done = (totaldur == offset)

                    # or if No data was sent
                    if len( lsr["LinearScheduleResult"]["Schedule"]["Airings"] ) == 0 :
                        Done = True

                    # Need to modify the json packet for chunking
                    # first packet search for ']' at end to strip off ] } } }
                    if first :
                        i = -1
                        while i>-10 and lsjson[i] != ']' :
                            i -= 1
                        lsjson = lsjson[0:i]       # strip off the end of the json list ] } } }
                        
                    # middle packet:  add ',' then strip leading '[' and trailing ']]
                    else : # we get an alist - could have [] or [{...},{...}]
                        lsjson = ',' + lsjson[1:-1]
                    
                    # end packet: strip leading [, add comma?, leave trailing ], add } } }
                    if Done :
                        lsjson += '] } } }'
                    
                    # stream compression with zlib
                    # https://docs.python.org/3.5/library/zlib.html
                    enc  = lsjson.encode()   # to bytes
                    lsgz = co.compress( enc ) 
                    
                    # if last chuck of data, flush the streaming compressor
                    if Done :
                        lsgz = lsgz + co.flush()
                    
                    zippedBytes = len(lsgz)
                    
                    if not Done and 0 == zippedBytes : continue
                    
                    logging.info("MAIN:Chunk: %s, lslen(%d) gzlen(%d)\n", ChunkCnt, len(lsjson), zippedBytes )
                    
                    #d = datetime.utcnow() - doGet_time
                    #d.seconds
                    
                    # ascii chunk header - e.g. '1AF\r\n' hex byte count
                    h = '{:X}\r\n'.format(zippedBytes)
                    #h = '{:X}\r\n'.format( len(enc) )
                    
                    # ascii chunk footer & if Done, append '0' length chunk (chunk header, no data, chunk footer)
                    f = '\r\n'  
                    if Done : f += '0\r\n\r\n'  

#                    logging.info("chunk %s %s", self.server.fileno(), self.wfile )
                    ba = h.encode() + lsgz + f.encode()

                    try :
                        self.wfile.write( ba )
                    except Exception as e:
                        d = datetime.utcnow() - doGet_time
                        #d.seconds
                        logging.warning("MAIN:LS %s secs.  Socket write Exception %s\n", d.seconds, e)

                    #time.sleep(0.5)  # testing broken pipe error

                    ChunkCnt += 1
                    first = False
        
            
        elif ("search" in o.path): # user doing a manual schedule search

            logging.info("MAIN:SEARCH: Query=%s", qs)

            self.send_response(200)
            #self.SendCommonHeaders()
            self.send_header("Vary", "Accept-Encoding")
            # with HTTP/1.1 keep-alive is default.  Server tells client it can close connection
            self.send_header("Connection", "Close")

            if not LSGetProgramSchedule() :
                logging.warning("MAIN:SEARCH: Schedule is not built.  Can't search! ")
                htmltxt = "Schedule is not built!  Can't search - See Logs."
            else :
                try :
                    htmltxt = SearchSchedule(LSGetProgramSchedule(), qs )
                    self.send_header("Content-Type", "text/html; charset=UTF-8")
                except:
                    htmltxt = "Search Exception! See Logs."
                    self.send_header("Content-Type", "text/plain; charset=UTF-8")
                    logging.warning("MAIN:SEARCH: Exception occurred! ", exc_info=True)
            
            # gzip encoded
            self.send_header("Content-Encoding", "gzip")
            segz = gzip.compress( htmltxt.encode() )  #to utf-8
            self.send_header("Content-Length", str( len(segz) )) 
            self.end_headers()
            self.wfile.write( segz )

        elif ("showlineup" in o.path): # user dumping lineup to webpage

            logging.info("MAIN:LINEUP: Query=%s", qs)

            self.send_response(200)
            self.send_header("Vary", "Accept-Encoding")
            self.send_header("Content-Type", "text/html; charset=UTF-8")

            try :
                htmltxt = DisplayLineup(GetServiceList(), qs )
            except:
                htmltxt = "Lineup Exception! See Logs."
                self.send_header("Content-Type", "text/plain; charset=UTF-8")
                logging.warning("MAIN:LINEUP: Exception occurred! ", exc_info=True)
            
            # gzip encoded
            self.send_header("Connection","Close")
            self.send_header("Content-Encoding", "gzip")
            segz = gzip.compress( htmltxt.encode() )  #to utf-8
            self.send_header("Content-Length", str( len(segz) )) 
            self.end_headers()
            self.wfile.write( segz )

        elif ("pigstatus" in o.path): #GET of 

            logging.info("MAIN:PIGSTATUS: PiGS Status request from %s at %s.", self.client_address, lts )

            self.send_response(200)
            self.SendCommonHeaders()
            self.send_header("Content-Type", "application/json")

            # add some more PiGS info to the results
            pigstatus = {}
            
            sts, lz = LocalTimeStr( PigsStartTime, True )
            uptime = datetime.utcnow() - PigsStartTime
            pigstatus['PiGS General'] = {}
            pigstatus['PiGS General']["StartTime"] = sts
            pigstatus['PiGS General']["  Up-Time"] = str(uptime)
            pigstatus['PiGS General']["  Version"] = dPiGS['PiGS_Version']
            pigstatus['PiGS General']["Directory"] = os.getcwd()
            pigstatus['PiGS General']["WebServer"] = gWebServerRestartTime
            
            # PiGS DNS Status 
            pigstatus["PiGS DNS Service"] = PigsDnsStatus()
            
            svcStatusDict = GetSVCBuildInfo()
            pigstatus.update(svcStatusDict)  # and merge it into to pigstatus dict
            
            # Then add the Next Build Time
            msg1 = NextBuildTimeStr if DvrBuildTimer.is_alive() else  "Not Scheduled."
            msg2 = AutoBuildTimeStr if WDogBuildTimer.is_alive() else "Not Scheduled."
            LSBuildStat.SetAlive( BuilderThread.is_alive() )
            LSBuildStat.SetNextBuild(msg1)
            LSBuildStat.SetAutoBuild(msg2)
            pigstatus.update(LSBuildStat._BuildInfo)
                
            pigstatus["DVR Status List"] = '---------------------'
                           
            # Get the status of all DVR's PiGS has updated
            dvrStateD = LsGetDVRStatus( dPiGS["DisplayDVR"]);  #
            pigstatus.update(dvrStateD)  # and merge it into to pigstatus dict
            
            dvrjson = json.dumps( pigstatus, indent=2 )
            
            self.send_header("Connection","Close")
            #unencoded response 
            self.send_header("Content-Length", str( len(dvrjson)) ) 
            self.end_headers()
            self.wfile.write( dvrjson.encode() )
        
        elif ("UIsettings" in o.path): #GET of 

            logging.info("MAIN:GET: UIsettings... from %s ", self.client_address) #, lts)

            self.send_response(200)
            self.send_header("Content-type", "application/json")

            # export dict to json string, then encode str to bytes object
            jstr = json.dumps( dPiGS )   # Send everything to the UI
            
            self.send_header("Connection","Close")
            self.send_header("Content-Length", str( len(jstr) )) 
            self.end_headers()
            
            self.wfile.write( jstr.encode() )
            return
        
        elif ("lineupPreview" in self.path) :  # UI requesting a SD Lineup Preview based on a Zip

            # Preview needs the SD resource lock b/c it uses the SD token.
            # if a schedule build was in progress this UI call would kill the build

            if not SDResourceLock.acquire(False) :  # if we get the lock, we can do the preview
                logging.warning("MAIN:PREVIEW: Can't Preview.  PiGS is busy rebuilding schedule!")
                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule. Please Retry.")
                return

            # Build a LineupDef to send to the UI fo display
            # LineupDef = {
            #   "LineupName" : "USA-OTA-53593",   # SD Name for the lineup
            #   "ChannelList": []                 # List of ChannelDefs.
            # }                                   # It will include every channel in the lineup

            # Builder will signal us when Preview data is ready
            PreviewReadyEvent = threading.Event() 
            PreviewReadyEvent.clear()

            # create a build command 
            pcmd = copy.copy(BldCmd)
            pcmd['Command'] = BuildPreview
            pcmd['Params']  = qs['postalcode'][0]
            pcmd['Event']   = PreviewReadyEvent
            
            logging.info("MAIN:GET:lineupPreview: Queuing BuildPreview for %s.  Waiting for Result...", qs['postalcode'][0])
            BuilderQueue.put( pcmd )
            
            # this command could take a long time if the Preview has to be built by the server.
            # it might not even complete on teh SD server in time, the user just has to retry.
            # The PiGS webserver is busy during this wait and cant respond to other commands
            
            PreviewReady = PreviewReadyEvent.wait(60) # wait here for up to 60 seconds
 
            if PreviewReady :
                logging.info("MAIN:GET:lineupPreview: Result is ready.")
                # Get the results
                d      = pcmd['Result']   # get reference to the result dictionary 
                code   = d['code']
                lujson = d['lujson']
                luname = d['luname']
            else :
                logging.warning("MAIN:GET:lineupPreview: Timeout waiting for Preview Result.")
                code = 503
                lujson = None
                luname = "Timeout"
            
            # Call Lineup Preview command which makes 2 SD calls to get the data
            # Request the lineup preview, we can get a 200 with an error block with a code in it
            # There is also a differnt type of error response with a non-zero code.
            # we dont know which we will get
                       
            if code == 200 :
                # check for error code in json data, it might be a 2107 Queued not ready error
                # check for an error code, the preview response doesnt have an error code
                
                resultcode = 0
                
                # SD returns List[] for preview or dict {} for error / status response
                if type(lujson) is dict :
                    resultcode = lujson.get('code',0)
               
                if resultcode == 0 :
                    
                    # The returned lujson IS the LineupDef ChannelList, but the ChannelDef's
                    # dont include the favorite key
                    # Create the 'favorite' key and default to True for all channels
                    for c in lujson :
                        c['favorite']=True
                        
                    # create the complete response message to send to the UI
                    msg = copy.copy(LineupDef)    # Start with a copy of my custom response
                    msg['LineupName']  = luname;  # Add the SD lineup name (for display on the UI side)
                    msg['ChannelList'] = lujson;  # append the lineup list of ChannelDefs with Fav key
                    msgj = json.dumps(msg)        # dict to json
                else :
                    logging.info("MAIN:PREVIEW: return code %d", resultcode)                
                    msg = {'code':2107, 'message':"The lineup is being generated, please retry."}
                    msgj = json.dumps(msg)
            else :
                logging.info("MAIN:PREVIEW: Error %d %s %s", code, lujson, luname)                

                
            if lujson == None :
                msg = {'code':code, 'message': luname }
                msgj = json.dumps(msg)

            # forward the SD response code to the UI
            self.send_response(code)
            self.SendCommonHeaders() 
            self.send_header("Content-Type", "application/json")
            self.send_header("Connection","Close")
        
            # gzip encocded response
            lugz = gzip.compress( msgj.encode() )  #to utf-8
            
            self.send_header("Content-Encoding", "gzip")
            self.send_header("Content-Length", str( len(lugz) )) 
            self.end_headers()

            logging.debug("  GET lineupPreview: lulen(%d)  gzlen(%d)\n", len(msgj), len(lugz) )

            self.wfile.write( lugz )
            return
             
        else: # we have not handled this GET.  Let SimpleHTTP handle file requests
            logging.info("MAIN:GET: File (%s) from %s %s", o.path, self.client_address, lts)
            # if it sees '/', it looks for and returns index.htm (or index.html)
            # 2 ways to call the Parent... name the parent, or super the child...
                 #super(PiGS_HTTPRequestHandler, self).do_GET()
           
            self._close_connection = True  # Add a 'Connection : Close' header 
            
            SimpleHTTPRequestHandler.do_GET(self)

            logging.debug("MAIN:do_GET Path: %s\nHeaders:\n%sRequestLine:%s\nMAIN:do_GET:Done.", str(self.path), str(self.headers), self.requestline)
 
    def do_POST(self):  # this receives URLs like POST /UIsettings or
        
        global dPiGS
        
        logging.debug("MAIN:POST: Path: %s Headers: %s", str(self.path), str(self.headers))

        # Break url into scheme, netloc, path, params, query,fraqment
        # then parse the query parameters into a dictionary
        o=urlparse(self.path)
        qs=parse_qs(o.query)

        # get and decode the body json
        content_length = int(self.headers['Content-Length'])
        body = self.rfile.read(content_length)
        
        # generic decode for error return/debug, may be re-decoded for specific command
        bj = json.loads( body.decode() )
        
        # Now process the individual Post commands
        
        if ("UIsettings" in self.path) :  # POST settings
           
            # Settings returned from the UI
            # ON Save Settings, lineup or icons may have changed we need to abort any build in progress
            # If a schedule build was in progress, UI Save would kill the build b/c the token would change.
            
            #Todo disable UI button until this command returns???
            
            if not SDResourceLock.acquire(False) :  # if we get the lock, we can do the settings
                
                logging.warning("MAIN:POST:UIsettings: PiGS is busy building schedule.  Requesting Abort...")
                
                BuildAbortEvent.set()   #signal Linear Schedule Builder to abort
        
                time.sleep(1)    # give thread time to end
                
                waitAttempts = 5 #seconds
                while waitAttempts and not SDResourceLock.acquire(False) :
                    time.sleep(1)
                    waitAttempts -= 1

                BuildAbortEvent.clear() # ensure flag is clear after abort
                
                if waitAttempts == 0 :
                    logging.warning("MAIN:POST:UIsettings: Timeout on Abort. PiGS could not save Settings!")
                    self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule")
                    #We never got the lock so I dont have to release it, but better to be safe
                    #if SDResourceLock.locked() :
                    #    SDResourceLock.release()
                    return
                
            # We have the Lock, we can make SD calls, or Queue commands
            SDResourceLock.release()  # Build init process in DoTimerBuild will get it's own lock
            
            # If this function returns before Queuing the Service build, must do a SDResourceLock.release()
            
            #--------------------------------------------------------------
            # Get the settings from the UI message block, json to dict conversion
            PigsUIreturned = json.loads( body.decode() )
            logging.info("MAIN:POST:UIsettings len: %s", content_length )
            self.send_response(200)
            self.send_header("Connection","Close")
            self.end_headers()

            #--------------------------------------------------------------
            # I could just replace the whole dPiGS dict with the received UI json.
            #   but maybe there are some keys we want to preserve on the server side
            # I do this with keys in pigsPreserveSettings.  See notes where dict is defined.
            # Copy some keys from current pigs settings to received settings
            # PigsPreserveSettings is just a list of keys only
            # Update the returned keys with current keys in case a few special keys changed while UI opened
            
            for key in PigsPreserveSettings :  # Use local definition structure
               PigsUIreturned[key] = dPiGS[key]
            
            # If icon type changes or SD theme changes, we will update Cached Icons
            #IconChanged  = ( PigsUIreturned['IconType']    != dPiGS['IconType'] )
            #IconChanged |= ( PigsUIreturned['SdIconTheme'] != dPiGS['SdIconTheme'] )
            #logging.info("MAIN: Icon Change State: %s", IconChanged )

            # Save change flag in config file so Service build can update icons.
            # after service build, should clear this flag and save settings
            
            #PigsUIreturned['UpdateIconCache'] = IconChanged  # service build will update Icon Cache
            

            # Do I need to sanity check the data ???
            # should I check the data before saving?
            # should I make the changes on the Temp settings like do the SD stuff first?
            # Just save what the User configured, if there is a fault at least we saved users work
            # when they restart, refresh UI and click save settings all should be well

            if PigsUIreturned['LoggingLevel'] != dPiGS['LoggingLevel'] :
                logging.info("MAIN: Logging level changed from %s to %s", dPiGS['LoggingLevel'], PigsUIreturned['LoggingLevel'] )
                
            
            #--------------------------------------------------------------
            # Recursively copy the UI data into our main config dict
            dPiGS = copy.deepcopy(PigsUIreturned)

            logging.info("MAIN: Saving DVR_States...")
            dPiGS['DVR_States'] = LS_GetDictOfDVRs()

            # This saves 5 backps as <name>N.json, where N = 1..5
            SaveSettingsFile(dPiGS, CONFIG_FILE)

            #--------------------------------------------------------------
            #https://docs.python.org/3/library/logging.html
            #See the heading 'Module Level Functions'
            #Get for a reference to the default root level logger
            logger = logging.getLogger()
            logger.setLevel( dPiGS['LoggingLevel'] )
            
            
            #--------------------------------------------------------------
            # Process PiGS DNS Server changes
            #global IPaddr
            
            # Starting DNS
            if dPiGS['PigsDoesDNS'] and not PigsDnsAlive() :  # do we need to start DNS?
                if dPiGS['PigsDelayDNS'] :
                    logging.info("MAIN:UIsettings: DNS will be started after Schedule is built...")
                else :
                    logging.info("MAIN:UIsettings: Starting DNS...")
                    PigsDnsStart(dPiGS['Server_IP'], dPiGS['PiGS_IP'], dPiGS['DnsHosts'])
            
            # Stopping DNS
            if not dPiGS['PigsDoesDNS'] and PigsDnsAlive() :
                logging.info("MAIN:UIsettings: Stopping DNS - waiting for thread to end...")
                dnsThread = PigsDnsStop()
                dnsThread.join(2.0)  # wait up to 2 seconds for DNS to Stop (should stop in less than 1 sec)
                logging.info("MAIN:UIsettings: DNS Stopped... Alive=%s", dnsThread.is_alive())

            #--------------------------------------------------------------
            # Process Lineup changes on Schedules direct
            # We need to make SD match what user configured by deleting and adding lineups
            # After User Saves settings we do a complete re-initialization
            
            logging.info("MAIN:POST:UIsettings: Clearing ReadyState to Re-initialize...")
            ResetReadyState()
            
            bts = SetDvrBuildTimer(1.5)  # set timer for 1.5 second to allow user to switch window
            
#            logging.info("MAIN:POST:UIsettings: Build time scheduled for %s", bts)
            # Done processing UI settings

        # User hits 'Verify Login' to create a sha1 password hash, Test login by getting SD token
        if ("sdlogin" in self.path) :

            sha1pwd = GetSha1Password(bj['password'])
            logging.info("MAIN:POST:SDLogin Attempting GetToken")

            if not SDResourceLock.acquire(False) :  # if we get the lock, we can do the sdlogin
                logging.warning("MAIN:POST:SDLogin Can't login.  PiGS is busy rebuilding schedule!")
                self.send_error(ServerErrorCode, message="PiGS is busy.  Please Retry.")
                return  # didnt get the lock, dont have to release anything

            # Builder will signal us when Login data is ready
            LoginReadyEvent = threading.Event() 
            LoginReadyEvent.clear()

            # create a build command
            lgcmd = copy.copy(BldCmd)
            lgcmd['Command'] = BuildLogin
            lgcmd['Params']  = {'un':bj['username'],'pw':sha1pwd }
            lgcmd['Event']   = LoginReadyEvent
            
            logging.info("MAIN:POST:SDLogin: Queuing BuildLogin & waiting for Result...")
            BuilderQueue.put( lgcmd )

            # The PiGS webserver is busy during this wait and cant respond to other commands
            LoginReady = LoginReadyEvent.wait(10) # wait here for up to 10 seconds
            
            if LoginReady :
                logging.info("MAIN:POST:SDLogin: BuildLogin is ready.")
                d = lgcmd['Result']   # get reference to the result dictionary 
                respcode  = d['code']
                rsj       = d['errj']
            else :
                logging.warning("MAIN:POST:SDLogin: Timeout waiting for BuildLogin Result.")
                respcode = 503
                rsj  = None
           
            # if it works, the username/sha1 combo is good and we can save in the config file
            if respcode == 200 :
                # update the config dictionary
                logging.info("MAIN:POST:SDLogin: SD Login Successful!")
                dPiGS['sdusername'] = bj['username']
                dPiGS['sdsha1pwd']  = sha1pwd
                SaveSettingsFile(dPiGS, CONFIG_FILE)     # Save setting file to disk
            else :
                logging.warning("MAIN:POST:SDLogin: code(%s) %s", respcode, rsj)

            self.send_response(respcode)
            self.send_header("Connection","Close")
            self.end_headers()

    def _do_unsupported(self):
        logging.warning("MAIN:Unsupported Command: do_ %s", self.command)
        logging.warning("MAIN: Path: %s\nHeaders:%s", str(self.path), str(self.headers))
        self.send_response(400)
        self.send_header("Connection","Close")
        self.end_headers()
        response = BytesIO()
        response.write(b'This is an unsupported request. ')
        
    #https://stackoverflow.com/questions/43156023/what-is-http-host-header
    #https://bugs.chromium.org/p/chromium/issues/detail?id=521009
    #https://superuser.com/questions/1398098/how-can-i-get-google-chrome-to-go-to-an-internal-page-without-performing-an-exte
    #https://docs.microsoft.com/en-us/troubleshoot/browsers/one-word-intranet-sites
    def do_HEAD(self):
        logging.info("MAIN:do_Head. Passing to parent.")
        self._close_connection = True  # Add a 'Connection : Close' header 
        SimpleHTTPRequestHandler.do_HEAD(self)
        logging.debug("MAIN:do_HEAD Path: %s\nHeaders:\n%sRequestLine:%s\nMAIN:do_HEAD:Done.", str(self.path), str(self.headers), self.requestline)
    def do_DELETE(self):
        self._do_unsupported()
    def do_PATCH(self):
        self._do_unsupported()
    def do_OPTIONS(self):
        self._do_unsupported()
    def do_PUT(self):
        self._do_unsupported()

#---------------------------------------------------------------------------------------
# Function to get my IP for display and wait for network to boot
# When starting PiGS at boot, the network is not always ready and
# returns: [Errno 101] Network is unreachable
# We will sleep a few times while waiting for network to boot
# 30 seconds max wait = 30 attempts, 1 sec wait each time
# 
# None is the return value if cant get IP

import socket

def WaitForNetwork(NetworkAttempts = 30):

    for i in range(NetworkAttempts, -1,-1) : # start, stop 1 before, step
    
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            IP = s.getsockname()[0]
        except Exception as err:
            logging.debug("  get_ip = %s", err )
            IP = None
        finally:
            s.close()
    
        if IP :        # If we get real IP, break out of wait loop
            break
        
        logging.debug("  Waiting for network %d", i )
        time.sleep(1)      # take a nap while the system boots the Network 

    #time.sleep(1)      # take a nap while the system boots the Network 

    return IP



        
##########################################################################
# OK, here we go... initialize and start things going

#====================================================================================
# Check for command line args.
# Pass in Port and maybe logging level to use config for logging level
# On Pi/Linux Port 80 is 'protected' so PiGS must be started with the sudo command
#  to allow the app to 'bind' to port 80.  No error checking is done here
#   sudo python3 pigs.py 80

#https://docs.python.org/3/howto/argparse.html
import argparse
cmdparser = argparse.ArgumentParser()
#group = cmdparser.add_mutually_exclusive_group(required=False)
cmdparser.add_argument("port", help="Specify http port other than default port 80", nargs='?', type=int, default=80)
cmdparser.add_argument("-serverIP", "-s",  help="Server_IP override")
cmdparser.add_argument("-force", "-f",  help="Force port to override IDE_Port for debug", action="store_true")
cmdparser.add_argument("-loglevel","-l",help="Set logging level. L=10 Debug, L=20 Info L=30 Warning", type=int)
cmdparser.add_argument("-ignore","-i",  help="Ignore 'No Ethernet Connection' error and continue startup.", action="store_true")
cmdparser.add_argument("-waitnet","-w", help="Wait 'waitnet' seconds for Network to become available.", type=int)
cmdparser.add_argument("-noload","-n",  help="Do NOT load Program Schedule or Cache on start", action="store_true")
cmdparser.add_argument("-api","-a",     help="Set another SD server or API")

#cmdparser.add_argument("-waitnet","-n", help="Set Network wait time", nargs='?', type=int, default=60)
args = cmdparser.parse_args()

#Group Mutually exclusive doesnt work for me.  doing it myself
#usage: pigs.py [-h] [-p P] [-loglevel LOGLEVEL] [port]
#pigs.py: error: one of the arguments port -p is required
#if args.p ('p' in args) and 'port' in args :
#    cmdparser.print_help()
#    cmdparser.print_usage()   # print brief command line format
#    cmdparser.exit(status=0, message='pigs.py: error: Use 1 or none of the arguments port -p, but not both\n')
#    print('port in args')

print("\nStarting...")
print("Command line args:", args)

# Some print()'s are used so something appears when logging to a file

#https://docs.python.org/3/howto/logging.html
#https://docs.python.org/3/library/logging.html#levels
#https://docs.python.org/3/library/logging.html#logging.basicConfig
#https://docs.python.org/3/library/logging.html#module-level-functions
#https://docs.python.org/3/library/logging.html
#See the heading Module Level Functions

#Level      When it’s used
#NOTSET      0
#DEBUG      10 Detailed information, typically of interest only when diagnosing problems.
#INFO       20 Confirmation that things are working as expected.
#WARNING    30 An indication that something unexpected happened, or indicative of some problem in the near future (e.g. ‘disk space low’). The software is still working as expected.
#ERROR      40 Due to a more serious problem, the software has not been able to perform some function.
#CRITICAL   50 A serious error, indicating that the program itself may be unable to continue running.
#The default level is WARNING

# Open CONFIG_FILE JSON file, load settings into a Dict
# Cant do any logging before the config file since we get the log file name from there
#https://stackoverflow.com/questions/9823936/python-how-do-i-know-what-type-of-exception-occurred
#https://docs.python.org/2/library/exceptions.html#exception-hierarchy

bDelayedLogging = False

print("PiGS Working Directory on launch:" , os.getcwd())

# V0.99 Change PiGS working directory to where program started
head, tail = os.path.split( os.path.realpath(__file__) )
os.chdir(head)

#====================================================================================
# Test for icons folders, create if not found.
# the mode params do not seem to work for mkdir

if not os.path.exists('myicons'):
    os.mkdir('myicons')  #, mode=0o644, *, dir_fd=None)
    os.chmod('myicons', mode=0o777)
if not os.path.exists('cmicons'):
    os.mkdir('cmicons')  #, mode=0o644, *, dir_fd=None)
    os.chmod('cmicons', mode=0o777)
if not os.path.exists('sdicons'):
    os.mkdir('sdicons')  #, mode=0o644, *, dir_fd=None)
    os.chmod('sdicons', mode=0o777)


#====================================================================================

from ut import PigsPath

cfgf = PigsPath(CONFIG_FILE)
    
try:
    f = open(cfgf)          # 'pigs.json'
    dPiGS = json.load(f)   # returns JSON object as a dictionary 
    f.close()              # Closing file

# JSONDecodeError exception isnt in Python 3.4 and earlier, dont really need it
#except json.JSONDecodeError:  
#    print("Error decoding config file.", CONFIG_FILE)
#    sys.exit(5)

except FileNotFoundError:
    print("Config file %s not found, creating...", cfgf)
    dPiGS = copy.copy(PigsConfigDict)
    with open(cfgf, "w") as outfile: 
        json.dump(dPiGS, outfile, indent=2)
        outfile.close()
        
    # need to delay logging this b/c logging not enabled yet - log file name in config file
    bDelayedLogging = True
    # We can contine, but this is pretty serious unless it is the very first run
    # I guess its not too serious...  if its gone, you just have to reconfigure.
    
except Exception as ex:
    template = "An exception of type {0} occurred. Arguments:\n{1!r}"
    message = template.format(type(ex).__name__, ex.args)

# update the version number in the config file in case the app version has changed.
dPiGS['PiGS_Version'] = PIGS_VERSION


#====================================================================================
# Auto-Upgrade pigs.json with missing keys when there are version upgrades

UpgradeConfigFile()


#====================================================================================
# Init Email alerts with Smtp login and destination emails

EM_Initialize(dPiGS['SmtpUser'], dPiGS['SmtpPass'], dPiGS['AlertEmails'], dPiGS['PiGS_IP'])

#====================================================================================
# Now have the configuration file.  http port over-ride?

port = args.port  # default 80 or what user specifies on cmd line

if not args.force and dPiGS["IDE_Port"] > 0 :
    port = dPiGS["IDE_Port"]
    print("IDE_Port Over-ride: %s", port)

#====================================================================================
# Get logging settings from config file, then Enable Logging

logFile  = dPiGS.get('LoggingFile',  PIGS_LOG_FILE)  # if logFile='' it will log to console
logLevel = dPiGS.get('LoggingLevel', logging.INFO)
logf     = PigsPath(logFile)

# Command line over-ride for this sesson only.  UI can change
if args.loglevel :
    logLevel = args.loglevel

# if running from the Dev environment, dont log to file so as not to
#   conflict with the port 80 process if it is running.
# running both at the same time could have side effects with the config file also

if port != 80 :     
    #logf = ""      # if not port 80, remove the filename to log to console only
    # if not port 80 append port to file name so diff pigs instances have their own logs
    logf = PigsPath("pigslog"+str(port)+".txt")
    
# replaced this code with the RotatingFileHandler
#https://stackoverflow.com/questions/24505145/how-to-limit-log-file-size-in-python
#https://gist.github.com/amamaenko/653743
# initialize logging before making calls to any modules that use logging or weird(nothing) thing happen

logger = logging.getLogger()

# add a Rotating File Handler to keep log file size in check 

LogMaxBytes = dPiGS.get("LogFileBytes", MAX_LOGFILE_BYTES)
LogBackups  = dPiGS.get("LogFileCount", LOGFILE_BACKUPS)

if logf :
    try :
        handler = logging.handlers.RotatingFileHandler(logf, maxBytes=LogMaxBytes, backupCount=LogBackups)
        logger.addHandler(handler)
        logger.setLevel( logLevel )
    except PermissionError:
        print('\nError - could not open log file. Insufficient permissions.')
        print("   On a Raspberry Pi or Linux you may need to start as 'sudo python3 pigs.py' (run as root)\n")
        exit(1)
else :
    logger = logging.basicConfig(filename=logf,level=logLevel)

# Also send Info logs to the console
# docs.python.org/2.4/lib/multiple-destinations.html
console = logging.StreamHandler()
console.setLevel(logging.INFO)
#console.setLevel(logLevel) #

#formatter =logging.Formatter('%(name-12s: %levelname)-8s %(message)s')
#console.setformatter(formatter)

#if port != 8000 :     # again, if command line start, log to console also
#    logging.getLogger("").addHandler(console)
logging.getLogger("").addHandler(console)

loggingto = 'file: ' + logf if logf else 'console ONLY <<<<<<<<<<<<<<<<<<<<<<<<<'
print("Logging to ", loggingto)
print("Logging Level is", logLevel)

# set global flag to over-ride annoying HTTP.server log.info prints
bLogHttp = dPiGS.get('LogHttpReq')

PigsStartTime = datetime.utcnow()
lst, lz = LocalTimeStr( PigsStartTime, WeekDay=False )
logging.critical("MAIN: PiGS started at %s", lst )
logging.critical("MAIN: Logging Level is %s", logLevel)
logging.critical("MAIN: PiGS Version %s", dPiGS['PiGS_Version'] )
logging.critical("MAIN: Current Working Directory %s" , os.getcwd())

# Had to delay logging this because the logfile did not yet exist when error occurred
if bDelayedLogging :
    logging.warning("MAIN: Config file %s not found, created!", cfgf)

#====================================================================================
#
logging.info("MAIN: Platform Info")
logging.info("  MaxSize(%s), Is64-bit(%s)", hex(sys.maxsize), sys.maxsize > 2**32 )

try :
    import platform
    logging.info("  Achritec: %s", platform.architecture() )
    logging.info("  System  : %s", platform.system() )
    logging.info("  Machine : %s", platform.machine() )
    logging.info("  Platform: %s", platform.platform() )
    logging.info("  Python I: %s", platform.python_implementation() )
    logging.info("  Python V: %s", platform.python_version() )
except :
    logging.warning("  Platform library not available")
    
#Python Version as a Float for comparisons
PythonVer = VersionStringToFloat(platform.python_version())

if dPiGS["ThreadedHTTP"] and PythonVer < 3.7 :
    print("MAIN:WARNING: PiGS 1.24 and up requires a minimum of Python 3.7!" )
    print("MAIN:WARNING: Update Python with 'sudo apt install python3.7'\n" )
    logging.warning("MAIN:WARNING: PiGS 1.24 and up requires a minimum of Python 3.7!" )
    logging.warning("MAIN:WARNING: Update Python with 'sudo apt install python3.7'\n" )
    # Exit pigs, no sense in running pigs if we dont have the required Python Version
    sys.exit(5)

logging.info("MAIN: http.server Info")
try :
    logging.info("    server_version: %s", PiGS_HTTPRequestHandler.server_version )
    logging.info("       sys_version: %s", PiGS_HTTPRequestHandler.sys_version  )
    logging.info("  protocol_version: %s", PiGS_HTTPRequestHandler.protocol_version)
except :
    logging.warning("MAIN: Error getting http.server version info.")


from ls import LSLoadProgramsFromDisk  #, utSaveObjectToDisk 

if not args.noload :  # if its not noload, then it means load
    logging.info('MAIN: Loading Program Schedule from Disk...')
    LSLoadProgramsFromDisk()
else : 
    logging.info('MAIN: -n Skipping Program Schedule Load from Disk...')

#====================================================================================
# Wait for Network Ready and Print the Pigs IP address.

waitnet = dPiGS['NetworkWait']

if args.waitnet :
    waitnet = args.waitnet

logging.info("MAIN: Checking Network.  Waiting up to %ss for Network ready...", waitnet )
IPaddr = WaitForNetwork(waitnet)
logging.info("MAIN: PiGS Discovered IP = %s", IPaddr )

if not IPaddr :  # PiGS did not detect an IP address - no Internet Connection?
    logging.warning("MAIN: Computer does not appear to have an Ethernet connection." )
    
    if not args.ignore :  # PiGS needs a network.  can override with -ignore
        #cmdparser.print_help()  # prints full help
        cmdparser.error('Fix network connection and retry.')    

dPiGS['PiGS_IP'] = IPaddr  # save PiGS_IP in config to display on UI

#====================================================================================
# manual over-ride of listen IP address for port 80 & port 53

hostIP = dPiGS["Server_IP"]    # defualt "0.0.0.0",

if args.serverIP :  # Cmd line override of ip
    hostIP = args.serverIP 

if hostIP != '0.0.0.0' :
    
    dPiGS['PiGS_IP'] = hostIP  # Display on UI
    
    # Print a warning if my discovered IP does not match the set IP
    # This can happen if a device has multiple Ethernet interfaces like a Wired, Wifi, or VPN
    # This message lets user know which is bound
    if hostIP != IPaddr :
        logging.warning("MAIN: *** Server_IP(%s) does not match discovered IP(%s) *** ", dPiGS['Server_IP'],IPaddr )

#====================================================================================
# Load state of Past DVRs to provide continuity
# ToDo should provide a method to clean this out if a DVR has not been seen for a while

if 'DVR_States' in dPiGS :
    logging.info("MAIN: Loading DVR_States...")
    LS_SetDictOfDVRs( dPiGS['DVR_States'] )
else :
    logging.info("MAIN: No DVR_States to load")

#====================================================================================
# version from http://pigs.goihl.com/zips/version.txt to print update msg 1.36

CheckForPigsUpdate()
            
#====================================================================================
# Initialize the Alt server name to reach CM servers for CM/Rovi Icons
# CMInitAltServer()  now done in builder

sdapi = dPiGS["sdApiVersion"]

if args.api :   # check for cmd line over-ride
    sdapi = args.api 
#if dPiGS["sdApiVersion"] :
logging.info("MAIN: Setting SD API Version to '%s'", sdapi)
SDSetAPIVersion(sdapi)  # set or display SD API 

#https://ywnd0zbafe.execute-api.us-east-1.amazonaws.com/20141201

#====================================================================================
# Start the Builder Thread
logging.info("MAIN: Starting BuilderThread")
BuilderThread = threading.Thread(target=Builder, name='Builder', daemon=True )
#BuilderThread = threading.Thread(target=Builder, name='Builder', daemon=True, args=(dPiGS, PiGS_SERVICEID, BuildAbortEvent,) )
BuilderThread.start()
logging.info("MAIN: BuilderThread.isalive(): %s", BuilderThread.is_alive())

#====================================================================================
# Start the DNS Server Thread
# PigsDelayDNS causes DNS to start after Sched build completes
# This ensures that a DVR wont ask for schedule data before it is ready

if dPiGS["PigsDoesDNS"] :
    if dPiGS["PigsDelayDNS"]:
        logging.info("MAIN: Delaying DNS Start until Schedule is Built.")
    else :
        PigsDnsStart(dPiGS['Server_IP'], dPiGS['PiGS_IP'], dPiGS['DnsHosts'])

#====================================================================================
# Build/Load the Service details so they are available to the Schedule builder.
# On very first run there is no user name so skip Initialization steps.
# It will get built when user does a Save Settings in the UI
# Also, if we never got an IP address, dont attempt to build the Service or Schedule lists

if IPaddr :
    
    if dPiGS['sdusername'] :

        ResetReadyState()
        
        # Build will do an Init to:
        # 0) Set the CM Alt Server url
        # 1) Verify SD account lineups match what user configured by deleting and adding lineups
        # 2) Build the Service list (Lineup list)
        # 3) Build the schedule, then enable DNS
        
        # Let webserver start - Schedule the build a number of seconds from now
        if dPiGS['FirstBuildDelay'] :
            fbt = SetDvrBuildTimer( dPiGS['FirstBuildDelay'] )
            logging.info("MAIN: Initial Schedule Build in %s secs", dPiGS['FirstBuildDelay'])
        else :
            logging.info("MAIN: Initial Schedule Build dsiabled.")
        
    else :
        logging.warning("MAIN: No SD username, can't Build Schedule! Configure at %s", 'http://'+IPaddr)
else :
    logging.warning("MAIN: PiGS could not determine IP address!  No Internet?")

    if dPiGS['FirstBuildDelay'] :
        SetDvrBuildTimer( dPiGS['FirstBuildDelay'] )  # set build timer to retry later
    else :
        logging.info("MAIN: Initial Schedule Build dsiabled.")


#====================================================================================
# Construct and start the webserver

# References for HttpServer and SocketServer objects
#https://github.com/python/cpython/blob/3.10/Lib/socketserver.py
#https://github.com/python/cpython/blob/3.8/Lib/http/server.py
#https://github.com/python/cpython/blob/3.10/Lib/socketserver.py

if dPiGS["ThreadedHTTP"] :
    from http.server import ThreadingHTTPServer
else :
    from http.server import HTTPServer


class PiGS_HTTPServer( HTTPServer ):

    # testing of httpserver,socketserver for 1.24,1.30 lockups
    
    # override/extend TCPServer in socketserver
    def XXget_request(self):
        print('getting...')
        rq = self.socket.accept()
        print('got ', rq)
        return rq
    
    # override/extend BaseServer in socketserver
    def XXshutdown_request(self, request):
        print('closing ', request)
        self.close_request(request)
        
    def service_actions(self) :  # override to see if socketserver is polling (it does)
        
        #how to access dunder private class variables in a base class
        #print( self.server_address )
        #print( self.RequestHandlerClass )
        #print( self._BaseServer__shutdown_request )
        #print( self._BaseServer__is_shut_down.is_set() )
        
        #print( 'service_actions() %s %s', httpd._BaseServer__shutdown_request, httpd._BaseServer__is_shut_down.is_set() )

        HTTPServer.service_actions(self)  # call parent  


#====================================================================================
# create/bind instance of a web server object with my custom handler
Handler = PiGS_HTTPRequestHandler

RestartServerRequest = True  # initial value  

#httpd = None  # webserver handle
            
gWebServerRestartTime,tz = LocalTimeStr( PigsStartTime, True )

# Restart server loop
while RestartServerRequest :

    RestartServerRequest = False   # Only run once unless requested to restart
    
    try :
        if dPiGS["ThreadedHTTP"] :
            logging.info('MAIN: ThreadingHTTPServer binding to IP=%s:%s',hostIP,port)
            
            # Change to threading server in 1.24 to handle Chrome favicon.ico requests
            httpd   = ThreadingHTTPServer((hostIP, port), Handler)
            # This class is identical to HTTPServer but uses threads to handle requests by using
            # the ThreadingMixIn. This is useful to handle web browsers pre-opening sockets, 
            # on which HTTPServer would wait indefinitely
        else :
            logging.info('MAIN: Non-threaded HTTPServer binding to IP=%s:%s',hostIP, port)
            #httpd   = HTTPServer((hostIP, port), Handler)
            httpd   = PiGS_HTTPServer((hostIP, port), Handler)
        
    except:
        logging.warning("MAIN: Could not bind Webserver to port %s. Another PiGS or Process runninng?", port )
        logging.warning("  On Linux did you start with root privs?  'sudo python3 pigs.py'?" )
        logging.warning("  On Linux try 'sudo lsof -i :80' to get Process ID #" )
        logging.warning("  On Windows, check firewall.  Allow PiGS full access on ports 53 & 80" )
        
        # Exit pigs, no sense in running pigs if we cant get the requested port, usually 80
        sys.exit(5)

    logging.info("MAIN: Waiting for DVR+ requests on Port %d!\n",port)

    WebserverResetEvent.set()  # Signal initiator Reset has occurred

    # Start webserver... server_forever() does not end.
    # This is the 'main' thread... But allow for clean exit from keyboard with Ctrl-C
    # or from code (another thread) httpd.shutdown() can be used to break
    
    try:
        httpd.serve_forever(poll_interval=0.5)
    except KeyboardInterrupt:
        logging.info("MAIN: Ctrl-C Keyboard Interrupt. Webserver stopped!\n")
        pass

    httpd.server_close()
    
    if RestartServerRequest :
        logging.info("\n******** MAIN: Restarting WebServer ********")
        gWebServerRestartTime, tz = LocalTimeStr( datetime.utcnow(), True )

        alertmsg = "PiGS WebServer not Responding!  Restarting...\r\n ." 
        SendEmailAlert(alertmsg, subjectMsg=': Webserver')
    
    # End of while RestartServerRequest : LOOP

# To exit, code will fall thru to here on a Ctrl-C

#====================================================================================
#Shut down Builder thread

logging.critical('\nMAIN: Shutting down threads and saving environment.')

if not SDResourceLock.acquire(False) :  # check if thread is busy
    logging.warning("MAIN: PiGS is busy building schedule.  Requesting Abort...")
    BuildAbortEvent.set()   #signal BuildLinearSchedule() to abort
    waitAttempts = 50 # 5 seconds
    while waitAttempts and not SDResourceLock.acquire(False) :
        time.sleep(0.1)
        waitAttempts -= 1
    BuildAbortEvent.clear() # ensure flag is clear after abort

# We have the lock(?), tell the Builder thread to terminate so we can exit nicely
logging.info("MAIN: Queuing BuildTerminate...")
BuilderQueue.put( {'Command':BuildTerminate} )
BuilderThread.join(2.0)  # give it a chance to end

# Shut down DNS Server Thread if alive
if PigsDnsAlive() :
    logging.info("MAIN: Stopping DNS.  Waiting for thread to end...")
    dnsThread = PigsDnsStop()
    dnsThread.join(2.0)  # wait up to 2 seconds for DNS to Stop (should stop in less than 1 sec)

# Save Settings on exit to preserve build times 
logging.info("MAIN: Saving DVR_States...")
dPiGS['DVR_States'] = LS_GetDictOfDVRs()

logging.critical('MAIN: Saving User Settings...')
SaveSettingsFile(dPiGS, cfgf)

#if dPiGS['SaveSchedule'] > 0:
#    logging.critical('MAIN: Saving Program Schedule to Disk...')
#    LSSaveProgramsToDisk( SaveLevel = dPiGS['SaveSchedule'] )

lst, lz = LocalTimeStr( datetime.utcnow(), WeekDay=False )
logging.critical("MAIN: PiGS done at %s.\n", lst)
