# sd.py
# functions to interface with or manipulate Schedule Direct
# Copyright June 2020, Timothy Goihl
#
# This module was developed for the '2014' SD API
# https://github.com/SchedulesDirect/JSON-Service/wiki/API-20141201
#

PIGS_USER_AGENT = 'PiGS/V1.0'
import json
import requests
import logging
import datetime
import sys
import time

#https://dateutil.readthedocs.io/en/stable/

# Changed to https in 1.25
SDapi   = "https://json.schedulesdirect.org/20141201"

# Test Server for Robert K at SD - Python Lamda function testing 
#SDapi   = "https://ywnd0zbafe.execute-api.us-east-1.amazonaws.com/20141201"

# To use these the full url is eg:  SDapi+PostTokenURL
PostTokenURL    = "/token"
GetStatusURL    = "/status"
GetHeadendsURL  = "/headends"  #/country=USA&postalcode=60030

GetUserLineupsURL     = "/lineups"
GetPreviewLineupsURL  = "/lineups/preview/"  #/lineupcode
PutUserLineupsURL     = "/lineups/"  #/lineup code
DelUserLineupURL      = "/lineups/"  #lineup code
GetLineupsMapURL      = "/lineups/" #+lineup code

PostSchedulesURL      = "/schedules"
PostProgramDetailsURL = "/programs"

# These API's are not currently used
AvailableGetURL       = "/available"
SchedulesMD5PostURL   = "/schedules/md5"
ProgramGenericPostURL = "/metadata/description"

def SDSetAPIVersion(NewServerApi=None) :
    
    global SDapi
   
    if NewServerApi : SDapi = NewServerApi

    logging.debug("SD:SDSetAPIVersion: SD TokenURL = %s", SDapi+PostTokenURL)

    return SDapi
    
#Useful Test website that echos the Post
# response = requests.post("http://httpbin.org/post", json={"username":"sdusername","password":'a sha1 hash pw goes here'})

#https://requests.readthedocs.io/en/master/user/quickstart/
#>>> payload = {'key1': 'value1', 'key2': 'value2'}
#>>> r = requests.get('https://httpbin.org/get', params=payload)

#Programming references
#https://pynative.com/python-post-json-using-requests-library/
#https://requests.readthedocs.io/en/master/user/quickstart/#json-response-content
#https://requests.readthedocs.io/en/master/user/advanced/#request-and-response-objects

#Chunked encoding
#https://requests.readthedocs.io/en/master/user/advanced/#chunk-encoded-requests

#==============================================================================
# Token Handling
#
# Utility function to Generate Sha1 password encoding required by SD
# Python shell usage:
#   from sd import GetSha1Password
#   GetSha1Password('mySDpassword')

import hashlib

def GetSha1Password(pwd) :
    #str = "mypassword"
    sha1pw = hashlib.sha1( pwd.encode() ).hexdigest()
    return sha1pw

# global gToken for use by other SD APIs when they need it

gToken = None

# Returns http error status 200,400, etc and the json so we can get error details
# If a token is passed in, just set that instead of asking for one
# this is for devlopent so that if using Postman on PC we are using the same token
# the token is saved in the config file and has to be generated manually

#REQ_TIMEOUT = 10.001   # one timeout uses same for both
REQ_TIMEOUT = (30.0, 90.0)  # Connection Timeout=30, Read Timeout = 90

SD_RETRY_SECS  = 2  # Initial retry secs for SD api's, then x2 for each

def SDGetToken(username, sha1pwd, token=None ) :

    global gToken

    # developement override...  use a fixed token to allow for concurrent access with PC Postman
    # Generate a token in Postman, then set the Token in pigs.json
    # The token should be good for about 24 hrs???
    
    # If a token is passed in, set gToken and then exit with a http success code 200
    if (token) :        
        gToken = token
        return 200, 0, gToken  #1.24

    gToken = None
    
    # Token Header doesnt need token
    tokenHeaders = {'User-Agent':PIGS_USER_AGENT}

    # do SD token request with retries
    status_code,b,msg = DoPostUrl(SDapi+PostTokenURL, jsbody={"username":username,"password":sha1pwd}, newHeaders=tokenHeaders)

    # if its a 200 response, check the return code in b json
    
    if status_code == 200 :
        if b.get('code',9999) == 0:
            gToken = b['token']
            logging.info("SD:SDGetToken codes: %d, %d, Token:%s", status_code, b['code'], b['token'])
            return status_code, b['code'], gToken
        else :
            logging.warning("SD:SDGetToken codes: %d, %d, SD-MSG:%s", status_code, b['code'], b['message'])
            return status_code, b['code'], b.get('message')
    
    # It's not a 200 response if it gets here
    
    # Bad or No username/password is a 403 error.
    
    logging.warning("SD:SDGetToken http(%d) Message: %s", status_code, msg )    

    return status_code, b.get('code',999), msg



#==============================================================================
# Process Lineup changes on Schedules direct
# We need to make SD match  what user configured by deleting and/or adding lineups
#
# Get list of lineups from SD into  sdLuList
# for lu in sdLuList, if lu not in MyLineup, delete SD list
# get sdLuList
# for ea lu in MyLineups, if not in sdLuList, ADD it
#
# Caller must get token since we dont have un/pwd

# I want to do this >> if uil not in dLineups['lineups'] :
#https://stackoverflow.com/questions/18210436/python-custom-class-operator-overloading
#https://realpython.com/operator-function-overloading/
#https://stackoverflow.com/questions/38542543/functionality-of-python-in-vs-contains/38542777
#https://stackoverflow.com/questions/1964934/what-does-contains-do-what-can-call-contains-function
        #if uil not in dLineups['lineups'] :
        # Cant use if uil not in dLinupes[] bc we need to compare to sdl['lineup']
#{'ID': 1, 'alist': [{'lineup': '123'}, {'lineup': 'abc'}]}


def SDAddDeleteLineups( UiLineupList ) :
    
    logging.info("SD:SDAddDel: PiGS Lineups %s ", UiLineupList)

    # Get current account status with lineups
    httpcode, result = SDGetStatus()
    
    #print(repr(result))
    
    # 400 seems to be No lineups in account
    if httpcode != 200 : 
        logging.warning('SD:SDAddDel: Error getting SD Lineups (Maybe no lineups) Code:%d, %s', httpcode, result)

    # We have to continue b/c we may be adding the first lineup
    logging.info("SD:SDAddDel: Current Account Lineups %s", result)
    
    # First do Deletes - if item in the SD Account list is not in the UI list
    
    for lu in result :
        if lu not in UiLineupList :
            logging.info('SD:SDAddDel: Deleting SD Lineup: %s', lu )
            httpcode, sdcode = DeleteLineup(lu)

    # Next do Add's - For each lu in UiList, if not in SA account, Add it
    for uil in UiLineupList :
        if uil not in result :
            logging.info('SD:SDAddDel: Adding Lineup: %s', uil)
            httpcode, sdcode = AddLineup(uil)
            logging.debug('SD:SDAddDel: Adding Lineup: hpptcode %s, sdcode %s', httpcode, sdcode)

    # now print what the account has in it for testing
    
    httpcode, r2 = SDGetStatus()
    logging.info("SD:SDAddDel:New Lineups in Account: %s", r2)
        
    return httpcode


#==============================================================================
# A 'Result' class so that I can access the lineups[] list items and do :
#   if 'lineup' in SDStatusResult
# or
#   for lu in SDStatusResult :
# 

class SDStatusResult :
    def __init__(self, AccountStatus) :
        self._statusJson = AccountStatus   # Json result of the SD Status call
        self._lineups    = self._statusJson['lineups']
        self.code        = self._statusJson['code']
        #print("Init normal")
    
    def __contains__(self,item) :   # this allows 'in' to work
        for l in self._lineups :    # iterate the dict obs in lineups[]
            if item == l['lineup'] : return True
        return False   # 0 length list or not in list
    
    def __str__(self) :  # print a list of lineups ['lu1','lu2',...]
        s = '['
        for l in self._lineups :
            if len(s) >1: s += ','  # add comma's after the fist item
            s += l['lineup']
        return(s+']')
    
    def __repr__(self) :
        # no F f-strings so PiGS will work with Python 3.5
        #s = f'{self.__class__.__name__}(' f'code={self.code}, ' + self.__str__() + ')'
        s = "Class: " + self.__class__.__name__ + ". Code(" + str(self.code) + ") " + self.__str__()
        
        return s
    
    def __len__(self) :
        return len(self._lineups)  # len of a list[]
    
    # These next two dunders allow for loops to work
    def __iter__(self) :
        self.current = 0
        return self
    
    def __next__(self) :  # return the name of the lineup
        if self.current < len(self) :
            self.current += 1
            return self._lineups[self.current-1]['lineup']
            #return self._lineups[self.current-1]  # lineup dict
        else :
            raise StopIteration
    


def SDGetStatus() :
    
    global gToken
    
    logging.debug("SD:GetStatus ")
    
    # header needs the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}

    # send a GET request
    try :
        r1 = requests.get(SDapi+GetStatusURL, headers=newHeaders, timeout=REQ_TIMEOUT)
    except requests.exceptions.Timeout:
        return 700, 0
    except Exception as e :
        print('e=', e)
        return 800, 0
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
#        return r1.status_code, rj
        return r1.status_code, SDStatusResult(rj)
    
    if r1.status_code == 400 :  # prob no lineups to get
        rj = r1.json()
        logging.warning("SD:GetStatus http(%d) Message: %s", r1.status_code, rj )    
        return r1.status_code, None   # no lineups to send to user
    
    if r1.status_code == 500 :
        logging.warning("SD:GetStatus http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    # some other error
    #rj = r1.json()
    
    return r1.status_code, None  #rj.get('code')



#==============================================================================
#
def GetLineups() :
    
    global gToken
    
    logging.debug("SD:GetLineups ")
    
    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
    #myparams = {'country': 'USA', 'postalcode': zipcode}

    # send a GET request
    try :
        r1 = requests.get(SDapi+GetUserLineupsURL, headers=newHeaders, timeout=REQ_TIMEOUT)
    except requests.exceptions.Timeout:
        return 700, 0
    except Exception as e :
        print('e=', e)
        return 800, 0
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        return r1.status_code, rj
    
    if r1.status_code == 400 :  # prob no lineups to get
        rj = r1.json()
        logging.warning("SD:GetLineups http(%d) Message: %s", r1.status_code, rj )    
        return r1.status_code, None   # no lineups to send to user
    
    if r1.status_code == 500 :
        logging.warning("SD:GetLineups http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    # some other error
    #rj = r1.json()
    
    return r1.status_code, None  #rj.get('code')

#==============================================================================
#
def AddLineup(luname) :  # Add line upt to user account
    
    global gToken
    
    logging.debug("SD:AddLineup: %s", luname)

    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
    #myparams = {'country': 'USA', 'postalcode': zipcode}

    # send a PUT request to add the lineup
    try :
        r1 = requests.put(SDapi+PutUserLineupsURL + luname, headers=newHeaders, timeout=REQ_TIMEOUT)
    except requests.exceptions.Timeout:
        return 700, 0
    except Exception as e :
        print('e=', e)
        return 800, 0
    
    #print(r1.status_code)
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        logging.debug("SD:AddLineup: %d, %d %s", r1.status_code, rj.get('code'), rj.get('message') )
        logging.info("SD:AddLineup: %d changes remaining.", rj.get('changesRemaining') )
        return r1.status_code, rj.get('code')  #0=ok, #2105 invalide lineup
    
    if r1.status_code == 500 :
        logging.warning("SD:AddLineup http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    # some other error
    logging.warning("SD:AddLineup http(%d) Message: %s", r1.status_code, r1.text )    

    return r1.status_code, None  #rj.get('code')

#==============================================================================
#
def DeleteLineup(luname) :
    
    global gToken
    
    logging.debug("SD:DelLineup Lineup: %s", luname)
    
    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
    #myparams = {'country': 'USA', 'postalcode': zipcode}

    # send a Delete request
    try :
        r1 = requests.delete(SDapi+DelUserLineupURL + luname, headers=newHeaders, timeout=REQ_TIMEOUT)
    except requests.exceptions.Timeout:
        return 700, 0
    except Exception as e :
        print('e=', e)
        return 800, 0
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        logging.debug("SD:DelLineup http:%d, sd:%d %s", r1.status_code, rj.get('code'), rj.get('message') )
        #logging.info("SD:DelLineup: %d, %d %s", r1.status_code, rj.get('code'), rj.get('message') )
        logging.debug("SD:DelLineup: %d changes remaining.", rj.get('changesRemaining') )
        return r1.status_code, rj.get('code')
    
    if r1.status_code == 500 :
        logging.warning("SD:DelLineup http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    # some other error
    #rj = r1.json()
    
    return r1.status_code, None  #rj.get('code')

#==============================================================================

#==============================================================================

def SDGetStationMapping(LineupName) :

    global gToken
    
    logging.debug("SD:SDGetStationMapping Lineup: %s", LineupName)

    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
    #myparams = {'country': 'USA', 'postalcode': zipcode}

    # send a GET request
    try :
        r1 = requests.get(SDapi+GetLineupsMapURL + LineupName, headers=newHeaders, timeout=REQ_TIMEOUT)
    except requests.exceptions.Timeout:
        return 700, 0
    except Exception as e :
        print('e=', e)
        return 800, 0
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        return r1.status_code, rj
    
    if r1.status_code == 500 :
        logging.warning("SD:SDGetStationMapping http(%d) Message: %s", r1.status_code, r1.text )    
        return r1.status_code, None
    
    # some other error
    #rj = r1.json()
    
    return r1.status_code, None   #rj.get('code')


#==============================================================================
# 2-stage call
# 1st - using postalcode, get the headends, search for Antenna with matching Zip
# 2nd - ask for a lineup preview. 
#

def SDGetLineupPreview(zipcode) :

    global gToken
    
    
    # ensure canadian zipcodes are uppdercase to match SD's Headend results
    zipcode = zipcode.upper()
    
    # header need the token
    newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}
    if len(zipcode) == 5 :
        myparams   = {'country': 'USA', 'postalcode': zipcode}
    else :
        myparams   = {'country': 'CAN', 'postalcode': zipcode}
        
    #--------------------------------------------------------
    # send a GET request for the HEADENDS
    try :
        r0 = requests.get(SDapi+GetHeadendsURL, headers=newHeaders, params=myparams, timeout=REQ_TIMEOUT)
    except requests.exceptions.Timeout:
        return 700, 0, "SD Timeout"
    except Exception as e :
        print('e=', e)
        return 800, 0, "SD Exception"

    if r0.status_code != 200 :
        #logging.warning("SD:SDGetLineupPreview http(%d) SDcode(%d). Message: %s", r0.status_code, rj.get('code'), rj.get('message') )    
        logging.warning("SD:SDGetLineupPreview http(%d) ", r0.status_code )
        
        return r0.status_code, None, None
    
    rj=r0.json()
    
    # search for a matching HeadEnd
    for lu in rj :
        if (lu['headend'] == zipcode) :
            logging.info('SD:SDGetLineupPreview: Found postal code %s', zipcode)
            break
        
    if (lu['headend'] != zipcode) :
        logging.warning("SD:SDGetLineupPreview : NO HEADEND for ZIP(%s)", zipcode)
#        return r0.status_code, None, None
        return 600, None, "NO HEADEND for ZIP(" + zipcode + ")"
    
    # if more than one lineup in this postalcode, then what????
    if ( len(lu['lineups']) > 1) :
        logging.warning("SD:SDGetLineupPreview: Multiple Antenna lineups for the ZIP.  Using LU 0.")
        
    #logging.warning("SDGetLineupPreview http(%d) SDcode(%d). Message: %s", r1.status_code, ys.get('code'), ys.get('message') )    
    
    # response could also be an error block
    # search the response list for headend=zip, transport= "Antenna"
    # lineups[0]['lineup'].
    #   lineup 'USA-OTA-53593'
    #   name '53593'
    #   uri '/20141201/lineups/USA-OTA-53593'
    
    # Get the Lineup Name
    sdlineup = lu['lineups'][0]['lineup'] 

    # Append the 'lineup' name to /lineups/preview/ to get a full preview URL
    requrl = SDapi+GetPreviewLineupsURL + sdlineup
    
    # Send the requesst to SD to Get the Channel Lineup
    try :
        r1 = requests.get(requrl, headers=newHeaders, timeout=REQ_TIMEOUT)
    except requests.exceptions.Timeout:
        return 700, 0, "SD Timeout"
    except Exception as e :
        print('e=', e)
        return 800, 0, "SD Exception"
    
    # Could get a 200 code with this json response
    # "response":"LINEUP_QUEUED"
    # "code": 2107
    # "serverID": 20141201.web.1"
    # "message":"The lineup is being generated, please retry."
    # "datetime":"dateTtimeZ"
    
    # Handle success first
    if r1.status_code == 200 :
        rj = r1.json()
        return r1.status_code, rj, sdlineup
    
    if r1.status_code == 500 :
        logging.warning("SD:SDGetLineupPreview http(%d) Message: %s", r1.status_code, r1.text )    
        logging.warning("  Sometimes trying a nearby zip code can help... " )
        return r1.status_code, None, None

    # handle some other error like 403???
    rj = r1.json()
    return r1.status_code, rj.get('code'), rj.get('message')

from urllib.parse import urlparse

#==============================================================================
# This does up to 5 Post retries when the SD api fails
# Generally it returns http_status_code, json_result, some_message

def DoPostUrl(url, jsbody, newHeaders=None) :
    
    global gToken

    # Get SD command o.path from url
    o=urlparse(url)   # break url into scheme,netloc,path,params, query,fraqment
    #logging.debug("SD:DoPostUrl(%s):", o.path )

    # header need the token
    if not newHeaders :
        newHeaders = {'User-Agent':PIGS_USER_AGENT, 'Content-type':'text/plain','Accept':'gzip, deflate','token':gToken}

    #https://stackoverflow.com/questions/16511337/correct-way-to-try-except-using-python-requests-module
    
    SD_Retries = 5
    retrySecs  = SD_RETRY_SECS   # 2 secs
    retrySecs /= 2               # adjust for loop structure           
    
    while SD_Retries :

        SD_Retries -=1
        retrySecs  *= 2   # each retry waits 2x as long. 2,4,8,16,32 for a total of about 1 min
        
        try :
            
            r1 = requests.post(url, json=jsbody, headers=newHeaders, timeout=REQ_TIMEOUT)
            
        except requests.exceptions.Timeout:
            if SD_Retries == 0 :
                logging.warning("SD:DoPostUrl:TimeoutErr:%s http(%d) RTC(%d) Giving up.", o.path, 700, SD_Retries )
                return 700, None, "SD Timeout"
            else :
                logging.warning("SD:DoPostUrl:TimeoutErr:%s http(%d) RTC(%d) Retry in %ss", o.path, 700, SD_Retries, retrySecs )
                time.sleep(retrySecs)  #do a retry
                continue
            
        except requests.exceptions.ConnectionError:
            if SD_Retries == 0 :
                logging.warning("SD:DoPostUrl:ConnectErr:%s http(%d) RTC(%d) Giving up.", o.path, 710, SD_Retries )
                return 710, None, "SD ConnectionError"
            else :
                logging.warning("SD:DoPostUrl:ConnectErr:%s http(%d) RTC(%d) Retry in %ss", o.path, 710, SD_Retries, retrySecs )
                time.sleep(retrySecs)  #do a retry
                continue
            
        except requests.exceptions.RequestException as e:
            if SD_Retries == 0 :
                logging.warning("SD:DoPostUrl:OtherException:%s http(%d) RTC(%d) Giving up.", o.path, 720, SD_Retries )
                return 720, None, "SD Exception"
            else :
                logging.warning("SD:DoPostUrl:OtherException:%s http(%d) RTC(%d) Retry in %ss", o.path, 720, SD_Retries, retrySecs )
                time.sleep(retrySecs)  #do a retry
                continue
            
        else : # No Exception if we made it here...

            #if SD_Retries > 2 : r1.status_code = 509 # testing, simulate 50x error
            
            if r1.status_code >= 500 :      # Retry on 5xx Server Errors
                if SD_Retries > 0 :         # Any retries remaining?
                    logging.warning("SD:DoPostUrl:%s: ServerError: http(%d) RTC(%d) Retry in %ss", o.path, r1.status_code, SD_Retries, retrySecs )
                    time.sleep(retrySecs)  #do a retry
                    continue
                else :    # If no retries left, fall out of loop, msg log below
                    logging.warning("SD:DoPostUrl:%s ServerError: http(%d) RTC(%d) Giving up.", o.path, r1.status_code, SD_Retries )
                    pass
            else :  #  no retry on other http errors like 4xx
                SD_Retries = 0    # disable retry counter to end loop

    # depending on http code, did the response have body json?
    
    if r1.status_code == 200 :
        logging.debug("SD:DoPostUrl:%s http(%d) Message: %s", o.path, r1.status_code, "OK" )    
        return r1.status_code, r1.json(), "OK"
    
    # may or may not have valid json
    try :
        rj = r1.json()  # try to get json.... 
    except :
        logging.debug("SD:DoPostUrl:%s http(%d) json Exception, r1:%s", o.path, r1.status_code, r1.text)
        rj = { 'code': r1.status_code }   # create some valid json with a SD type 'code'
    
    # if here, its some <>200 http error, like 40x, 50x, etc
    
    logging.warning("SD:DoPostUrl:%s Error http(%d) Message: %s", o.path, r1.status_code, r1.text )
    
    return r1.status_code, rj, r1.text 


#==============================================================================
# get data for all days available days=0, or n days if days >0 

def SDGetProgramDetails(pidlist) :
    
    # Do SD request with retries
    status_code,js,msg = DoPostUrl(SDapi+PostProgramDetailsURL, jsbody=pidlist)

    if status_code != 200 :
        logging.warning("SD:SDGetProgramDetails http(%d) Message: %s", status_code, msg )    

    return status_code, js, msg



#==============================================================================
# get data for all days available days=0, or n days if days >0 

from datetime import timedelta
from datetime import datetime
from time import gmtime, strftime, mktime

def SDGetStationSchedule(stationID, ndays) :
    
    # To handle InProgress requests that occur near GMT=0, we need the previous days listings
    # This is a boundary condition but for the USA it affects 'Prime time.'
    # at CST of -5Z GMT 'today' starts ~7pm cst, so a 6pm(18:00) InProgress show is on
    # 'yesterdays' 23 GMT data

    # recommended way to create an object representing the current time in UTC is by calling
    # t=datetime.utcnow()
    
    # build a request for yesterday and all days
    #t = datetime.now(timezone.utc)
    t  = datetime.utcnow()       # today
    td = t.strftime("%Y-%m-%d")  # td todays date as a string
    y  = t + timedelta(days=-1)  # yesterday is today -1 day
    yd = y.strftime("%Y-%m-%d")  # yd yesterday as a string
    
    # Sched Direct api allows for multiple station id's / dates in 1 call
    # I'm using this to request a single day / yesterday in 1 line
    # ommiting the date in line 2 requests all dates available
    # duplicating dates is handled by the SD server
    sl = [ {"stationID": stationID, 'date':[yd] },  # yesterday
           {"stationID": stationID } ]              # and everything else

    #sl = [ {"stationID": stationID } ]              # and everything else

    # for debugging, limit data to 'ndays' of program data
    # yes the dates are added in reverse order but SD does not seem to care

    if ndays>0 :
        # make a new sl object structure with 1 station, yesterday plus ndays specific dates
        sl = [ {"stationID": stationID, 'date':[yd] }]
        while ndays :
            dy = t + timedelta(days=ndays-1)
            sl[0]['date'].append( dy.strftime("%Y-%m-%d") )
            ndays -= 1          

    #logging.info("SD:SDGetStationSchedule DaysList: %s", sl )
    
    # Try SD request with retries
                                               
    status_code,js,msg = DoPostUrl(SDapi+PostSchedulesURL, jsbody=sl)

    # an outdated token would give a server error and no data

    if status_code != 200 :
        logging.warning("SD:SDGetStationSchedule http(%d) js(%s) Message: %s", status_code, js, msg )    

    return status_code, js


