# ut.py
#
# Utility Functions to support PiGS application
#
# Copyright July 2020, Timothy Goihl
#
# return local time string and timezone minutes


#---------------------------------------------------------------------------------------
# this helps put the log file and config files in the pigs.py directory
# even when app is launched when current working dir is '/'

import os
from os import path

def PigsPath(fn) :
    
    pigspypath = os.path.realpath(__file__)
    
    # separate file path/name from its extention
    #root, ext  = path.splitext(configFile)
    # separate path from 'pigs.py'
    head, tail = os.path.split(pigspypath)

    return os.path.join(head, fn)


import time
from datetime import timedelta, datetime

#https://docs.python.org/3.5/library/time.html

#Returns a LocalTime String 2020-07-15 23:11:27(CDT)
#and utc offset integer in minutes, like -300
#usage: timestr,offset = LocalTimeStr( datetime.utcnow() ) 
# doing my own local time convesion for my 'unaware' times

def LocalTimeStr(at, showSeconds=False, showDate=True, showTimeZone=True, WeekDay=True) :
    
    # fixed in 2.13 to display correct local-time
    t = time.localtime()
    tzstr = "(" + t.tm_zone + ")"
    tzmins = int(t.tm_gmtoff/60)
    localTime = at + timedelta(minutes=tzmins)
    
    dispFmt=''
    WkDyYr = "%a " if WeekDay else "%Y-"  # show Mon- or 2020- 
    if showDate : dispFmt += WkDyYr + '%m-%d '
    dispFmt += '%H:%M'
    if showSeconds : dispFmt += ':%S'
    
    lts = localTime.strftime(dispFmt)
    
    if showTimeZone : lts += tzstr
        
    return( lts, tzmins )


def VersionStringToFloat(vStr) :
    version = 0.0
    firstdp = -1
     
    for c in vStr :   # convert '5.4.1' into 5.41 for comparisons
        if c.isdigit() :
            version = version*10 + int(c)
            if firstdp > -1 : firstdp += 1  # count number of digits after first DP
        elif c == '.' and firstdp == -1 : firstdp = 0
    version = version / (10**firstdp)        

    return version


import requests

def test_WebServer(address) :  # addr str '192.168.1.200' or '192.168.1.200:80' 

    # send a HEAD request to see if PiGS responds
    try :
        #REQ_TIMEOUT = (30.0, 90.0)  # Connection Timeout=30, Read Timeout = 90
        # Doing a quick timeout since we expect it to respond and its on same PC
        r1 = requests.head('http://'+address, timeout=(0.5, 2.0) )
    except requests.exceptions.Timeout:
        logging.debug("Connection to %s failed.  Timeout!", address )
        return -1
    except Exception as e :
        logging.debug("Connection to %s failed.  Exception: %s", address,e )
        return -2

    return 0


def utGetPigsVersion(address = 'http://pigs.goihl.com/zips/version.txt') :

    # send a Get request for current version on server
    try :
        # REQ_TIMEOUT = (30.0, 90.0)  # Connection Timeout=30, Read Timeout = 90
        myheaders = {'Accept': 'text/html'}
        r = requests.get(address, timeout=(1.5, 5.0) ) #headers=myheaders
    except requests.exceptions.Timeout:
        logging.debug("Connection to %s:80 failed. Timeout!", address )
        return -1.0
    except Exception as e :
        logging.debug("Connection to %s:80 failed.  Exception: %s", address,e )
        return -2.0
    
    #print(r.status_code, r.text, r.headers)
    
    if r.status_code == 200 :
        return float(r.text)
    else :
        logging.info("UT: Error reading PiGS server version %s", r.status_code )
        return -3.0



# adapted from a stackoverflow post
# used to test if pigs webserver is responding
import socket
def test_serverport(address, port) :
    
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM) # create socket
    
    s.settimeout(0.2)
    
    #e = s.connect_ex((address, port))
    rv = False
    
    try: # to make a connection
        s.connect((address, port))
        rv = True
    except socket.error as e:
        logging.debug("Connection to %s:%s failed! (%s)", address, port, e )
        rv = False
    finally: # always executed
        s.close()
        
    return rv

#Function to get total size of Airing list
#https://code.activestate.com/recipes/577504/

from sys import getsizeof, stderr
from itertools import chain
from collections import deque
try:
    from reprlib import repr
except ImportError:
    pass

def total_size(o, handlers={}, verbose=False):
    """ Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    """
    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {tuple: iter,
                    list: iter,
                    deque: iter,
                    dict: dict_handler,
                    set: iter,
                    frozenset: iter,
                   }
    all_handlers.update(handlers)     # user handlers take precedence
    seen = set()                      # track which object id's have already been seen
    default_size = getsizeof(0)       # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:       # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)



# https://stackabuse.com/how-to-send-emails-with-gmail-using-python/
# https://docs.python.org/3.5/library/smtplib.html

# in your gmail account you must turn on allow-less-secure-apps or this will fail

import smtplib
import logging

gmail_user = ""
gmail_password = ""
toEmails = ""
emailIP = ""

def EM_Initialize(xgmail_user, xgmail_password, xtoEmails, PigsIP) :
    
    global gmail_user
    global gmail_password
    global toEmails
    global emailIP    
    
    gmail_user     = xgmail_user
    gmail_password = xgmail_password
    toEmails       = xtoEmails
    emailIP        = PigsIP

def SendEmailAlert(alertmsg, subjectMsg ='', UpdateMsg='') :
    
    global gmail_user
    global gmail_password
    global toEmails
    global emailIP    
   
    if toEmails == "" :  # exit if no email setup
        return
    
    if UpdateMsg : subjectMsg = ', Update Available'
    
    sent_from = gmail_user
    subject = 'PiGS Server Notification' + subjectMsg
    body = alertmsg
    body += "\r\n\r\nLaunch PiGS http://"+ emailIP + "\r\n"
    body += "PiGS Status http://"+ emailIP + "/pigstatus\r\n"
    if UpdateMsg :
        body += "\n" + UpdateMsg + "\r\n"

    email_text = "From: PiGS Server<%s>\r\nTo: %s\r\nSubject: %s\r\n\r\n%s\r\n\r\n" \
         % (sent_from, toEmails, subject, body)
#         % (sent_from, ", ".join(toEmails), subject, body)

    try:
        server = smtplib.SMTP_SSL('smtp.gmail.com', 465)
        server.ehlo()
        server.login(gmail_user, gmail_password)
        server.sendmail(sent_from, toEmails, email_text)
        server.close()
        logging.debug('UT:EML: Alert Email sent! %s', alertmsg)
    except Exception as err:
        logging.warning('UT:EML: Something went wrong...  %s', err)

#SendEmailAlert(gmail_u, gmail_p,toEms,"PiGS Alert Text")
